#' @include type_robiatomic.R
#' @importFrom  vctrs vec_data vec_cast vec_default_ptype2 vec_default_cast new_factor
#' @importFrom glue glue
#' @import stringr
#' @import tibble
NULL

#' Vector of 4mers for multiplexing PCR.
#'
#' A `robi4mer` is character vector that is for storing short oligo nucleotides
#' used to 4mer sample. 4mers are always represented as uppercase letters and
#' contain only A, C, G, and T and every 4mer
#'
#' @details The `robi4mer()` constructor builds a default `robi4mer` instance of
#'   length defined by the `length` argument and composed of 4mers of length
#'   zero.
#'
#' @param length A non-negative integer specifying the desired length. Double
#'   values will be coerced to integer: supplying an argument of length other
#'   than one is an error.
#'
#' @return the `robi4mer()`and `new_robi4mer()` functions return a `robi4mer`
#'   instance.
#'
#' @examples
#'
#' rt <- robi4mer(10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
#' @md
robi4mer <- function(length = 0) {
  new_robi4mer(character(length))
}


#' @details The contructor `new_robi4mer` is the main function to create a
#'   `robi4mer` instance. For simpler conversion to that class, it can be easier
#'   to use the `as_robi4mer()` utility function.
#'
#' @param object a character vector or something coercible.
#' @param adapte_length if `TRUE` longuest 4mers are shortened by trimming them
#'   on the left side.
#' @param ... Passed on to [`base::structure()`]. It allows for defining
#'   attributes, specified in 4mer = value form, which will be attached to data
#'   (see [`robiatomic`]).
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   constructor takes some decision because of missing values and estimates
#'   default values these missing data.
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#'
#' rt <- new_robi4mer(c("ATC", "ATG", "CCC", "CTG"))
#'
#' @rdname robi4mer
#' @export
#' @md
new_robi4mer <- function(object, adapte_length=TRUE,
                        ...,
                        verbose = is_robi_verbose(),
                        class = character()) {

  object <- toupper(vec_cast(object,character()))

  lx <- nchar(object)
  lx_min <- if (length(object) == 0) 0 else min(lx, na.rm = TRUE)

  if (lx_min > 0 && any(lx != lx_min  & !is.na(object))) {
    if (adapte_length) {
      if (verbose)
        warning(sprintf("Every 4mers have not the same length. They will be all shortened to %dbp",
                        lx_min))

      object <- str_sub(object,-lx_min,-1)
    }
  }

  validate_robiobject(
    new_robiatomic(as.character(object),
      ...,
      class = c(class, "robi4mer")
    )
  )
}

#' @details The [`robi4mer`] version of that method ckecks that :
#'
#' * All the 4mers have the same length
#' * Every 4mers are composed only of upper case A, C, G, or T
#'
#' @examples
#'
#' rt <- new_robi4mer(c("AGT","GGT","CGT","GTA","CCC"))
#' validate_robiobject(rt)
#'
#' @rdname validate_robiobject
#' @export
#' @md
validate_robiobject.robi4mer <- function(object) {

  lx <- nchar(object)
  same_length <- all(duplicated(lx)[-1])

  if (!same_length) {
    strange = object[which(duplicated(lx)[-1]) + 1]
    if (length(strange) > 20)
      strange = c(strange[1:19], "and others...")
  }

  robiassert_arg(same_length,
                 "object",
                 "Every 4mers must have the same length ({nchar(object[1])}). ({paste(strange,collapse=', ')} seem different)")


  if (any(is.na(str_match(object,"^[ACGT]*$")) &&
          !is.na(object) &&
          object != "__NA__")) {
    stop("Only uppercase A, C, G, and T are allowed in 4mers.",
    call. = TRUE
    )
  }

  object
}


#' @@details `is_robi4mer` tests if an object belongs the `robi4mer` class
#'
#' @param `object` the object to test.
#'
#' @return the `is_robi4mer()`, `is_robi4mer_forward()`, and `is_robi4mer_reverse()`
#'    functions return a [`logical`] value.
#'
#' @examples
#'
#' is_robi4mer(rt)
#' is_robi4mer("toto")
#'
#' @rdname robi4mer
#' @export
#' @md
is_robi4mer = function(object) {
  inherits(object,'robi4mer')
}

#' @details `as_robi4mer` converts an object to a `robi4mer` instance.
#'
#' @param object the object to convert.
#'
#' @return the `as_robi4mer()` funtion returns a `robi4mer` value.
#'
#' @examples
#'
#' x <- c("AGTT","GGCT","CCGT","GATA","CCCC")
#' rt <- as_robi4mer(x)
#'
#' @rdname robi4mer
#' @export
#' @md
as_robi4mer <- function(object) {
  vec_cast(object,robi4mer())
}


#' @export
lcs_score <- function(x,y,lx=nchar(x),ly=nchar(y),
                      threshold = -1,
                      normalize = FALSE,
                      reference = c("alignment","longest","shortest"),
                      similarity_mode = c("similarity","distance"),
                      lcs=FALSE,alilength=FALSE) {

  data <- tibble(x=x,y=y,lx=as.integer(lx),ly=as.integer(ly))

  theshold  <- as.double(threshold[1])
  normalize <- as.logical(normalize[1])
  reference <- as.integer(match(as.character(reference[1]),
                                c("alignment","longest","shortest"),
                                nomatch = 0) - 1)

  robiassert_arg(reference>=0,
                 "reference",
                 "must be one of the following values : \"alignment\",\"longest\",\"shortest\"")

  similarity_mode <- match(as.character(similarity_mode[1]),c("similarity","distance"),0)

  robiassert_arg(similarity_mode>0,
                 "similarity_mode",
                 "must be one of the following values : \"similarity\",\"distance\"")

  similarity_mode <- similarity_mode == 1

  lcs <- as.logical(lcs[1])
  alilength <- as.logical(alilength[1])


  results = .Call("ROBI_lcs",
                  data$x,data$y,data$lx,data$ly,
                  threshold,normalize,reference,similarity_mode,
                  lcs,alilength)

  n <- "score"
  if (lcs) n <- c(n,"lcs")
  if (alilength) n <- c(n,"alignment_length")

  names(results) <- n

  do.call(tibble,results);
}

#' @export
robidnadist <- function(x,y,lx=nchar(x),ly=nchar(y)) {

  data <- tibble(x=x,y=y,lx=as.integer(lx),ly=as.integer(ly))
  results = .Call("ROBI_dnadist",data$x,data$y,data$lx,data$ly)

  results
}
