
#' Qualitative color schemes by Paul Tol.
#'
#' Provides a set of color palettes covenient even for blind color people
#'
#' @param n the minimal number of color you are interested in
#' @param alpha the transparency value between $0$ full transparency
#'        and $1$ for opaque colors.
#' @return a character vector containing hexadecimal color codes
#'
#' @examples
#'     x = colors_tol(10,0.5)
#'     x
#'     plot(1:10,1:10,col=x,cex=1:10,pch=16)
#'
#' @seealso  <http://www.sron.nl/~pault/>
#' @author Eric Coissac
#' @export
colors_tol = function(n,alpha=1) {

  color_tol = list(
    "1" = c("#4477AA"),

    "2" = c("#4477AA", "#CC6677"),

    "3" = c("#4477AA", "#DDCC77", "#CC6677"),

    "4" = c("#4477AA", "#117733", "#DDCC77", "#CC6677"),

    "5" = c("#332288", "#88CCEE", "#117733", "#DDCC77",
            "#CC6677"),

    "6" = c("#332288", "#88CCEE", "#117733", "#DDCC77",
            "#CC6677","#AA4499"),

    "7" = c("#332288", "#88CCEE", "#44AA99", "#117733",
            "#DDCC77", "#CC6677","#AA4499"),

    "8" = c("#332288", "#88CCEE", "#44AA99", "#117733",
            "#999933", "#DDCC77", "#CC6677","#AA4499"),

    "9" = c("#332288", "#88CCEE", "#44AA99", "#117733",
            "#999933", "#DDCC77", "#CC6677", "#882255",
            "#AA4499"),

    "10" = c("#332288", "#88CCEE", "#44AA99", "#117733",
             "#999933", "#DDCC77", "#661100", "#CC6677",
             "#882255", "#AA4499"),

    "11" = c("#332288", "#6699CC", "#88CCEE", "#44AA99",
             "#117733", "#999933", "#DDCC77", "#661100",
             "#CC6677", "#882255", "#AA4499"),

    "12" = c("#332288", "#6699CC", "#88CCEE", "#44AA99",
             "#117733", "#999933", "#DDCC77", "#661100",
             "#CC6677", "#AA4466", "#882255", "#AA4499"),

    "14" = c("#882E72", "#B178A6", "#D6C1DE", "#1965B0",
             "#5289C7", "#7BAFDE", "#4EB265", "#90C987",
             "#CAE0AB", "#F7EE55", "#F6C141", "#F1932D",
             "#E8601C", "#DC050C"),

    "15" = c("#114477", "#4477AA", "#77AADD", "#117755",
             "#44AA88", "#99CCBB", "#777711", "#AAAA44",
             "#DDDD77", "#771111", "#AA4444", "#DD7777",
             "#771144", "#AA4477", "#DD77AA"),

    "18" = c("#771155", "#AA4488", "#CC99BB", "#114477",
             "#4477AA", "#77AADD", "#117777", "#44AAAA",
             "#77CCCC", "#777711", "#AAAA44", "#DDDD77",
             "#774411", "#AA7744", "#DDAA77", "#771122",
             "#AA4455", "#DD7788"),

    "21" = c("#771155", "#AA4488", "#CC99BB", "#114477",
             "#4477AA", "#77AADD", "#117777", "#44AAAA",
             "#77CCCC", "#117744", "#44AA77", "#88CCAA",
             "#777711", "#AAAA44", "#DDDD77", "#774411",
             "#AA7744", "#DDAA77", "#771122", "#AA4455",
             "#DD7788")
  )

  if (n > 21)
    stop("Only color schemas with at maximum 21 colors are handled")

  limits = as.integer(names(color_tol))

  n_tol = as.character(limits[n <= limits][1])
  tol   = color_tol[[n_tol]]

  hex_alpha = toupper(as.hexmode(floor(alpha * 255)))
  tol_alpha = paste(tol,hex_alpha,sep="")

  return(tol_alpha)
}


#' Colors for variations around 0 (blue-red)
#'
#' @rdname colors_tol
#' @export

gradient_tol1 = function(x,alpha=1) {
  rcol = function(x) 0.237 - 2.13*x + 26.92*x**2 - 65.5*x**3 + 63.5*x**4 - 22.36*x**5
  gcol = function(x) ((0.572 + 1.524*x - 1.811*x**2)/(1 - 0.291*x + 0.1574*x**2))**2
  bcol = function(x) 1/(1.579 - 4.03*x + 12.92*x**2 - 31.4*x**3 + 48.6*x**4 - 23.36*x**5)

  return(rgb(rcol(x),gcol(x),bcol(x),alpha))
}

#' Gradient colors,
#'
#' @rdname colors_tol
#' @export

gradient_tol2 = function(x,alpha=1) {
  rcol = function(x) (1 - 0.392*(1 + erf((x - 0.869) / 0.255)))
  gcol = function(x) (1.021 - 0.456*(1 + erf((x - 0.527) / 0.376)))
  bcol = function(x) (1 - 0.493*(1 + erf((x - 0.272) / 0.309)))

  return(rgb(rcol(x),gcol(x),bcol(x),alpha))
}

#' Rainbow gradient colors,
#'
#' @rdname colors_tol
#' @export
rainbow_tol = function(x,alpha=1) {
  rcol = function(x) (0.472 - 0.567 * x + 4.05 * x ** 2) / (1. + 8.72 * x - 19.17 * x ** 2 + 14.1 * x ** 3)
  gcol = function(x) 0.108932 - 1.22635 * x + 27.284 * x ** 2 - 98.577 * x ** 3 + 163.3 * x ** 4 - 131.395 * x ** 5 + 40.634 * x ** 6
  bcol = function(x) 1./(1.97 + 3.54 * x - 68.5 * x ** 2 + 243 * x ** 3 - 297 * x ** 4 + 125 * x ** 5)

  return(rgb(rcol(x), gcol(x), bcol(x), alpha))
}
