#' @include robimetabar.R
#' @import dplyr
#'
NULL

#' Computes the Hellinger transformation of the read counts
#'
#' On a per sample base, the square root of the relative frequency of each
#' *MOTU*.
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **relative_frequency** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#'
#' @seealso `relative_frequencies()`, `cum_relative_frequencies()`,
#'          `motus_ranks()`
#'
#' @return the modified `robimetabar` instance.
#' @export
#'
#' @examples
#' @md
hellinger <- function(metabar, results_in) {
  if (missing(results_in)) {
    results_in <- "hellinger"
  }

  metabar$data %>%
    ungroup() %>%
    group_by(sample) %>%
    mutate(!!results_in := sqrt(count/sum(count))) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}

#' Computes the relative frequencies of the read counts
#'
#' On a per sample base, the relative frequency of each
#' *MOTU* is computed by dividing the read count of the motu for that sample
#' by the total read count of every *MOTUs* occurring in the given sample.
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **relative_frequency** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#'
#' @seealso `hellinger()`, `cum_relative_frequencies()`,
#'          `motus_ranks()`
#'
#' @return the modified `robimetabar` instance.
#' @export
#'
#' @examples
#' @md
relative_frequencies <- function(metabar, results_in) {
  if (missing(results_in)) {
    results_in <- "relative_frequency"
  }

  metabar$data %>%
    ungroup() %>%
    group_by(sample) %>%
    mutate(!!results_in := count/sum(count)) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}

#' Computes the cumulative relative frequencies in a sample.
#'
#' MOTUs are firstly ordered by decreasing read abundances,
#' sample per sample and and then the cumulative relative
#' frequency of each *MOTU* summed to the more abundant is
#' estimated.
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **cum_rel_freq** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#'
#' @return the modified `robimetabar` instance.
#' @export
#'
#' @seealso `relative_frequencies()`, `hellinger()`,
#'          `motus_ranks()`
#'
#' @examples
#' @md
cum_relative_frequencies <- function(metabar, results_in) {
  if (missing(results_in)) {
    results_in <- "cum_rel_freq"
  }

  metabar$data %>%
    ungroup() %>%
    arrange(sample,desc(count)) %>%
    group_by(sample) %>%
    mutate(!! results_in := cumsum(count/sum(count))) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}

#' Computes the cumulative relative frequencies in a sample.
#'
#' MOTUs are firstly ordered by decreasing read abundances,
#' sample per sample and and then the cumulative relative
#' frequency of each *MOTU* summed to the more abundant is
#' estimated.
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **cum_hill** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#'
#' @return the modified `robimetabar` instance.
#' @export
#'
#' @seealso `relative_frequencies()`, `hellinger()`,
#'          `motus_ranks()`
#'
#' @examples
#' @md
cum_hill <- function(metabar, q=1, results_in) {

  .cumhill <- function(x) sapply(1:length(x),
                                 function(i) D_q(x[1:i]))

  if (missing(results_in)) {
    results_in <- "cum_hill"
  }

  metabar$data %>%
    ungroup() %>%
    arrange(sample,desc(count)) %>%
    group_by(sample) %>%
    mutate(!! results_in := .cumhill(count)) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}

#' Computes the cumulative relative frequencies in a sample.
#'
#' MOTUs are firstly ordered by decreasing read abundances,
#' sample per sample and and then the cumulative relative
#' frequency of each *MOTU* summed to the more abundant is
#' estimated.
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **cum_hill** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#'
#' @return the modified `robimetabar` instance.
#' @export
#'
#' @seealso `relative_frequencies()`, `hellinger()`,
#'          `motus_ranks()`
#'
#' @examples
#' @md
cum_hill <- function(metabar, q=1, results_in) {

  .cumhill <- function(x) sapply(1:length(x),
                                 function(i) D_q(x[1:i]))

  if (missing(results_in)) {
    results_in <- "relative_cum_hill"
  }

  metabar$data %>%
    ungroup() %>%
    arrange(sample,desc(count)) %>%
    group_by(sample) %>%
    mutate(!! results_in := .cumhill(count)) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}

#' Computes the cumulative relative frequencies in a sample.
#'
#' MOTUs are firstly ordered by decreasing read abundances,
#' sample per sample and and then the cumulative relative
#' frequency of each *MOTU* summed to the more abundant is
#' estimated.
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **cum_hill** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#'
#' @return the modified `robimetabar` instance.
#' @export
#'
#' @seealso `relative_frequencies()`, `hellinger()`,
#'          `motus_ranks()`
#'
#' @examples
#' @md
relative_cum_hill <- function(metabar, q=1, results_in) {

  .relcumhill <- function(x) {
    total = D_q(x,q=q)
    sapply(1:length(x),
           function(i) D_q(x[1:i],q=q)/total)
  }

  if (missing(results_in)) {
    results_in <- "relative_cum_hill"
  }

  metabar$data %>%
    ungroup() %>%
    arrange(sample,desc(count)) %>%
    group_by(sample) %>%
    mutate(!! results_in := .relcumhill(count)) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}


#' Computes the rank of each MOTU in a sample.
#'
#'
#' This function is modifying the `robimetabar` instance provided
#' as input.
#'
#' @param metabar a `robimetabar` instance.
#' @param results_in by default the result of the computation
#'                   is stored in the **motus_rank** column of
#'                   the data table.. Name of the destination column
#'                   can be changed by providing the desired name through
#'                   that argument.
#' @export
#'
#' @seealso `relative_frequencies()`, `hellinger()`,
#'          `cum_relative_frequencies()`
#'
#' @examples
motus_ranks <- function(metabar, results_in) {
  if (missing(results_in)) {
    results_in <- "motus_rank"
  }

  metabar$data %>%
    ungroup() %>%
    group_by(sample) %>%
    mutate(!!results_in := order(count,decreasing = TRUE)) %>%
    ungroup() -> metabar$data

  invisible(metabar)
}

