#' Returns sample ids column name
#'
#' The structure describing samples in a metabar object
#' contains a column containing the sample's names.
#' \code{sample_ids_name} returns the name of that column.
#'
#' @param data a metabar object
#' @return a charactere value
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#'   metabar <- robimetabar(reads = reads)
#'   sample_ids_name(metabar)
#'
#' @seealso \code{\link[ROBITools2]{motu_ids_name}},\code{\link[ROBITools2]{motu_ids}},
#'          \code{\link[ROBITools2]{sample_ids}}
#'
#' @author Eric Coissac
#' @export
samples_ids_colname <- function(data) {
  UseMethod("samples_ids_colname", data)
}

#' @rdname samples_ids_colname
#' @author Eric Coissac
#' @export
samples_ids <- function(data) {
  UseMethod("samples_ids", data)
}

#' Returns motu ids column name
#'
#' The structure describing MOTUs in a metabar object
#' contains a column containing the MOTU's names.
#' \code{motu_ids_name} returns the name of that column.
#'
#' @param data a metabar object
#' @return a charactere value
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#'   metabar <- robimetabar(reads = reads)
#'   motus_ids_colname(metabar)
#'
#' @seealso \code{\link[ROBITools2]{sample_ids_names}},
#'          \code{\link[ROBITools2]{sample_ids}},
#'          \code{\link[ROBITools2]{motus_ids}}
#'
#' @author Eric Coissac
#' @export
motus_ids_colname <- function(data) {
  UseMethod("motus_ids_colname", data)
}

#' @rdname motus_ids_colname
#' @author Eric Coissac
#' @export
motus_ids <- function(data) {
  UseMethod("motus_ids", data)
}


#' Returns number of MOTUs
#'
#' Returns the number of MOTUs stored in a metabar object.
#' That number corresponds to the culumn count of the contengency matrix.
#'
#' @param data a metabar object
#' @return an numeric value
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#'   metabar <- robimetabar(reads = reads)
#'   nmotus(metabar)
#'
#' @seealso \code{\link[ROBITools2]{nsamples}}
#' @author Eric Coissac
#' @export
`nmotus` <- function(data) {
  UseMethod("nmotus", data)
}


#' Returns number of samples
#'
#' Returns the number of samples stored in a metabar object.
#' That number corresponds to the row count of the contengency matrix.
#'
#' @param data a metabar object
#' @return an numeric value
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#'   metabar <- robimetabar(reads = reads)
#'   nsamples(metabar)
#'
#' @seealso \code{\link[ROBITools2]{nmotus}}
#' @author Eric Coissac
#' @export
`nsamples` <- function(data) {
  UseMethod("nsamples", data)
}

#' @param data
#'
#' @author Eric Coissac
#' @export
#'
#' @example
motus <- function(data) {
  UseMethod("motus", data)
}

#' @rdname motus
#' @export
store_into_motus <- function(value, data) {
  UseMethod("store_into_motus", value)
}

#' @rdname motus
#' @export
`motus<-` <- function(data, value ) {
  value %>% store_into_motus(data)

  invisible(data)
}

#' @rdname motus
#' @export
samples <- function(data) {
  UseMethod("samples", data)
}

#' @rdname motus
#' @export
store_into_samples <- function(value, data) {
  UseMethod("store_into_samples", value)
}

#' @rdname motus
#' @export
`samples<-` <- function(data, value ) {
  value %>% store_into_samples(data)

  invisible(data)
}

#' @export
store_into <- function(value, data) {
  UseMethod("store_into", value)
}


