#' @include taxonomy_robi.R
#'
NULL

#' Returns taxids associated to the names.
#'
#' Return the set of taxids having their name matching the given pattern.
#'
#' @param taxonomy an object able to provide a taxonomy.
#' @param patterns one or several regular pattern used to select the the taxa.
#' @param   rank a \code{character} indicating a taxonomic rank. If not \code{NULL}
#'               only taxids correponding to this rank are returned.
#' @param   alternative  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the function must only look for a scientific name.
#' @param exact
#' @param full_length
#' @param case_sensitive
#'
#' @return if just one pattern is given, an integer vector is returned with the
#'         corresponding taxids. If a list of patterns is given, the function
#'         returns a list of integer vectors, each vector containing the taxids
#'         corresponding to a pattern. The returned list is in the same order
#'         than the given patern list.
#'
#' @author Eric Coissac
#' @export
ecofind = function(patterns,
                   taxonomy,
                   rank = NULL,
                   subclade_of = NULL,
                   alternative = FALSE,
                   exact = TRUE,
                   full_length = FALSE,
                   case_sensitive = FALSE) {
  UseMethod("ecofind", taxonomy)
}


#' @export
ecofind.robitaxonomy <- function(patterns,
                                 taxonomy,
                                 rank = NULL,
                                 subclade_of = NULL,
                                 alternative = FALSE,
                                 exact = TRUE,
                                 full_length = TRUE,
                                 case_sensitive = FALSE) {
  if (full_length) {
    if (exact) {
      if (case_sensitive) {
        matcher <- function(pattern) {
          nodes(taxonomy)$scientific_name == pattern
        }
      } else {
        matcher <- function(pattern) {
          toupper(nodes(taxonomy)$scientific_name) == toupper(pattern)
        }
      }
    } else {
      matcher <- function(pattern) {
        str_detect(
          nodes(taxonomy)$scientific_name,
          regex(paste0("^", pattern, "$"),
            ignore_case = !case_sensitive
          )
        )
      }
    }
  } else
  if (exact) {
    matcher <- function(pattern) {
      str_detect(
        nodes(taxonomy)$scientific_name,
        fixed(pattern,
          ignore_case = !case_sensitive
        )
      )
    }
  } else {
    matcher <- function(pattern) {
      str_detect(
        nodes(taxonomy)$scientific_name,
        regex(pattern,
          ignore_case = !case_sensitive
        )
      )
    }
  }


  if (.has_doParallel && getDoParRegistered()) {
    `%dp%` <- `%dopar%`
  }
  else {
    `%dp%` <- `%do%`
  }

  foreach(
    i = seq_along(patterns),
    .combine = bind_rows
  ) %dp% {
    nodes(taxonomy)[matcher(patterns[[i]]), ] %>%
      mutate(pattern = patterns[[i]])
  } -> candidates

  irank <- rank
  if (!is.null(irank)) {
    candidates %>%
      dplyr::filter(rank == irank) -> candidates
  }

  if (!is.null(subclade_of)) {
    candidates %>%
      dplyr::filter(is_subclade_of(data = taxid,
                                   taxonomy=taxonomy,
                                   parent = subclade_of,
                                   results_in = "ok")$ok) -> candidates
  }

  candidates %>%
    group_by(pattern) %>%
    summarise(taxid=list(taxid),
              scientific_name=list(scientific_name)) %>%
    as.list() -> candidates


  nc <- candidates$pattern
  candidates$pattern <- NULL
  candidates <- mapply(function(t,n) {names(t)<- n
                                      t <- as_robitaxid(t)
                                      attr(t,"reference_taxonomy") <- taxonomy
                                      t
                                     },
                candidates$taxid,
                candidates$scientific_name,
                SIMPLIFY = FALSE)
  names(candidates) <- nc

  candidates
}

