
#' Merge the id parsing tibble with the original  \code{\link[ROBITools2]{robisample}}.
#'
#' Internal function. Please do not use.
#'
#' @param i the id parsing tibble
#'
#' @param object the object to parse a \code{\link[ROBITools2]{robisample}},
#'        or a \code{\link[ROBITools2]{robimetabar}} instance.
#' @param parts a string vector indicating the name of each sub-pattern
#'        produced. It can include or not a name for the last sub-pattern,
#'        that actually corresponds to the full id. if the \code{include_full}
#'        parameter is set to \code{FALSE} don't provide name for the full id.
#' @param n an integer value, used if no \code{parts} are specified. The
#'        parts names will be build by pasting the \code{prefix} with a
#'        number ranging from 1 to \code{n}.
#' @param pattern a string used to split the id. It cannot be a regular
#'        pattern. It as to be a simple string.
#' @param category_mapping a category mapping object mapping actual category
#'        names to the levels of a \code{\link[ROBITools2]{robicategory}}
#'        instance (see \code{\link[ROBITools2]{robicategory_mapping}}.
#' @param include_full a logical value indicating if the full id has to
#'        be conserved as a new field in the \code{\link[ROBITools2]{robisample}}.
#' @param sample_categories a string indicating which part of the id contains
#'        the sample
#' @param replace if the initial \code{\link[ROBITools2]{robisample}} had columns
#'        with the same names, do the new ones replace the originals.
#' @return a \code{\link[ROBITools2]{robisample}} instance
#'
#' @rdname internal.merged_parsed_ids_and_samples
#' @examples
.merged_parsed_ids_and_samples <- function(i,object,parts,
                                  category_mapping,
                                  sample_categories,
                                  replace,
                                  prefix) {
  sids <- ids(object)
  sids_name <- ids_colname(object)

  colnames(i) <- parts

  if (sample_categories %in% parts && !missing(category_mapping))
    i[[sample_categories]] <- category_mapping(i[[sample_categories]])

  i[[sids_name]] = sids[[sids_name]]

  names(sids_name) <- sids_name

  i %>%
    select(sids_name,everything()) %>%
    left_join(object,by = sids_name,
              suffix = c("._new_","._old_")) -> i


  if (replace) {
    cnames <- colnames(i)
    old    <- cnames[str_detect(cnames,"._old_$")]
    new    <- cnames[str_detect(cnames,"._new_$")]
    names(new) <- str_replace(new,"._new_$","")

    i %>%
      select(-old) %>%
      rename(new) %>%
      robisample(sample_categories = sample_categories) -> i
  }

  i
}

#' Parse the column containing sample ids
#'
#' Often the sample ids are composed of sevral parts
#' giving some information about the sample. Is it a
#' true sample, a control, or a blanck. To which PCR
#' replicate it belongs.
#'
#' The parsing consists in spliting the id by a pattern
#' and to compose sub-ids by concatenating the firsts
#' (\code{left_parse_sample_ids}) or the last
#' (\code{right_parse_sample_ids}) of the intial id.
#' With an id looking like \code{P1_P2_P3} and a pattern
#' sets to underscore,  \code{left_parse_sample_ids} will generate
#' patterns \code{P1}, \code{P1_P2}, and \code{P1_P2_P3},
#' when \code{right_parse_sample_ids} will produce
#' \code{P3}, \code{P2_P3} and \code{P1_P2_P3}.
#'
#' @param object the object to parse a \code{\link[ROBITools2]{robisample}},
#'        or a \code{\link[ROBITools2]{robimetabar}} instance.
#' @param parts a string vector indicating the name of each sub-pattern
#'        produced. It can include or not a name for the last sub-pattern,
#'        that actually corresponds to the full id. if the \code{include_full}
#'        parameter is set to \code{FALSE} don't provide name for the full id.
#' @param n an integer value, used if no \code{parts} are specified. The
#'        parts names will be build by pasting the \code{prefix} with a
#'        number ranging from 1 to \code{n}.
#' @param pattern a string used to split the id. It cannot be a regular
#'        pattern. It as to be a simple string.
#' @param category_mapping a category mapping object mapping actual category
#'        names to the levels of a \code{\link[ROBITools2]{robicategory}}
#'        instance (see \code{\link[ROBITools2]{robicategory_mapping}}.
#' @param include_full a logical value indicating if the full id has to
#'        be conserved as a new field in the \code{\link[ROBITools2]{robisample}}.
#' @param sample_categories a string indicating which part of the id contains
#'        the sample
#' @param replace if the initial \code{\link[ROBITools2]{robisample}} had columns
#'        with the same names, do the new ones replace the originals.
#' @param prefix a string used as prefix when the \code{parts} parameter is missing.
#'
#'
#' @return a \code{\link[ROBITools2]{robisample}} instance
#' @export
#'
#' @examples
#'
#'     i <- c("blank_01_R1",   "blank_01_R2",
#'            "coExt_01_R1",   "coExt_01_R2",
#'            "PCRNeg_02_R1",  "PCRNeg_02_R2",
#'            "s_0001_R1",     "s_0002_R1",
#'            "s_0001_R2",     "s_0002_R2"  )
#'
#'    s <- robisample(i)
#'    s %>% left_parse_sample_ids(parts = c("type","samples"),
#'                                include_full = FALSE)
#'
#'    s %>% left_parse_sample_ids(parts = c("category","samples"),
#'                                category_mapping = robicategory_mapping(),
#'                                sample_categories = "category",
#'                                include_full = FALSE)
#'
#'    s %>% right_parse_sample_ids(parts = c("type","samples"),
#'                                include_full = FALSE)
#'
left_parse_sample_ids <- function(object,parts,n,pattern="_",
                                  category_mapping,
                                  sample_categories = "category",
                                  include_full = FALSE,
                                  replace = TRUE,
                                  prefix="Left") {
  UseMethod("left_parse_sample_ids",object)
}

#' @rdname left_parse_sample_ids
#' @export
left_parse_sample_ids.robidata <- function(object,parts,n,pattern="_",
                                             category_mapping,
                                             sample_categories = "category",
                                             include_full = FALSE,
                                             replace = TRUE,
                                             prefix="Left") {
  if (missing(parts) && missing(n))
    stop("at least on of the parts or n parameters hace to be specified")

  if (!missing(parts))
    n = length(parts)
  else
    parts = paste0(prefix, 1:n)

  if (include_full)
    ns = n
  else
    ns = n + 1

  sids <- ids(object)
  sids_name <- ids_colname(object)

  pattern <- fixed(pattern)

  suppressMessages(sids %>%
                     pull() %>%
                     as.character() %>%
    str_split_fixed(pattern = pattern,n = ns) %>%
    as_tibble(.name_repair = "unique") %>%
    apply(MARGIN = 1,
          function(p) sapply(1:n,
                             function(x) paste(p[1:x],
                                                collapse = pattern))) %>%
    t() %>% as_tibble(.name_repair = "unique")) %>%
    .merged_parsed_ids_and_samples(object,parts,
                                   category_mapping,
                                   sample_categories,
                                   replace,
                                   prefix) -> result

  if (any(vapply(result,is_robicategory,TRUE)))
    result %>% robisample()
  else
    result
}

#' @export
#'
#' @examples
right_parse_sample_ids <- function(object,parts,n,pattern="_",
                                  category_mapping,
                                  sample_categories = "category",
                                  include_full = FALSE,
                                  replace = TRUE,
                                  prefix="Right") {
  UseMethod("right_parse_sample_ids",object)
}

#' @rdname left_parse_sample_ids
#' @export
right_parse_sample_ids.robisample <- function(object,parts,n,pattern="_",
                                             category_mapping,
                                             sample_categories = "category",
                                             include_full = FALSE,
                                             replace = TRUE,
                                             prefix="Right") {
  if (missing(parts) && missing(n))
    stop("at least on of the parts or n parameters hace to be specified")

  if (!missing(parts))
    n = length(parts)
  else
    parts = paste0(prefix, 1:n)

  if (include_full) {
    ns = n
    nb = 1
  } else {
    ns = n + 1
    nb = 2
  }
  sids <- ids(object)
  sids_name <- ids_colname(object)

  pattern <- fixed(pattern)

  suppressMessages(sids %>%
                     pull() %>%
                     as.character() %>%
    str_split_fixed(pattern = pattern,n = ns) %>%
    as_tibble(.name_repair = "unique") %>%
    apply(MARGIN = 1,
          function(p) sapply(nb:ns,
                             function(x) paste(p[x:ns],
                                                collapse = pattern))) %>%
    t() %>% as_tibble(.name_repair = "unique"))  %>%
    .merged_parsed_ids_and_samples(object,parts,
                                   category_mapping,
                                   sample_categories,
                                   replace,
                                   prefix)
}


#' @rdname left_parse_sample_ids
#' @export
left_parse_sample_ids.robimetabar <- function(object,parts,n,pattern="_",
                                             category_mapping,
                                             sample_categories = "category",
                                             include_full = FALSE,
                                             replace = TRUE,
                                             prefix="Left") {
  call <- match.call()
  call[[1]] <- quote(left_parse_sample_ids.robidata)
  call[[2]] <- quote(object$samples)
  eval(call) -> i
  i
}


#' @rdname left_parse_sample_ids
#' @export
right_parse_sample_ids.robimetabar <- function(object,parts,n,pattern="_",
                                             category_mapping,
                                             sample_categories = "category",
                                             include_full = FALSE,
                                             replace = TRUE,
                                             prefix="Right") {
  call <- match.call()
  call[[1]] <- quote(right_parse_sample_ids.robidata)
  call[[2]] <- quote(object$samples)
  eval(call) -> i
  i
}