
every_kmer <- function(size) {
  if (size == 1) {
    return(c("A", "C", "G", "T"))
  }

  k1 <- every_kmer(size - 1)
  k2 <- character(length(k1) * 4)
  for (i in seq_along(k1)) {
    for (j in 1:4) {
      k2[(i - 1) * 4 + j] <- paste0(k1[i], c("A", "C", "G", "T")[j])
    }
  }

  k2
}

#' Generates random barcodes
#'
#' @param nseq number of barcodes to generate.
#' @param kmer size of the kmer you don't want to repeat over every generated sequences.
#' @param sizes a numeric vector with the lengths of the generated sequences, if shorter than \code{nseq}, it is repeated as needed.
#' @param max_homo the maximum length of an homopolymers.
#' @param max_trial maximum number of trial to build a sequenceof the requested length
#' @param verbose if \code{TRUE} some messages about failures during computation are emited
#'
#' @return a vector of character
#'
#' @examples
#'
#'   rbarcode(nseq = 32, kmer = 6, sizes = 100, max_homo = 3)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
rbarcode <- function(nseq = 16, kmer = 6, sizes = 100, max_homo = 3, max_trial = 1000, verbose = FALSE) {
  lseq <- 0
  k1 <- every_kmer(kmer - 1)

  if (max_homo > 0)
    k1 <- k1[-grep(pattern = sprintf("(.)\\1{%d}",max_homo),k1,perl = TRUE)]

  nrep = ceiling(nseq / length(sizes))
  sizes = rep(sizes,nrep)[1:nseq]

  extension <- rep(list(c("A", "C", "G", "T")), length(k1))
  names(extension) <- k1

  responses <- character(length = length(nseq))


  if (max_homo > 0) {
    homo_pat = k1[grep(sprintf("(.)\\1{%d}$",max_homo - 1),k1,perl = TRUE)]
    for (k in homo_pat) {
      e <- c("A", "C", "G", "T")[which(c("A", "C", "G", "T") != substr(k,start = kmer - 1, stop = kmer - 1))]
      extension[[k]] <- e
      if (verbose) message(sprintf("reduce extensions of seed %s -> %s\n",substr(k,start = kmer - 1, stop = kmer - 1),k,deparse(e)))
    }
  }

  ex <- extension

  for (j in seq_along(sizes)) {
    size <- sizes[j]
    trial <- 0
    lseq  <- 0

    while (lseq < size && trial <= max_trial && length(extension)) {
      if (verbose && trial > 0) message(sprintf("Retry %d for sequence %d (%d/%d)",trial - 1,j,lseq,size))
      trial <- trial + 1
      ex <- extension

      seq  <- character(size - kmer + 2)
      seed <- sample(names(ex), 1)
      seq[1] <- seed

      for (i in 2:length(seq)) {
        possible <- ex[[seed]]

        if (!length(possible)) break

        next_nuc <- sample(1:length(possible), 1)
        nuc <- possible[next_nuc]
        possible <- possible[-next_nuc]

        ex[[seed]] <- if (length(possible))
                        possible
                      else
                        NULL

        seed <- paste0(substr(seed, start = 2, stop = kmer - 1),
                       nuc
                      )

        seq[i] <- nuc

      }

      seq = paste(seq,collapse = "")
      lseq = nchar(seq)

    }

    extension <- ex
    responses[j] <- seq
  }

  responses[nchar(responses) == sizes]
}
