
# EcoPCR tabular file format :
#
#   column 1 : accession number
#   column 2 : sequence length
#   column 3 : taxonomic id
#   column 4 : rank
#   column 5 : species taxonomic id
#   column 6 : scientific name
#   column 7 : genus taxonomic id
#   column 8 : genus name
#   column 9 : family taxonomic id
#   column 10 : family name
#   column 11 : super kingdom taxonomic id
#   column 12 : super kingdom name
#   column 13 : strand (direct or reverse)
#   column 14 : first oligonucleotide
#   column 15 : number of errors for the first strand
#   column 16 : Tm for hybridization of primer 1 at this site
#   column 17 : second oligonucleotide
#   column 18 : number of errors for the second strand
#   column 19 : Tm for hybridization of primer 1 at this site
#   column 20 : amplification length
#   column 21 : sequence
#   column 22 : definition



#' Reads an ecoPCR file
#'
#' @param file
#' @param verbose
#'
#' @return
#' @export
#'
#' @examples
#'    filename <- system.file("extdata", "chlo02.ecopcr",
#'                            package = "ROBITools2")
#'
#'    read_ecopcr(filename)
#'
read_ecopcr <- function(file,
                        verbose = is_robi_verbose()) {

  split.line = function(line) {
    l = strsplit(line,split=" +\\| +")[[1]]
    l = c(l[1:21],paste(l[-c(1:21)],sep="|"))
    return(l)
  }

  if (is.character(file)) {
    file <- file(file, "rb")
    on.exit(close(file))
  }
  if (!inherits(file, "connection"))
    stop("'file' must be a character string or connection")
  if (!isOpen(file, "rb")) {
    open(file, "rt")
    on.exit(close(file))
  }

  data <- read_lines(file)
  #close(file)

  if (data[1] == "#@ecopcr-v2")
    version <- 2
  else
    version <- 1

  mode <- str_split(data[str_which(data,pattern = "^# output in")[1]],
                    pattern = " ")[[1]][4]
  robimessage("EcoPCR run in {mode} mode")

  data <- data[-str_which(data,"^#")]
  file  <- rawConnection(charToRaw(paste0(data,collapse = "\n")))

  if (version == 1)
    col_types="ciiciciciciccciciicc"
  else
    col_types="ciicicicicicccifcificc"


  data <- read_delim(file,delim = "|",
                     trim_ws = TRUE,
                     col_names = FALSE,
                     col_types=col_types,
                     na = "###")
  #close(file)

  if (version == 1) {
    robimessage("EcoPCR file format V1")
    cnames <- c("ac","seq_length","TAXID","rank",
                "species","species_name",
                "genus","genus_name",
                "family","family_name",
                mode,paste0(mode,"_name"),
                "strand",
                "forward_match","forward_mismatch",
                "reverse_match","reverse_mismatch",
                "amplicon_length","sequence","definition")
    } else {
    robimessage("EcoPCR file format V2")
    cnames <- c("ac","seq_length","TAXID","rank",
                "species","species_name",
                "genus","genus_name",
                "family","family_name",
                mode,paste0(mode,"_name"),
                "strand",
                "forward_match","forward_mismatch","forward_tm",
                "reverse_match","reverse_mismatch","reverse_tm",
                "amplicon_length","sequence","definition")
    }

  colnames(data) <- cnames

  features <- data %>%
    select(-sequence,-definition) %>%
    mutate(TAXID = as_robitaxid_master(TAXID),
           species = ifelse(species==-1,NA,species),
           species = as_robitaxid(species),
           genus = ifelse(genus==-1,NA,genus),
           genus = as_robitaxid(genus),
           family = ifelse(family==-1,NA,family),
           family = as_robitaxid(family),
           "{as.symbol(mode)}" :=  ifelse(.data[[mode]]==-1,NA,.data[[mode]]),
           "{as.symbol(mode)}" :=  as_robitaxid(.data[[mode]])
           )

  tibble (id=new_robiuniqueid(data$ac,make_unique = TRUE,verbose = FALSE),
          features = features,
          sequence = data$sequence,
          definition=data$definition) %>% as_robidata()
}