#' @include read_obifasta.R
#' @include robimetabar.R
NULL

#' Title
#'
#' @param file
#' @param keys
#' @param motu_ids
#' @param read_count
#' @param obiclean_prefix
#' @param obiclean_status
#' @param sumaclust_prefix
#' @param main_taxid
#' @param other_taxid
#' @param verbose
#'
#' @return
#' @export
#'
#' @examples
read_metabar <- function(file,
                         keys = NULL,
                         ngsfilter = NULL,
                         motu_ids = "id",
                         read_count = "sample",
                         obiclean_prefix = "obiclean",
                         obiclean_status = paste0(obiclean_prefix,"_status"),
                         obiclean_cluster = paste0(obiclean_prefix,"_cluster"),
                         obiclean_count = paste0(obiclean_prefix,"_count"),
                         sumaclust_prefix = "cluster",
                         main_taxid = "taxid",
                         other_taxid = c("family","genus","order","species","taxid_by_db"),
                         verbose = is_robi_verbose()) {

  mfasta <- read_obifasta(file = file,keys = keys, verbose = verbose)

  features <- mfasta$features

  merged_read_count <- sprintf("merged_%s",read_count)

  features$sequence <- mfasta$sequence
  features$id <- as_robiuniqueid(mfasta$id)

  data <- feature_as_three_columns(features[[merged_read_count]],
                                   motus_ids = features$id) %>%
          rename(count = value)

  features[[merged_read_count]] <- NULL

  if (obiclean_status %in% colnames(features)) {
    data <- data %>%
              full_join(features[[obiclean_status]] %>%
                          feature_as_three_columns(motus_ids = features$id),
                        by = c("motu", "sample"))

    cd <- colnames(data)
    cd[cd == "value"] <-  obiclean_status
    colnames(data) <- cd
    data[[obiclean_status]] = factor(data[[obiclean_status]],levels = c("h","s","i"))

    features[[obiclean_status]] <- NULL
  }

  if (obiclean_cluster %in% colnames(features)) {
    data <- data %>%
              full_join(features[[obiclean_cluster]] %>%
                          feature_as_three_columns(motus_ids = features$id),
                        by = c("motu", "sample"))

    cd <- colnames(data)
    cd[cd == "value"] <-  obiclean_cluster
    colnames(data) <- cd

    features[[obiclean_cluster]] <- NULL
  }

  if (obiclean_count %in% colnames(features)) {
    data <- data %>%
              full_join(features[[obiclean_count]] %>%
                          feature_as_three_columns(motus_ids = features$id),
                        by = c("motu", "sample"))

    cd <- colnames(data)
    cd[cd == "value"] <-  obiclean_count
    colnames(data) <- cd

    features[[obiclean_count]] <- NULL
  }

  features %>% lapply(function(x) {
    if (is(x,"list")) feature_as_matrix(x) %>% as_tibble()
    else x}) %>% as_tibble() -> features

  obiclean_idx <- which(str_detect(colnames(features),
                                   pattern = paste0('^', obiclean_prefix)))

  if (length(obiclean_idx) > 0) {
    obiclean <- features[,obiclean_idx]
    features <- features[,-obiclean_idx]
    features$obiclean <- obiclean
  }

  sumaclust_idx <- which(str_detect(colnames(features),
                                    pattern = paste0('^', sumaclust_prefix)))

  if (length(sumaclust_idx) > 0) {
    sumaclust <- features[,sumaclust_idx]
    features <- features[,-sumaclust_idx]
    features$sumaclust <- sumaclust
  }

  if (main_taxid %in% colnames(features)) {
    features[[main_taxid]] <- as_robitaxid_master(features[[main_taxid]])
  }

  for (c in other_taxid)
    if (c %in% colnames(features)) {
      fs <- features[[c]]
      if (is_tibble(fs))
        for (j in seq_along(fs))
          fs[[j]] <- as_robitaxid(fs[[j]])
      else
        fs <- as_robitaxid(fs)
      features[[c]] <- fs
    }

  usampleids <- unique(data$sample)

  if (!is.null(ngsfilter)) {
    m <- match(usampleids,ids(ngsfilter) %>% pull(),NA)
    missing <- usampleids[which(is.na(m))]
    robiassert_arg(length(missing) == 0,
                   "ngsfilter",
                   "The provided ngsfilter doesn't define all the samples. ({paste(missing,collapse=', ')}) are missing ")
    samples <- robisample(ngsfilter)
  } else {
    samples <- robisample(usampleids)
  }

  robimetabar(motus   = robimotu(features %>% select(id,everything())),
              samples = samples,
              data    = data,
              verbose = verbose)
}
