#' @include utils.R
#' @include robiseq.R
#' @import stringr
#' @import readr
#' @import foreach
#' @import dplyr
#' @import progressr
#'
#' @author Eric Coissac
NULL



features_as_tibble <- function(features) {
  all_names <- sort(unique(unlist(lapply(features, names))))

  data <- lapply(all_names, function(n) vapply(features,
                                               function(y)
                                                 if (is.null(v <- y[n])) NA else v,
                                               FUN.VALUE = "") %>%
                                        .coerce_vector()
                   )
  names(data) <- all_names

  return(as_tibble(data))

  data <- do.call(rbind,lapply(features,
                               function(x)  {
                                 x <- x[all_names]
                                 x[vapply(x, is.null, TRUE)] <- NA
                                 names(x) <- all_names
                                 x <- lapply(x, function(y) if (length(y)<=1) y else list(y))
                                 as_tibble(x)
                               }
                               ))

  data
}

#' Title
#'
#' @param feature
#' @param key
#'
#' @return
#' @export
#'
#' @examples
feature_as_matrix <- function(feature,na_value=NA) {
  feature <- lapply(feature,
                    function(x) {
                      y <-as.character(x)
                      names(y)<-names(x)
                      y}
                    )

  all_names <- sort(unique(unlist(lapply(feature, names))))
  data <- do.call(rbind,lapply(feature, function(x) x[all_names]))

  if (!is.na(na_value))
    data[is.na(data)] <- na_value

  colnames(data) <- all_names

  as_tibble(data) %>% lapply(.coerce_vector) %>% as_tibble()
}

feature_as_three_columns <- function(feature,motus_ids=NULL) {
  samples <- unlist(lapply(feature, names))
  all_names <- sort(unique(samples))
  if (is.null(motus_ids))
    motus_ids <- build_motu_names(n = seq_along(feature))

  lengths <- vapply(feature, length, 0)
  motus <-   rep(motus_ids,lengths)
  values  <- unlist(feature)

  tibble(motu   = motus,
         sample = samples,
         value  = values)
}


.coerce_vector <- function(data) {

  data[data == "None"] <- NA

  converted <- suppressWarnings(as.logical(data))
  if (all(!is.na(converted) | is.na(data)))
    return(converted)

  converted <- suppressWarnings(as.integer(data))
  na_conv <- is.na(converted)
  if (all(!na_conv | is.na(data))) {
    num <- suppressWarnings(as.numeric(data))
    if (all(converted == num | na_conv))
      return(converted)
    else
      return(num)
  }
  converted <- .Call("R_as_named_list",data,TRUE)
  if (all(suppressWarnings(vapply(converted,length,0)) > 0 | is.na(data))) {
    converted <- lapply(converted, function(x) {
      if (length(x) > 0) {
        value <- .coerce_vector(x[[2]])
        names(value) <- x[[1]]
        value
      }
      else x
    })
    return(converted)
  }

  .Call("R_trim_and_unquote_strings",data)
}



#' Builds a regex pattern matching key=value; patterns
#'
#' The pattern is usable to parse the key=value; patterns
#' included in the obifasta sequence header lines.
#'
#' @param keys the \code{NULL} value or a character vector indicating the list
#'        of tags to extract from the fasta header.
#'
#' @return a regular pattern from the `ore` library
#'
#' @md
#' @examples
#'   ROBITools2:::.build_key_value_pattern(c("taxid","species"))
.build_key_value_pattern <- function(keys=NULL) {

  .or_rx  <- function(...) paste0('(',paste(list(...),collapse = "|"),')')
  .and_rx <- function(...) paste0(list(...),collapse = "")
  .sub_rx <- function(x) paste0("(",x,")")
  .eventually_rx <- function(x) paste0(.sub_rx(x),"?")
  .eventually_several_rx <- function(x) paste0(.sub_rx(x),"*")
  .several_rx <- function(x) paste0(.sub_rx(x),"+")

  .sq_string = "'[^']*'"
  .dq_string = '"[^"]*"'
  .string = .or_rx(.sq_string,.dq_string)

  .space = "[ \t]*"

  .integer <- "[+-]?[0-9]+"
  .exponent <- .and_rx("[eE]",.integer)
  .numeric1 <- "[+-]?\\.[0-9]+"
  .numeric2 <- .and_rx(.integer,"(\\.[0-9]*)?")
  .numeric  <- .and_rx(.or_rx(.numeric1,.numeric2),.eventually_rx(.exponent))

  .value <- .or_rx(.string,.numeric)

  .key_val <- .and_rx(.string,.space,":",.space,.value)
  .dict <- .and_rx("\\{",.eventually_rx(.and_rx(.key_val,.eventually_several_rx(.and_rx(.space,",",.space,.key_val)))),"\\}")

  .left <- if (is.null(keys))
                  "[^0-9'\";= \t][^'\";= \t]*"
           else
             do.call(.or_rx,as.list(keys))
  .right <- .or_rx(.numeric,.string,.dict,"[^;]+")

  ore::ore(.and_rx(.sub_rx(.left),.space,"=",.space,.sub_rx(.right)))
}


parse_sequence_header <- function(title,keys_pattern) {

  tags <- ore::groups(ore::ore.search(keys_pattern,title,
                                      all = TRUE,
                                      simplify = TRUE))[,1:2,drop=FALSE]

  parsed_values <- tags[,2]
  names(parsed_values) <- tags[,1]

  parsed_values

}

#' Reads OBITools extended version of fasta files.
#'
#' @param file a character string containing the filename to read
#'
#' @param keys the `NULL` value or a character vector indicating the list
#'        of tags to extract from the fasta header.
#' @param verbose
#'
#' @md
#' @export
#' @examples
#'    filename <- system.file("extdata", "B5.fasta.gz",
#'                            package = "ROBITools2")
#'
#'    read_obifasta(filename)
read_obifasta <- function(file,
                          keys = NULL,
                          verbose = is_robi_verbose()) {
  keys_pattern <- .build_key_value_pattern(keys)

  robimessage("Loading the {file} file in memory...")

  mfasta <- read_file_raw(file)

  robimessage("Looking for the beginning of the sequences...")

  mfasta <- .Call("R_parse_fasta", mfasta)
  names(mfasta) <- c("id", "features", "sequence")

  robimessage("{length(mfasta[[1]])} sequences found")



  features <- as.list(mfasta$features)
  pbar <- progressor(along = features)

  for (seq_idx in seq_along(features)) {
    if (! seq_idx %% 50) pbar(amount = 50)
    features[[seq_idx]] <- parse_sequence_header(features[[seq_idx]],
      keys_pattern = keys_pattern
    )
  }

  mfasta$features <- features_as_tibble(features)

  as_tibble(mfasta)
}
