#' @include type_robiatomic.R
#' @importFrom  vctrs vec_data vec_cast vec_default_ptype2 vec_default_cast new_factor
#'
NULL

.robi_all_categories <- c("sample", "positive_ctrl", "extraction_blk", "pcr_blk", "sequencing_blk")


#' Robicateroy vector to classify PCRs in a DNA metbarcoding experiment.
#'
#' In the [ROBITools2] library every PCR referenced in a [`robisample`]
#' have to be associated to a category. The set of categories
#' is predefined and cannot be changed by the user. If users need another more
#' precise classification system of their PCR they can us a standard [`base::factor`]
#' with their own defined levels.
#'
#' The `robicategory` system define five categories:
#'
#' * **sample**         : which corresponds to PCR run on DNA corresponding to a true sample.
#' * **positive_ctrl**  : for PCRs corresponding to the amplification of a
#'                        positive control (mock community).
#' * **extraction_blk** : for PCRs corresponding to an extraction control.
#' * **pcr_blk**        : for PCRs corresponding to a PCR control (no DNA template).
#' * **sequencing_blk** : for combinaisons of tags corresponing to no PCR
#'
#' A `robicategory` is implemented as a factor. `NA` values are not allowed in `robicategory`.
#'
#' @details The `robicategory()` constructor builds a default `robicategory` instance of length
#'  defined by the `length` argument and composed of elements corresponding to the `sample`
#'  modality.
#'
#' @param length A non-negative integer specifying the desired length. Double
#'   values will be coerced to integer: supplying an argument of length other
#'   than one is an error.
#'
#' @return a \code{robicategory} instance
#'
#' @examples
#'
#' rc <- robicategory(10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
#' @md
robicategory <- function(length = 0) {

  robiassert_arg(length(length) == 1,
                 "length",
                 "supplying an argument of length ({length(length)}) other than one is an error")

  new_robicategory(rep("sample",length))
}

#' @details The contructor `new_robicategory` is the main function to
#'   create a `robicategory` instance. For simpler conversion to that class,
#'   it can be easier to use the `as_robicategory()` utility function.
#'
#' @param object a [`base::character`] vector or an object coercible to such
#'   vector, to convert in to a `robicategory` instance. The object once
#'   transform into a character vector, must only contain string that extactly
#'   correspond to one of the imposed level. See [`robicategory_mapping()`] for
#'   more complex conversions.
#' @param ... Passed on to [`base::structure()`]. It allows for
#'   defining attributes, specified in tag = value form, which will be attached
#'   to data (see [`robiatomic`]).
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   constructor takes some decision because of missing values and estimates
#'   default values these missing data.
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#'
#' rc <- new_robicategory(rep(c("sample", "positive_ctrl", "pcr_blk"),
#'                            c(3, 2, 1)))
#'
#' @rdname robicategory
#' @export
#' @md
new_robicategory <- function(object, ...,
                             verbose = is_robi_verbose(),
                             class = character()) {

  object <- vec_cast(object,character())

  fac <- match(object, .robi_all_categories)

  if (verbose && any(is.na(fac))) {
    warning(ngettext(
      sum(is.na(fac)),
      sprintf(
        "The element %s doesn't belong one of the following categories",
        (object[is.na(fac)])[1],
        substr(paste(deparse(.robi_all_categories), collapse = ""), 2, 1000)
      ),
      sprintf(
        "The elements %s don't belong one of the following categories",
        substr(paste(deparse(unique(object[is.na(fac)])), collapse = ""), 2, 1000),
        substr(paste(deparse(.robi_all_categories), collapse = ""), 2, 1000)
      ),
    ))
  }

  fac <- new_factor(fac,levels = .robi_all_categories,
                    class = c(class, "robicategory",
                              "robiatomic","robiobject"))

  validate_robiobject(fac)
}

#' @details The [`robicategory`] version of that method ckecks that :
#'
#'  * the instance does not contain any `NA` value.
#'  * the levels of the instance corresponds to the predefined levels
#'
#' @examples
#'
#' rc <- robicategory(10)
#' validate_robiobject(xd)
#'
#' @export
#' @rdname validate_robiobject
#' @md
validate_robiobject.robicategory <- function(object) {
  stopifnot(identical(levels(object),.robi_all_categories))

  if (any(is.na(object))) {
    stop('<NA> values not allowed in robicategory instance')
  }
  object
}

#' @details `is_robicategory` tests if an object belongs the `robicategory` class
#'
#' @param `object` the object to test
#' @return a [`logical`] value
#'
#' @examples
#'
#' is_robicategory(rc)
#' is_robiatomic(1:10)
#'
#' @rdname robiatomic
#' @export
#' @md
is_robicategory <- function(object) {
  inherits(object, "robicategory")
}

#' @details `as_robiatomic` converts an object to a `robicategory` instance.
#'
#' @param object the object to convert
#' @return a `robicategory` instance
#'
#' @examples
#'
#' x <- factor(rep(c("sample", "positive_ctrl", "pcr_blk"),
#'             c(3, 2, 1)))
#' as_robicategory(x)
#'
#' @rdname robicategory
#' @export
#' @md
as_robicategory <- function(object) {
  vec_cast(object,robicategory())
}


#' Mapping of a character vector to a `robicategory` instance.
#'
#' The fonction `robicategory_mapping()` builds a mapping fonction
#' to translate a [`base::character`] vector into a [`robicategory`]
#' instance.
#'
#' @param sample the string used to qualifying samples.
#' @param positive_ctrl the string used to qualifying positive controles
#' @param extraction_blk the string used to qualifying extraction blank.
#' @param pcr_blk the string used to qualifying PCR blank.
#' @param sequencing_blk the string used to sequencing blank.
#' @param default the default category when the name is matching no pattern
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   constructor takes some decision because of missing values and estimates
#'   default values these missing data.
#'
#' @return a [`robicategory`] instance.
#'
#' @examples
#'
#' s <- c("blk","s","s","s","POS","Ext","NEG")
#'
#' mapper <- robicategory_mapping(sample="s",
#'                                positive_ctrl="POS",
#'                                extraction_blk="blk",
#'                                pcr_blk="NEG",
#'                                sequencing_blk="blank")
#'
#' rc <- mapper(s)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
#' @md
robicategory_mapping <- function(sample="s",positive_ctrl="PCRPos",
                                 extraction_blk="coExt",
                                 pcr_blk="PCRNeg",
                                 sequencing_blk="blank",
                                 default = NA,
                                 verbose = is_robi_verbose()) {

  matcher <- function(data) {
      m <- rbind(str_detect(data,sample),
                 str_detect(data,positive_ctrl),
                 str_detect(data,extraction_blk),
                 str_detect(data,pcr_blk),
                 str_detect(data,sequencing_blk))

      ms = colSums(m)
      robiassert(all(ms <= 1),message = "A sample name match several categories")

      cat <- colSums(m * 1:5)
      cat
  }

  m <- as.character(c(sample, positive_ctrl, extraction_blk, pcr_blk, sequencing_blk))

  mapping <- function(data) (
    new_robicategory(c(default,.robi_all_categories)[matcher(as.character(data))+1],
                 verbose = verbose)
  )

  mapping
}

#' @details This implementation of the \code{\link[base]{levels}} method for
#' the \code{robicategory} just allows for substituting levels of a
#' `robicategory` by themselves.
#'
#' @param x a \code{robicategory} instance
#' @param value A valid value for levels, actually not used.
#'
#' @examples
#'
#' x <- robicategory(10)
#' levels(x) <- c("sample", "positive_ctrl",
#'                "extraction_blk", "pcr_blk",
#'                "sequencing_blk")
#'
#' \dontrun{
#' levels(x) <- c("A", "B", "C", "D")
#' }
#'
#' @rdname robicategory
#' @export
#' @md
`levels <- .robicategory` <- function(x, value) {
  robiassert_arg(identical(value,.robi_all_categories),
                 "value",
                 "levels of a robicategory cannot be change")
  levels(x) <- value
}



##
#
# robicategory type mapping for the vctrs cast system
#
##

#' Vctrs cast methods for robicategory
#' @examples
#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory <- function(x, y, ...) {
  UseMethod("vec_ptype2.robicategory")
}

#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory.default <- function(x, y, ..., x_arg = "x", y_arg = "y") {
  vec_default_ptype2(x, y, x_arg = x_arg, y_arg = y_arg)
}

#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory.robicategory <- function(x, y, ...) robicategory()

#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory.integer <- function(x, y, ...) integer()

#' @rdname robicategory_cast
#' @export
vec_ptype2.integer.robicategory <- function(x, y, ...) integer()

#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory.double <- function(x, y, ...) double()

#' @rdname robicategory_cast
#' @export
vec_ptype2.double.robicategory <- function(x, y, ...) double()

#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory.character <- function(x, y, ...) character()

#' @rdname robicategory_cast
#' @export
vec_ptype2.character.robicategory <- function(x, y, ...) character()

#' @rdname robicategory_cast
#' @export
vec_ptype2.robicategory.factor <- function(x, y, ...) factor()

#' @rdname robicategory_cast
#' @export
vec_ptype2.factor.robicategory <- function(x, y, ...) factor()


#' @rdname robicategory_cast
#' @export
vec_cast.robicategory <- function(x, to, ...)
  UseMethod("vec_cast.robicategory")

#' @rdname robicategory_cast
#' @export
vec_cast.robicategory.default <- function(x, to, ...)
  vec_default_cast(x, to)

#' @rdname robicategory_cast
#' @export
vec_cast.robicategory.robicategory <- function(x, to, ...) x

#' @rdname robicategory_cast
#' @export
vec_cast.integer.robicategory <- function(x, to, ...)
  vec_data(x)

#' @rdname robicategory_cast
#' @export
vec_cast.double.robicategory <- function(x, to, ...)
  vec_cast(vec_data(x),double())

#' @rdname robicategory_cast
#' @export
vec_cast.factor.robicategory <- function(x, to, ...)
  new_factor(vec_data(x),levels = levels(x))

#' @rdname robicategory_cast
#' @export
vec_cast.robicategory.factor <- function(x, to, ...)
  new_robicategory(vec_cast(x,character()))

#' @rdname robicategory_cast
#' @export
vec_cast.robicategory.character <- function(x, to, ...)
  new_robicategory(x)

#' @rdname robicategory_cast
#' @export
vec_cast.character.robicategory <- function(x, to, ...)
  levels(x)[vec_data(x)]


#' @rdname robicategory_cast
#' @export
vec_ptype_abbr.robicategory <- function(x, ...) {
"robicategory"
}

#' @rdname robicategory_cast
#' @importFrom pillar type_sum
#' @export
type_sum.robicategory <- function(x) {
  "robicat"
}

#' @rdname robicategory_cast
#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.robicategory <- function(x, ...) {
  out <- format(x)
  out[is.na(x)] <- NA
  pillar::new_pillar_shaft_simple(out, align = "left")
}


#' Title
#'
#' @param x
#' @param incomparables
#' @param fromLast
#' @param nmax
#' @param ...
#'
#' @return
#' @export
#'
#' @examples
unique.robicategory <- function(x, incomparables = FALSE,
                                fromLast = FALSE, nmax = NA,
                                ...) {
  call = match.call()
  call[[1]] <- quote(unique.default)
  as_robicategory(eval(call))
}

