#' @include type_robiobject.R
#' @import tibble
#' @import dplyr
#' @import vctrs
NULL

#' Test if an object belongs \code{robidata} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#' x <- tibble(x=1:10,y=rnorm(10))
#' xd <- new_robidata(x,class="robifoo")
#' is_robidata(xd)
#' is_robidata(x)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robidata <- function(object) {
  inherits(object, "robidata")
}


#' @details The \code{\link[ROBITools2]{robidata}} version
#'     ckecks that the instance contains a one and only one
#'     \code{\link[ROBITools2]{robiuniqueid}} instance.
#'
#' @export
#'
#' @rdname validate_robiobject
validate_robiobject.robidata <- function(object, verbose=is_robi_verbose()) {
  cuid <- sum(vapply(object, is_robiuniqueid, FUN.VALUE = TRUE))

  robiassert_arg(cuid == 1,
                 "object",
    "the {class} instance must have one and only one column of type robiuniqueid, not {count}",
    class = class(object)[1],
    count = cuid)

  object
}


#' @param object the object to modify
#' @param ... Passed on to \code{\link[base]{structure}}()
#' @param class Subclasses to assign to the new object, default: none
#'
#' @rdname robidata
#' @export
new_robidata <- function(object,...,class=NULL) {

  robiassert_arg(
    is_tibble(object),
    "object",
    "Object of class {class} can only be built from a tibble instance",
    class = class)

  validate_robiobject(new_robiobject(object,
                                 ...,
                                 class = c(class, "robidata")))
}

#' Builds a new \code{robidata} instance.
#'
#' robidata is a subclass of \code{\link[tibble]{tibble}}.
#' It is notably the super class of \code{\link[ROBITools2]{robisample}}
#' and \code{\link[ROBITools2]{robimotu}}.
#' \code{robidata} must have an column belonging the
#' \code{\link[ROBITools2]{robiuniqueid}} class to insure a unique identifier.
#'
#'
#' @param object
#' @param id_column
#'
#' @return a  \code{robidata} instance
#' @export
#'
#' @examples
robidata <- function(object,id_column) {

  object <- as_tibble(object)

  if(!missing(id_column)) {
    robiassert_arg(is_character(id_column) && length(id_column)==1,
                   "id_column",
                   "Must be a single character string")

    object[[id_column]] <- robiuniqueid(object[[id_column]])
  }

  new_robidata(object)
}

#' @rdname robidata
#' @export
as_robidata <- function(object) {
  UseMethod("as_robidata", object)
}

#' @rdname robidata
#' @export
as_robidata.default <- function(object) {
  new_robidata(object)
}

#' @rdname robidata
#' @export
as_robidata.robidata <- function(object) {
  classes <- class(object)
  rclass <- which(classes == 'robidata')
  class(object) <- classes[rclass:length(classes)]
  object
}



#' Ids of a \code{robidata}
#'
#' @details Extract the \code{\link[ROBITools2]{robiuniqueid}} column
#' from a \code{\link[ROBITools2]{robidata}} instance.
#'
#' @param object the object to analyse
#'
#' @return a \code{\link[tidyr]{tibble}} instance with a single
#'        \code{\link[ROBITools2]{robiuniqueid}} column.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
ids <- function(object) {
  UseMethod("ids", object)
}


#' @details Extract the name of the \code{\link[ROBITools2]{robiuniqueid}}
#' column from a \code{\link[ROBITools2]{robidata}} instance.
#'
#' @return a \code{character} string.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @rdname ids
#' @export
ids_colname <- function(object) {
  UseMethod("ids_colname", object)
}


#' @rdname ids
#' @export
ids.default <- function(object) {
  stop(sprintf("No ids method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname ids
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   ids_colname(s)
#'
ids_colname.robidata <- function(object) {
  colnames(object)[which(vapply(object, is_robiuniqueid, FUN.VALUE = TRUE))]
}

#' @rdname ids
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   ids(s)
#'
ids.robidata <- function(object) {
  select(object,
         ids_colname(object))
}


#' @export
rbind.robidata <- function(..., deparse.level = 1) {

 result <- base::rbind.data.frame(..., deparse.level = deparse.level,
                                  stringsAsFactors = FALSE)

 if (is_robidata(result))
   result %>%
     mutate_if(is_robiatomic,validate_robiobject) %>%
     validate_robiobject()
 else
   result
}


#' @export
rbind.robidata <- function(..., deparse.level = 1) {
  vec_rbind(...)
}
#
#  fcall <- match.call()
#  fcall[[1]] <- quote(base::rbind.data.frame)
#
#  d <- list(...)
#  all_robidata <- all(sapply(d,is_robidata))
#  if (all_robidata) {
#    ids_name <- unique(sapply(d,ids_colname))
#    same_ids_name <- length(ids_name) == 1
#    classes <- lapply(d, class)
#  }
#
#  result <- eval(fcall)
#
#  if (all_robidata && same_ids_name) {
#    classes <- lapply(classes, rev)
#    mclass  <- sapply(classes,length)
#    common_class <- classes[[1]][(which(sapply(1:mclass, function(i)
#                       unique(sapply(classes,function(x) x[[i]])) == 1
#                    )) %>% rev())[1]]
#
#    fcall = parse(text = "robidata(result)")
#    fcall[[1]] = as.name(common_class)
#
#    result[[ids_name]] <- robiuniqueid(result[[ids_name]])
#    result <- eval(fcall)
#  }
#
#  result
# }
