#' @title robimetabar.R
#'
#' @import R6
#' @import dplyr
#' @import tidyr
#' @import tibble
#'
#' @include robisample.R
#' @include robimotu.R
#'
NULL

.robimetabar = R6::R6Class("robimetabar",

                         public = list(
      data = NULL,
      samples = NULL,
      motus = NULL,
      distances = NULL,


      initialize = function(data,
                            samples,
                            motus,
                            distances
      ) {

        robiassert_arg(!missing(data),"data",
                   "the data parametter is mandatory")


        robiassert_arg(is_tibble(data),
                       "data",
                       "Must be a tibble not a {wrongclass}.",
                       wrongclass = class(data)[1])

        self$data <- data


        robiassert_arg(!missing(samples),"samples",
                   "the samples parametter is mandatory")

        robiassert_arg(is_robisample(samples),
                       "samples",
                       "Must be a robisample not a {wrongclass}.",
                       wrongclass = class(samples)[1])

        self$samples <- samples

        robiassert_arg(!missing(motus),"motus",
                   "the motus parametter is mandatory")

        robiassert_arg(is_robimotu(motus),
                       "motus",
                       "Must be a robimotu not a {wrongclass}.",
                       wrongclass = class(motus)[1])

        self$motus <- motus

        if (!missing(distances))
          self$distances <- distances

        validate_robiobject(self)
      },


      dim = function()
        c(nrow(self$samples), nrow(self$motus)),

      dimnames = function()
                   list(samples = as.character(samples_ids(self)),
                        motus   = as.character(motus_ids(self)))
  )
)


#' @details The \code{robimetabar} version of this methode checks that
#' taxonomy declared as full verified that constaint.
#'
#' @rdname validate_robiobject
#' @export
validate_robiobject.robimetabar <- function(object, verbose=is_robi_verbose()) {

  object
}


#' Test if an object belongs \code{robicategory} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#' m <- matrix(1:20,nrow=4)
#' me <- robimetabar(m)
#' is_robimetabar(me)
#' is_robimetabar(m)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robimetabar <- function(object) {
  inherits(object, "robimetabar")
}


#' Title
#'
#' @param object
#' @param alternative_names
#' @param reference_taxonomy
#' @param ...
#' @param class
#'
#' @return
#' @export
#'
#' @examples
new_robimetabar <- function(data,
                            samples,
                            motus,
                            distances,
                             ...) {

  .robimetabar$new(data = data,
                   samples = samples,
                   motus = motus,
                   distances = distances)

}


#' Constructor for `robimetabar` instances.
#'
#' `robimetabar` builds a `robimetabar` instance from
#' simple R data structure.
#'
#' @param samples must be a `tibble` or a data structure coercible to that type.
#'                This table describes the samples. Each row correponds to a sample,
#'                each column to a description variable. The table must have the same
#'                number of row than the number of columns of the `reads` matrix.
#'                This is not a mandatory parameter. If missing a default `samples`
#'                table will be construct from the column names of the `reads` matrix.
#'
#' @param motus   must be a `tibble` or a data structure coercible to that type.
#'                This table describes the MOTUs Each row correponds to a MOTU,
#'                each column to a description variable. The table must have the same
#'                number of row than the `reads` matrix.
#'                This is not a mandatory parameter. If missing a default `motus`
#'                table will be construct from the row names of the `reads` matrix.
#'
#' @param distances a named list containing or \code{dist} object of the same size than
#'                the number of row of the `reads` matrix, or function accepting as single
#'                argument an instance of `robimetabar` and returning \code{dist} object
#'                of the same size than the number of row of the `reads` matrix
#'
#' @param verbose    if `TRUE` warnings are emitted if the constructor takes
#'                   some decision because of missing values and estimates
#'                   default values these missing data.
#' @param data
#'
#' @return a \code{robimetabar} instance.
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#'   metabar <- robimetabar(data = reads)
#'
#'   samples(metabar)
#'   motus(metabar)
#'   metabar$reads
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
robimetabar <- function(data, samples, motus,
                        distances = vector(mode = "list", 0),
                        verbose = is_robi_verbose()) {

  if (missing(data))
    stop("data argument must be provided",
         call. = FALSE)

  if (!missing(samples) && !is_robisample(samples)) {
    if (verbose) {
      message("the samples table is converted to a robisample tibble")
    }
    samples <- robisample(samples)
  }

  if (!missing(motus) && !is_robimotu(motus)) {
    if (verbose) {
      message("the motus table is converted to a robimotu tibble")
    }
    motus <- robisample(motus)
  }

  if (is.matrix(data)) {
    if (verbose) message("Build data structure from matrix")

    if (!is.integer(data) || !is.numeric(data)) {
      dm <- dim(data)
      rn <- as.numeric(data)
      dim(rn) <- dm
      data <- rn
    }

    motus_names <- colnames(data)
    sample_names <- rownames(data)


    ##
    ##
    ## Case 1 : the provided data is a matrix
    ##
    ##

    if (!missing(samples))
    {
      all_ids = which(vapply(samples,is_robiuniqueid,FALSE))
      if (length(ids) == 1) {
        all_ids <- samples[[all_ids]]
        if (!is.null(sample_names))
        {
          sample_names <- factor(sample_names,levels = levels(all_ids))
          if (any(is.na(sample_names)))
            stop("The rownames of the data matrix and the samples table don't correspond",
                 call. = FALSE)
        }
        else {
          if (nrow(data) == length(all_ids))
            sample_names <- factor(all_ids,levels = levels(all_ids))
          else
            stop("The count of row of the samples table and of the data matrix are not consistant",
                 call. = FALSE)
        }
      }
      else
        stop("samples must have one and only one id column",
             call. = FALSE)
    } else {
      if (verbose) message("missing sample descriptions")

      if (!is.null(sample_names))
        samples = robisample(tibble(id = as_robiuniqueid(sample_names)))
      else {
        if (verbose) message("no row names on the data matrix")
        samples = robisample(n = c(pcr = nrow(data)))
      }
      sample_names = samples$id

    }

    if (!missing(motus))
    {
      all_ids = which(vapply(motus,is_robiuniqueid,FALSE))
      if (length(ids) == 1) {
        all_ids <- samples[[all_ids]]
        if (!is.null(motus_names))
        {
          motus_names <- factor(motus_names,levels = levels(all_ids))
          if (any(is.na(motus_names)))
            stop("The colname of the data matrix and the motus table don't correspond",
                 call. = FALSE)
        }
        else {
          if (ncol(data) == length(all_ids))
            motus_names <- factor(all_ids,levels = levels(all_ids))
        else
           stop("The count of row of the samples table and of column of the data matrix are not consistant",
                 call. = FALSE)
        }
      }
      else
        stop("motus must have one and only one id column",
             call. = FALSE)
    }
    else {
      if (verbose) message("missing motus descriptions")
      if (!is.null(motus_names))
        motus = robimotu(tibble(id = as_robiuniqueid(motus_names)))
      else {
        if (verbose) message("no column names on the data matrix")
        motus = robimotu(n = ncol(data))
      }
      motus_names = motus$id
    }

    colnames(data) <- as.character(motus_names)
    data <- cbind(tibble(sample = sample_names), as_tibble(data)) %>%
      tidyr::gather(key = "motu",
                    value = "count",
                    -sample)
  }


  ##
  ##
  ## At that level if the provided data was a matrix
  ## it is now a three columns tibble
  ##
  ##

  ##
  ## The tibble must have at least three columns
  ##
  ##  - motu
  ##  - sample
  ##  - count
  ##

  if (any(is.na(match(c("motu","sample","count"),colnames(data))))) {
    missed <- which(is.na(match(c("motu","sample","count"),colnames(data))))
    nmissed <- length(missed)
    missed <- paste0("(",paste(c("motu","sample","count")[missed],
                               collapse = ", "),
                     ")")
    stop(ngettext(nmissed,
             sprintf("the column %s is missing from the data table",missed),
             sprintf("the columns %s are missing from the data table",missed)
             ))
  }

  if (missing(motus)) {
    if (verbose) message("missing MOTU descriptions")

    motu_names <- sort(unique(data$motu))
    motus <- robimotu(motu_names)
  }

  #
  # If no samples table provided, a fake one is
  # created from the row names of reads or
  # row positions if no row names are available
  #

  if (missing(samples)) {
    if (verbose) message("missing sample descriptions")

    sample_names <- sort(unique(data$motu))
    samples <- robisample(sample_names)

  }

  #
  # motus and samples must be cohercible to tibble
  #

  data  <- as_tibble(data)
  motus <- as_robimotu(motus)
  samples <- as_robisample(samples)

  new_robimetabar(data = data,
                  samples = samples,
                  motus = motus,
                  distances = distances)
}



#' Retrieve the dimension of a `robimetabar` object.
#'
#' @param x a `robimetabar` instance
#'
#' @return Dim returns a vector of two numbers respectively the number
#'         of samples and the number of MOTUs.
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#'   metabar <- robimetabar(reads = reads)
#'   dim(metabar)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
dim.robimetabar <- function(x) {
  x$dim()
}

#' Retrieve the dimnames of an object.
#'
#' @param x a `robimetabar` instance
#'
#' @return The dimnames function returns a list of length equals to two.
#'         Each element of this list is a character vector.
#'         The first element contains sample names, the second MOTU names
#'
#' @examples
#'   reads <- matrix(1:12, nrow = 3)
#'   rownames(reads) <- c("sample_1", "sample_2", "sample_3")
#'   colnames(reads) <- c("motu_1", "motu_2", "motu_3", "motu_4")
#'
#' metabar <- robimetabar(reads = reads)
#' dimnames(metabar)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
dimnames.robimetabar <- function(x) {
  x$dimnames()
}


#' @param x
#'
#' @author Eric Coissac
#' @export
names.robimetabar <- function(x) {
  c("reads",
    eval(quote(names(layers)),envir = x)
   )
}


#' @rdname motus
#' @export
samples.robimetabar <- function(data, value) {
  data$samples
}

#' @rdname motus
#' @export
motus.robimetabar <- function(data) {
  data$motus
}


#' @author Eric Coissac
#' @export
dimnames.robidata <- function(x) {
  ux <- unmake.robiobject(x)
  list(
    ux[[attributes(x)$robi.ids]],
    names(ux)
  )
}


#' @rdname nmotus
#' @export
`nmotus.robimetabar` <- function(data) {
  ncol(data)
}

#' @rdname nsamples
#' @export
`nsamples.robimetabar` <- function(data) {
  nrow(data)
}

#' Converts a R object into a robimetabar instance
#'
#' @author Eric Coissac
#' @export
as_robimetabar <- function(data, ...) {
  UseMethod("as_robimetabar", data)
}

#' @rdname taxids
#' @export
#'
taxids_colname.robimetabar <- function(object) {
  motus(object) %>% taxids_colname()
}

#' @rdname taxids
#' @export
#'
master_taxids_colname.robimetabar <- function(object) {
  motus(object) %>% master_taxids_colname()
}

#' @rdname taxids
#' @export
#'
taxids.robimetabar <- function(object) {
  motus(object) %>% taxids()
}

#' @rdname taxids
#' @export
#'
master_taxids.robimetabar <- function(object) {
  motus(object) %>% master_taxids()
}

#' @rdname tags
#' @export
#'
tags_colname.robimetabar <- function(object) {
  samples(object) %>% tags_colname()
}

#' @rdname tags
#' @export
#'
forward_tags_colname.robimetabar <- function(object) {
  samples(object) %>% forward_tags_colname()
}

#' @rdname tags
#' @export
#'
reverse_tags_colname.robimetabar <- function(object) {
  samples(object) %>% reverse_tags_colname()
}

#' @rdname tags
#' @export
#'
tags.robimetabar <- function(object) {
  samples(object) %>% tags()
}

#' @rdname tags
#' @export
#'
forward_tags.robimetabar <- function(object) {
  samples(object) %>% forward_tags()
}

#' @rdname tags
#' @export
#'
reverse_tags.robimetabar <- function(object) {
  samples(object) %>% reverse_tags()
}


#' @rdname categories
#' @export
#'
categories.robimetabar <- function(object) {
  samples(object) %>% categories()
}

#' @rdname categories
#' @export
#'
categories_colname.robimetabar <- function(object) {
  samples(object) %>% categories_colname()
}

#' @rdname motus_ids_colname
#' @export
motus_ids_colname.robimetabar <- function(object) {
   motus(object) %>% ids_colname()
}

#' @rdname motus_ids_colname
#' @export
motus_ids.robimetabar <- function(object) {
   motus(object) %>% ids()
}

#' @rdname samples_ids_colname
#' @export
samples_ids_colname.robimetabar <- function(object) {
   samples(object) %>% ids_colname()
}

#' @rdname samples_ids_colname
#' @export
samples_ids.robimetabar <- function(object) {
   samples(object) %>% ids()
}

#' Clones a robimetabar instance
#'
#' @param object the robimetabar instance to be cloned
#'
#' @return a copy of object
#' @export
#'
#' @examples
clone.robimetabar <- function(object) {
  object$clone()
}

