#' @include robimetabar.R
#' @import dplyr
#'
NULL

filter_data_orphean <- function(metabar) {
  metabar$data[ metabar$data$motu %in% pull(motus_ids(metabar)) &
                metabar$data$sample %in% pull(samples_ids(metabar)),] -> metabar$data

  invisible(metabar)
}

#' Remove samples with no more reads associated to them.
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
filter_sample_empty <- function(metabar) {
    metabar$data %>%
      group_by(sample) %>%
      summarise(.total__ = sum(count)) %>%
      filter(.total__ > 0) %>%
      pull(sample) -> skeep

    samples(metabar)[(samples_ids(metabar) %>% pull())  %in% skeep,] -> metabar$samples

    invisible(filter_data_orphean(metabar))
}

#' Remove MOTUs with no more reads associated to them.
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
filter_motus_empty <- function(metabar) {
    metabar$data %>%
      group_by(motu) %>%
      summarise(.total__ = sum(count)) %>%
      filter(.total__ > 0) %>%
      pull(motu) -> skeep

    motus(metabar)[(motus_ids(metabar) %>% pull())  %in% skeep,] -> metabar$motus

    invisible(filter_data_orphean(metabar))
}

#' Filters out samples based on read count.
#'
#' Filters out samples having less than the specified minimum
#' number of reads in total.
#'
#' @param metabar
#' @param min_read
#' @param remove_empty_samples
#' @param remove_empty_motus
#'
#' @return the modified robimetabar object
#' @export
#' @md
#' @seealso `filter_sample_empty`, `filter_motus_empty`
#' @examples
filter_sample_min_read <- function(metabar,min_read,
                                   remove_empty_samples = TRUE,
                                   remove_empty_motus = TRUE) {
  metabar$data %>%
    group_by(sample) %>%
    mutate(.total__ = sum(count)) %>%
    filter(.total__ >= min_read) %>%
    select(-.total__) %>%
    ungroup() -> metabar$data

  if (remove_empty_samples)
     filter_sample_empty(metabar)

  if (remove_empty_motus)
     filter_motus_empty(metabar)

  invisible(metabar)
}

#' Filters out samples based on read count.
#'
#' `filter_sample_max_read` filters out samples having more than the specified maximum
#' number of reads in total.
#'
#' @param max_read
#'
#'
#' @rdname filter_sample_min_read
#' @md
#' @examples
#' @export
filter_sample_max_read <- function(metabar,max_read,
                                   remove_empty_samples = TRUE,
                                   remove_empty_motus = TRUE) {
  metabar$data %>%
    group_by(sample) %>%
    mutate(.total__ = sum(count)) %>%
    filter(.total__ <= max_read) %>%
    select(-.total__) %>%
    ungroup() -> metabar$data

  if (remove_empty_samples)
     filter_sample_empty(metabar)

  if (remove_empty_motus)
     filter_motus_empty(metabar)

  invisible(metabar)
}

#' Filters out samples based on diversity.
#'
#' Filters out samples having a diversity below the specified minimum
#' Hill's number specified.
#'
#' @param metabar
#' @param remove_empty_samples
#' @param remove_empty_motus
#' @param min_diversity
#' @param q
#'
#' @return the modified robimetabar object
#' @export
#'
#' @examples
filter_sample_min_diversity <- function(metabar,min_diversity, q = 1,
                                   remove_empty_samples = TRUE,
                                   remove_empty_motus = TRUE) {
  metabar$data %>%
    group_by(sample) %>%
    mutate(.diversity__ = D_q(count,q = q)) %>%
    filter(.diversity__ >= min_diversity) %>%
    select(-.diversity__) %>%
    ungroup() -> metabar$data

  if (remove_empty_samples)
     filter_sample_empty(metabar)

  if (remove_empty_motus)
     filter_motus_empty(metabar)

  invisible(metabar)
}

#' Filters out samples based on diversity.
#'
#' Filters out samples having a diversity below the specified minimum
#' Hill's number specified.
#'
#' @param metabar
#' @param remove_empty_samples
#' @param remove_empty_motus
#' @param min_diversity
#' @param q
#'
#' @return the modified robimetabar object
#' @export
#'
#' @rdname filter_sample_min_diversity
#' @examples
filter_sample_max_diversity <- function(metabar,max_diversity, q = 1,
                                   remove_empty_samples = TRUE,
                                   remove_empty_motus = TRUE) {
  metabar$data %>%
    group_by(sample) %>%
    mutate(.diversity__ = D_q(count,q = q)) %>%
    filter(.diversity__ <= min_diversity) %>%
    select(-.diversity__) %>%
    ungroup() -> metabar$data

  if (remove_empty_samples)
     filter_sample_empty(metabar)

  if (remove_empty_motus)
     filter_motus_empty(metabar)

  invisible(metabar)
}

#' Filters out samples based on category
#'
#' Filters out samples having not belonging the specified sample categories.
#'
#' @param metabar
#' @param remove_empty_samples
#' @param category
#' @param remove_empty_motus
#'
#' @return the modified robimetabar object
#' @export
#'
#' @examples
filter_sample_category <- function(metabar,category,
                                   remove_empty_samples = TRUE,
                                   remove_empty_motus = TRUE) {

  samples(metabar)[pull(categories(metabar)) %in% category,] -> metabar$samples

  if (remove_empty_samples)
     filter_sample_empty(metabar)

  if (remove_empty_motus)
     filter_motus_empty(metabar)

  invisible(metabar)
}