#' @include robimetabar.R
#' @import dplyr
#'
NULL


#' Makes stats about the read counts per motu
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
motus_read_count <- function(metabar) {
  bys = motus_ids_colname(metabar)

  suppressWarnings(
    select(metabar$motus, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := motu) %>%
                  summarise(read_count = sum(count)),
                by = bys)) %>%
    mutate(read_count = ifelse(is.na(read_count),
                               0,
                               read_count))
}

#' Makes stats about in how many samples a MOTU occur
#'
#' @param metabar
#' @param threshold
#'
#' @return
#' @export
#'
#' @examples
motus_sample_count <- function(metabar, threshold = 1) {
  bys = motus_ids_colname(metabar)

  suppressWarnings(
    select(metabar$motus, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := motu) %>%
                  summarise(sample_count = sum(count >= threshold)),
                by = bys)) %>%
    mutate(sample_count = ifelse(is.na(sample_count),
                               0,
                               sample_count))
}

#' Makes stats about in how many samples a MOTU occur
#'
#' @param metabar
#' @param threshold
#'
#' @return
#' @export
#'
#' @examples
motus_sample_max <- function(metabar) {
  bys = motus_ids_colname(metabar)

  suppressWarnings(
    select(metabar$motus, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := motu) %>%
                  summarise(sample_max = sample[which.max(count)][1],
                            sample_count_max = max(count)),
                  by = bys)) %>%
    mutate(sample_count_max = ifelse(is.na(sample_count_max),
                               0,
                               sample_count_max))
}


#' Makes stats about in how many samples a MOTU occur
#'
#' @param metabar
#' @param threshold
#'
#' @return
#' @export
#'
#' @examples
motus_sample_min <- function(metabar) {
  bys = motus_ids_colname(metabar)

  suppressWarnings(
    select(metabar$motus, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := motu) %>%
                  summarise(sample_min = sample[which.min(count)][1],
                            sample_count_min = min(count)),
                  by = bys)) %>%
    mutate(sample_count_min = ifelse(is.na(sample_count_min),
                               0,
                               sample_count_min))
}

#' Makes stats about the read counts per sample
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
samples_read_count <- function(metabar) {

  bys = samples_ids_colname(metabar)

  suppressWarnings(
    select(metabar$samples, bys) %>%
      left_join(group_by(metabar$data,
                         !!bys := sample) %>%
                  summarise(read_count = sum(count)),
                by = bys)) %>%
    mutate(read_count = ifelse(is.na(read_count),
                               0,
                               read_count))
}

#' Makes stats about the motu counts per sample
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
samples_motu_count <- function(metabar, threshold = 1) {

  bys = samples_ids_colname(metabar)

  suppressWarnings(
    select(metabar$samples, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := sample) %>%
                  summarise(motu_count = sum(count >= threshold)),
                by = bys)) %>%
    mutate(motu_count = ifelse(is.na(motu_count),
                               0,
                               motu_count))
}

#' Makes stats about the motu counts per sample
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
samples_motu_max <- function(metabar, threshold = 1) {

  bys = samples_ids_colname(metabar)

  suppressWarnings(
    select(metabar$samples, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := sample) %>%
                  summarise(motu = motu[which.max(count)][1],
                            motu_count_max = max(count)),
                  by = bys)) %>%
    mutate(motu_count_max = ifelse(is.na(motu_count_max),
                               0,
                               motu_count_max))
}

#' Makes stats about the motu counts per sample
#'
#' @param metabar
#'
#' @return
#' @export
#'
#' @examples
samples_motu_min <- function(metabar, threshold = 1,
                             verbose = is_robi_verbose()) {

  bys = samples_ids_colname(metabar)

  suppressWarnings(
    select(metabar$samples, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := sample) %>%
                  summarise(motu = motu[which.min(count)][1],
                            motu_count_min = min(count)),
                  by = bys)) %>%
    mutate(motu_count_min = ifelse(is.na(motu_count_min),
                               0,
                               motu_count_min))
}

#' Title
#'
#' @param metabar
#' @param q
#' @param verbose
#'
#' @return
#' @export
#'
#' @examples
samples_hill <- function(metabar, q = 0, verbose = is_robi_verbose()) {
  if (length(q) > 1) {
    message("length of the q parameter is greater than 1, only the first value will be used")
    q <- q[[1]]
  }

  bys = samples_ids_colname(metabar)

  suppressWarnings(
    select(metabar$samples, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := sample) %>%
                  summarise(hill = D_q(count,
                                       q = q,
                                       normalize = TRUE),
                            q = q),
                  by = bys)) %>%
    mutate(hill = ifelse(is.na(hill),
                               0,
                               hill))

}

#' Title
#'
#' @param metabar
#' @param q
#' @param verbose
#'
#' @return
#' @export
#'
#' @examples
samples_entropy <- function(metabar, q = 0, verbose = is_robi_verbose()) {
  if (length(q) > 1) {
    message("length of the q parameter is greater than 1, only the first value will be used")
    q <- q[[1]]
  }

  bys = samples_ids_colname(metabar)

  suppressWarnings(
    select(metabar$samples, all_of(bys)) %>%
      left_join(group_by(metabar$data,
                         !!bys := sample) %>%
                  summarise(entropy = H_q(count,
                                       q = q,
                                       normalize = TRUE),
                            q = q),
                  by = bys)) %>%
    mutate(entropy = ifelse(is.na(entropy),
                               0,
                               entropy))
}
