#' @include type_robiuniqueid.R
#' @include type_robitag.R
#' @include type_robitaxid.R
#' @include robidata.R
#' @import tibble
#' @import dplyr
NULL

#' Builds a serie of motu names.
#'
#' motu names are build by concatenating a group name (facutative), a motu
#' set id, a motu number in the set and a PCR repeat id (facutative).
#'
#' @param n an \code{integer} vector indicating the number of motu names to
#'   generate per category. If \code{n} has names, they are used as category
#'   names. Otherwhise category names are generated by concatenating the word
#'   \code{motu} to a serie of letters.
#'
#' @param ndigits a singme integer value indicating on how many digits the motu
#'   numerical serie number have to be padded with 0s.
#'
#' @param pcr_replicates a number indicating the number of pcr replicates or a
#'   \code{character} vector with the name of the PCR replicates.
#'
#' @param groups a \code{character} vector of motu group names for which the
#'   motu design have to be repeated.
#'
#' @param sep a character string to separate the name parts
#' @param verbose    if `TRUE` warnings are emitted if the constructor takes
#'   some decision because of missing values and estimates default values these
#'   missing data.
#'
#' @examples
#'   build_motu_names(5)
#'
#'   build_motu_names(c(5, 3))
#'
#'   build_motu_names(c(SAMP=5, CTRL=3), ndigits = 2)
#'
#'   build_motu_names(c(SAMP=5, CTRL=3), ndigits = 2,
#'                      pcr_replicates = 3)
#'
#'   build_motu_names(c(SAMP=5, CTRL=3), ndigits = 2,
#'                      pcr_replicates = c("A","B","C"))
#'
#'   build_motu_names(c(SAMP=5, CTRL=3), ndigits = 2,
#'                      pcr_replicates = c("A","B","C"),
#'                      groups=c("S1","S2"))
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
build_motu_names <- function(n,
                             ndigits = max(floor(log10(n)) + 1),
                             sep = "_",
                             verbose = is_robi_verbose()) {

    if (!is.numeric(n))
    {
      stop("'n' must be a numeric value")
    }

    prefixes = names(n)

    if (is.null(prefixes)) {
      prefixes <- if (length(n) == 1)
                      "motu"
                  else
                      paste0("motu",
                             combine_LETTERS(length(n)))
    }

    prefixes <- rep(prefixes,n)
    ns <- unlist(lapply(n, seq_len))

    snames <- sprintf("%s%0*d",prefixes,ndigits,ns)

    as_robiuniqueid(snames)
}


#' @details The \code{\link[ROBITools2]{robimotu}} version
#'     ckecks that if \code{\link[ROBITools2]{robitaxid}} columns
#'     are present that at least one, but no more, is a
#'     \code{\link[ROBITools2]{robitaxid_master}} instance.
#'
#' @rdname validate_robiobject
#' @export
#'
#' @examples
#'
#' # For robimotu instance
#' s <- robimotu(n=20)
#' validate_robiobject(s)
#'
validate_robiobject.robimotu <- function(object, verbose=is_robi_verbose()) {

  count_taxid = sum(vapply(object,is,TRUE,"robitaxid"))

  robiassert_arg(
    (count_taxid == 0) || sum(vapply(object,is,TRUE,"robitaxid_master")) == 1,
    "object",
    ngettext(count_taxid,
            "The unique column describing taxids, must be master",
            "Among the {count_taxid} columns describing taxids, one must be master not {master}"
            ),
    count_taxid = count_taxid,
    master = sum(vapply(object,is,TRUE,"robitaxid_master")))

  object
}

#' Test if an object belongs \code{robimotu} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#'   x <- robimotu(n=10)
#'   is_robimotu(x)
#'   is_robimotu("toto")
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robimotu = function(object) {
  inherits(object,'robimotu')
}


#' Title
#'
#' @param object the object to modify
#' @param ... Passed on to \code{\link[base]{structure}}()
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#'
#' @return
#' @export
new_robimotu <- function(object,...,class=NULL) {

  validate_robiobject(
    new_robidata(object,
      ...,
      class = c(class, "robimotu")
    )
  )
}

#' Build a \code{robimotu} instance.
#'
#' \code{robimotu} class is a specialisation of \code{\link[dplyr]{tibble}}.
#' It aims to store information on motus
#'
#' @param motus
#' @param n an integer value indicating the number of motu. By default equat
#' @param motu_ids the column name in the `motus` table containing
#'                   the motu identifiers.
#'
#' @param motu_categories
#' @param verbose    if `TRUE` warnings are emitted if the constructor takes
#'                   some decision because of missing values and estimates
#'                   default values these missing data.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
robimotu <- function(table,
                      n = ifelse(is.null(dim(table)),
                                 length(table),
                                 nrow(table)),
                      motu_ids = "id",
                      verbose = is_robi_verbose()) {


  if (missing(table)) {
    if (missing(n)) {
      stop("'table' or 'n' must be provided")
    }

    if (!is.numeric(n))
    {
      stop("'n' must be a numeric value")
    }

    motu=build_motu_names(n)

    table <- tibble(motu)
    colnames(table) <- motu_ids
  }

  if (is_atomic(table)) {
    table = tibble(table)
    colnames(table) = motu_ids
  }

  data <- as_tibble(table)

  if (!motu_ids %in% colnames(data))
    data[[motu_ids]] <- build_motu_names(nrow(data))

  if (!is_robiuniqueid(data[[motu_ids]]))
    data[[motu_ids]] <- robiuniqueid(data[[motu_ids]])

  as_robimotu(data)
}

#' @export
as_robimotu <- function(object) {
  UseMethod("as_robimotu", object)
}

#' @export
as_robimotu.default <- function(object) {
  new_robimotu(object)
}

#' @export
as_robimotu.robimotu <- function(object) {
  classes <- class(object)
  rclass <- which(classes == 'robimotu')
  class(object) <- classes[rclass:length(classes)]
  object
}

#' @export
`[[<-.robimotu` <- function(x, i, j, value) {
  as_robimotu(NextMethod())
}

#' @export
`$<-.robimotu` <- function(x, i, j, value) {
  as_robimotu(NextMethod())
}


#' taxids of a \code{robimotu} instance
#'
#' @details Extracts the \code{\link[ROBITools2]{robiuniqueid}} columns
#' from a \code{\link[ROBITools2]{robidata}} instance.
#'
#' @param object the object to analyse
#'
#' @return a \code{\link[tidyr]{tibble}} instance. Moreover the
#'        column containing te sample ids, it contains
#'        a  \code{\link[ROBITools2]{robitaxid_master}} column,
#'        and eventually several other \code{\link[ROBITools2]{robitaxid}}
#'        columns or no more column if no taxids are defined for the motus.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
taxids <- function(object) {
  UseMethod("taxids", object)
}


#' @details Extracts the names of the \code{\link[ROBITools2]{robitaxid}}
#' column from a \code{\link[ROBITools2]{robimotu}} instance.
#'
#' @return a \code{character} string.
#'
#' @rdname taxids
#' @export
taxids_colname <- function(object) {
  UseMethod("taxids_colname", object)
}


#' @rdname taxids
#' @export
taxids.default <- function(object) {
  stop(sprintf("No taxids method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname taxids
#' @export
#'
#' @examples
#'
#'   s <- robimotu(n=4)
#'   taxids_colname(s)
#'
taxids_colname.tbl <- function(object) {
  colnames(object)[which(vapply(object, is_robitaxid, FUN.VALUE = TRUE))]
}

#' @rdname taxids
#' @export
#'
#' @examples
#'
#'   s <- robimotu(n=4)
#'   taxids(s)
#'
taxids.robimotu <- function(object) {
  to_keep = c(ids_colname(object),
              sort(taxids_colname(object)))
  select(object,to_keep)  %>% as_robimotu()
}

#' @rdname taxids
#' @export
master_taxids <- function(object) {
  UseMethod("master_taxids", object)
}


#' @details Extracts the name of the \code{\link[ROBITools2]{robitaxid_master}}
#' column from a \code{\link[ROBITools2]{robimotu}} instance.
#'
#' @return a \code{character} string.
#'
#' @rdname taxids
#' @export
master_taxids_colname <- function(object) {
  UseMethod("master_taxids_colname", object)
}


#' @rdname taxids
#' @export
master_taxids.default <- function(object) {
  stop(sprintf("No taxids method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname taxids
#' @export
#'
#' @examples
#'
#'   s <- robimotu(n=4)
#'   master_taxids_colname(s)
#'
master_taxids_colname.tbl <- function(object) {
  colnames(object)[which(vapply(object, is,
                                FUN.VALUE = TRUE,
                                class2 = "robitaxid_master"))] -> response

  if (length(response) == 0) {
    rescue = taxids_colname(object)
    if (length(rescue) == 1)
      response = rescue
  }

  response
}

#' @rdname taxids
#' @export
#'
#' @examples
#'
#'   s <- robimotu(n=4)
#'   taxids(s)
#'
master_taxids.robimotu <- function(object) {
  to_keep = c(ids_colname(object),
              sort(master_taxids_colname(object)))
  select(object,to_keep) %>% as_robimotu()
}

