#' @import vctrs
#' @import rlang
#' @include type_robiatomic.R
#'
NULL

#' Test if an object belongs \code{robimutation} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#'   x <- as_robimutation(256)
#'   is_robimutation(x)
#'   is_robimutation("toto")
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robimutation = function(object) {
  inherits(object,'robimutation')
}

#' @rdname validate_robiobject
#' @export
validate_robiobject.robimutation <- function(object) {

  if (any(!are_na(object) & object < 0)) {
    bad = unique(object[!is.na(object) && object < 1])
    stop(ngettext(length(bad),
                  sprintf("The mutation %d is negative or null",
                          bad[1]),
                  sprintf("The mutation (%s) are negative or null",
                          paste(bad,collapse = ","))))
  }

  object
}

#' Constructs a new \code{robimutation} instance.
#'
#' @param object an integer vector
#' @param ... Passed on to \code{\link[base]{structure}}()
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#' new_robimutation(1:10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
new_robimutation <- function(object,
                          ...,
                          verbose = is_robi_verbose(),
                          class = character()) {

    object <- vec_cast(object,integer())

  validate_robiobject(
      new_robiatomic(object,
      ...,
      class = c(class, "robimutation")
    )
  )
}


#' Build a \code{robimutation} instance
#'
#' a \code{robimutation} is an integer vector that is for storing numerical taxonomic
#' identifiers. They are normally used to store NCBI taxids. They have to be positive
#' and not null.
#'
#' @param x an object coercible to an \code{integer}
#' @param verbose    if `TRUE` warnings are emitted if the constructor takes
#'                   some decision because of missing values and estimates
#'                   default values these missing data.
#'
#' @return a \code{robimutation} instance
#'
#' @examples
#'
#'   robimutation(1:10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
robimutation <- function(x = 0) {
  new_robimutation(rep(NA,x))
}


#' @export
vec_ptype2.robimutation <- function(x, y, ...) {
  UseMethod("vec_ptype2.robimutation")
}

#' @export
vec_ptype2.robimutation.default <- function(x, y, ..., x_arg = "x", y_arg = "y") {
  vec_default_ptype2(x, y, x_arg = x_arg, y_arg = y_arg)
}

#' @export
vec_ptype2.robimutation.robimutation <- function(x, y, ...) robimutation()

#' @export
vec_ptype2.robimutation.integer <- function(x, y, ...) integer()

#' @export
vec_ptype2.integer.robimutation <- function(x, y, ...) integer()

#' @export
vec_ptype2.robimutation.double <- function(x, y, ...) double()

#' @export
vec_ptype2.double.robimutation <- function(x, y, ...) double()


#' @export
vec_cast.robimutation <- function(x, to, ...) UseMethod("vec_cast.robimutation")

#' @export
vec_cast.robimutation.default <- function(x, to, ...) vec_default_cast(x, to)

#' @export
vec_cast.robimutation.robimutation <- function(x, to, ...) x

#' @export
vec_cast.robimutation.integer <- function(x, to, ...)
  new_robimutation(x)

#' @export
vec_cast.integer.robimutation <- function(x, to, ...)
  vec_data(x)

#' @export
vec_cast.robimutation.double <- function(x, to, ...)
  new_robimutation(x)

#' @export
vec_cast.double.robimutation <- function(x, to, ...)
  vec_cast(vec_data(x),double())


#' @export
vec_cast.character.robimutation <- function(x, to, ...)
  .Call("ROBI_as_chararacter_mutation",vec_data(x))

#' @export
vec_cast.logical.robimutation <- function(x, to, ...)
  vec_cast(vec_data(x),logical())


#' Converts an object to a \code{robimutation} instance
#'
#' @param object the object to convert
#' @return a \code{robimutation} instance
#'
#' @examples
#'   x <- robimutation(c(1,2,2759))
#'   is_robimutation(x)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
as_robimutation = function(object) {
  UseMethod("as_robimutation", object)
}

#' @rdname as_robimutation
#' @export
as_robimutation.default = function(object) {
  vec_cast(object,robimutation())
}

#' @export
format.robimutation <- function(x, ...) {
  out <- .Call("ROBI_as_chararacter_mutation",vec_data(x))
  out[is.na(x)] <- NA
  out
}

#' @export
vec_ptype_abbr.robimutation <- function(x, ...) {
  "mutation"
}

#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.robimutation <- function(x, ...) {
  out <- format(x)
  out[is.na(x)] <- NA
  pillar::new_pillar_shaft_simple(out, align = "left")
}


#' @export
is_dist_1_or_0 <- function(x,y,lx=nchar(x),ly=nchar(y)) {
  data = tibble(x=x,y=y,lx=lx,ly=ly)
  as_robimutation(.Call("ROBI_is_d0_or_d1",data$x,data$y,data$lx,data$ly))
}

#' @export
mutation_position <- function(x) {
  robiassert_arg(is_robimutation(x),x,"x must be a robimutation instance")
  .Call("ROBI_position_mutation",x)
}


#' @export
origin_site <- function(x) {
  robiassert_arg(is_robimutation(x),x,"x must be a robimutation instance")
  .Call("ROBI_origin_mutation",x)
}

#' @export
mutated_site <- function(x) {
  robiassert_arg(is_robimutation(x),x,"x must be a robimutation instance")
  .Call("ROBI_to_mutation",x)
}

#' @export
substitution <- function(x) {
  robiassert_arg(is_robimutation(x),x,"x must be a robimutation instance")
  .Call("ROBI_substitution_mutation",x)
}


