#' @include type_robiuniqueid.R
#' @include robicategory.R
#' @include robidata.R
#' @import tibble
#' @import purrr
#' @import dplyr
NULL

#' Builds a serie of sample names.
#'
#' Sample names are build by concatenating a group name (facutative), a sample
#' set id, a sample number in the set and a PCR repeat id (facutative).
#'
#' @param n an \code{integer} vector indicating the number of sample names to
#'   generate per category. If \code{n} has names, they are used as category
#'   names. Otherwhise category names are generated by concatenating the word
#'   \code{sample} to a serie of letters.
#'
#' @param ndigits a singme integer value indicating on how many digits the sample
#'   numerical serie number have to be padded with 0s.
#'
#' @param pcr_replicates a number indicating the number of pcr replicates or a
#'   \code{character} vector with the name of the PCR replicates.
#'
#' @param groups a \code{character} vector of sample group names for which the
#'   sample design have to be repeated.
#'
#' @param sep a character string to separate the name parts
#' @param verbose    if `TRUE` warnings are emitted if the constructor takes
#'   some decision because of missing values and estimates default values these
#'   missing data.
#'
#' @examples
#'   build_sample_names(5)
#'
#'   build_sample_names(c(5, 3))
#'
#'   build_sample_names(c(SAMP=5, CTRL=3), ndigits = 2)
#'
#'   build_sample_names(c(SAMP=5, CTRL=3), ndigits = 2,
#'                      pcr_replicates = 3)
#'
#'   build_sample_names(c(SAMP=5, CTRL=3), ndigits = 2,
#'                      pcr_replicates = c("A","B","C"))
#'
#'   build_sample_names(c(SAMP=5, CTRL=3), ndigits = 2,
#'                      pcr_replicates = c("A","B","C"),
#'                      groups=c("S1","S2"))
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
build_sample_names <- function(n,
                               ndigits = max(floor(log10(n)) + 1),
                               pcr_replicates=NULL,
                               sep = "_",
                               groups = c(),
                               verbose = is_robi_verbose()) {

    if (!is.numeric(n))
    {
      stop("'n' must be a numeric value")
    }

    prefixes = names(n)

    if (is.null(prefixes)) {
      prefixes <- if (length(n) == 1)
                      "sample"
                  else
                      paste0("sample",
                             combine_LETTERS(length(n)))
    }

    prefixes <- rep(prefixes,n)
    ns <- unlist(lapply(n, seq_len))

    snames <- sprintf("%s%0*d",prefixes,ndigits,ns)

    if (length(groups)) {
      snames <- expand.grid(snames,groups)
      snames <- paste(snames[,2],snames[,1],sep = sep)
    }

    if (!missing(pcr_replicates)) {
      if (is.numeric(pcr_replicates) &&
          length(pcr_replicates) == 1) {
        nd = floor(log10(pcr_replicates)) + 1
        pcr_replicates <- sprintf("R%0*d",
                                  nd,
                                  1:pcr_replicates)
      }

      snames <- expand.grid(pcr_replicates,snames)
      snames <- paste(snames[,2],snames[,1],sep = sep)
    }

  as_robiuniqueid(snames)
}

#' @details The \code{\link[ROBITools2]{robisample}} version
#'     ckecks that a single column of.\code{\link[ROBITools2]{robicategory}}
#'     exist, and if tags columns are present it must be a single
#'     \code{\link[ROBITools2]{robitag}} or a single couple of
#'     \code{\link[ROBITools2]{robitag_forward}} and
#'     \code{\link[ROBITools2]{robitag_reverse}}
#'     columns
#' @rdname validate_robiobject
#' @export
#'
#' @examples
#'
#' # Validation of a robisample
#' s <- robisample(n=20)
#' validate_robiobject(s)
#'

validate_robiobject.robisample <- function(object, verbose=TRUE) {
  columns <- table(sapply(object,function(i) class(i)[1]))
  cc <- function(x) ifelse(is.na(columns[x]),0,columns[x])

  robiassert_arg(cc("robicategory") == 1,
                 "object",
                 "A {class} instance must a one robicategory column not {count}",
                 class = class(object)[1],
                 count = cc("robicategory"))

  robiassert_arg(cc("robitag_forward") <= 1,
                 "object",
                 "A {class} instance cannot have mot than a single robitag_forward column not {count}",
                 class = class(object)[1],
                 count = cc("robitag_forward"))

  robiassert_arg(cc("robitag_reverse") <= 1,
                 "object",
                 "A {class} instance cannot have mot than a single robitag_reverse column not {count}",
                 class = class(object)[1],
                 count = cc("robitag_reverse"))

  robiassert_arg(cc("robitag_forward") == cc("robitag_reverse"),
                 "object",
                 "A {class} instance must have a robitag_forward and robitag_reverse column",
                 class = class(object)[1])

  robiassert_arg(cc("robitag") <= 1,
                 "object",
                 "A {class} instance cannot have mot than a single robitag column not {count}",
                 class = class(object)[1],
                 count = cc("robitag"))

  robiassert_arg(cc("robitag") !=  cc("robitag_forward") ||
                   (cc("robitag") == 0 && cc("robitag_forward") == 0),
                 "object",
                 "A {class} instance cannot have robitag and robitag_forward columns simultaneously",
                 class = class(object)[1])

  object
}

#' Test if an object belongs \code{robisample} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#'   x <- robisample(n=10)
#'   is_robisample(x)
#'   is_robisample("toto")
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robisample = function(object) {
  inherits(object,'robisample')
}


#' Title
#'
#' @param object the object to modify
#' @param ... Passed on to \code{\link[base]{structure}}()
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#'
#' @return
#' @export
new_robisample <- function(object,...,class=NULL) {

  validate_robiobject(
    new_robidata(object,
      ...,
      class = c(class, "robisample")
    )
  )
}

#' Build a \code{robisample} instance.
#'
#' \code{robisample} class is a specialisation of \code{\link[dplyr]{tibble}}.
#' It aims to store information on samples
#'
#' @param samples
#' @param n an integer value indicating the number of sample. By default equat
#' @param sample_ids the column name in the `samples` table containing
#'                   the sample identifiers.
#'
#' @param sample_categories
#' @param verbose    if `TRUE` warnings are emitted if the constructor takes
#'                   some decision because of missing values and estimates
#'                   default values these missing data.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
robisample <- function(table,
                        n = ifelse(is.null(dim(table)),
                                   length(table),
                                   nrow(table)),
                        sample_ids = "id",
                        sample_categories = "category",
                        verbose = is_robi_verbose()) {


  if (missing(table)) {
    if (missing(n)) {
      stop("'samples' or 'n' must be provided")
    }

    if (!is.numeric(n))
    {
      stop("'n' must be a numeric value")
    }

    sample=build_sample_names(n)

    table <- tibble(sample)
    colnames(table) <- sample_ids
  }

  if (is_atomic(table)) {
    table = tibble(table)
    colnames(table) = sample_ids
  }

  data <- unclass_robiobject(table)

  if (! is_tibble(data))
    data <- as_tibble(data)

  if (!sample_ids %in% colnames(data))
    data[[sample_ids]] <- build_sample_names(nrow(data))

  if (!is_robiuniqueid(data[[sample_ids]]))
    data[[sample_ids]] <- as_robiuniqueid(data[[sample_ids]])

  if (!sample_categories %in% colnames(data))
    data[[sample_categories]] <- robicategory(nrow(data))

  if (!is_robicategory(data[[sample_categories]]))
    data[[sample_categories]] <- as_robicategory(data[[sample_categories]])

  as_robisample(data)
}


#' @export
as_robisample <- function(object) {
  UseMethod("as_robisample", object)
}

#' @export
as_robisample.default <- function(object) {
  new_robisample(object)
}

#' @export
as_robisample.robisample <- function(object) {
  classes <- class(object)
  rclass <- which(classes == 'robisample')
  class(object) <- classes[rclass:length(classes)]
  object
}

#' @export
`[[<-.robisample` <- function(x, i, j, value) {
  as_robisample(NextMethod())
}

#' @export
`$<-.robisample` <- function(x, i, j, value) {
  as_robisample(NextMethod())
}


#' tags of a \code{robisample} instance
#'
#' @details Extracts the \code{\link[ROBITools2]{robiuniqueid}} column
#' from a \code{\link[ROBITools2]{robidata}} instance.
#'
#' @param object the object to analyse
#'
#' @return a \code{\link[tidyr]{tibble}} instance. Moreover the
#'        column containing te sample ids, it can contain :
#'        a  single \code{\link[ROBITools2]{robitag}} column,
#'        a  \code{\link[ROBITools2]{robitag_forward}} and a
#'         \code{\link[ROBITools2]{robitag_reverse}} instance
#'         or no more column if no tags are defined for the samples.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
tags <- function(object) {
  UseMethod("tags", object)
}


#' @details Extracts the name of the \code{\link[ROBITools2]{robitag}}
#' column from a \code{\link[ROBITools2]{robisample}} instance.
#'
#' @return a \code{character} string.
#'
#' @rdname tags
#' @export
tags_colname <- function(object) {
  UseMethod("tags_colname", object)
}


#' @rdname tags
#' @export
tags.default <- function(object) {
  stop(sprintf("No tags method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname tags
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   tags_colname(s)
#'
tags_colname.robisample <- function(object) {
  colnames(object)[which(vapply(object, is_robitag, FUN.VALUE = TRUE))]
}

#' @rdname tags
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   tags(s)
#'
tags.robisample <- function(object) {
  to_keep = c(ids_colname(object),
              sort(tags_colname(object)))
  select(object,to_keep) %>% as_robidata()
}

#' @rdname tags
#' @export
forward_tags <- function(object) {
  UseMethod("forward_tags", object)
}


#' @details Extracts the name of the \code{\link[ROBITools2]{robitag_forward}}
#' column from a \code{\link[ROBITools2]{robisample}} instance.
#'
#' @return a \code{character} string.
#'
#' @rdname tags
#' @export
forward_tags_colname <- function(object) {
  UseMethod("forward_tags_colname", object)
}


#' @rdname tags
#' @export
forward_tags.default <- function(object) {
  stop(sprintf("No tags method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname tags
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   forward_tags_colname(s)
#'
forward_tags_colname.robisample <- function(object) {
  tname <- colnames(object)[which(vapply(object, is,
                                FUN.VALUE = TRUE,
                                class2 = "robitag_forward"))]

  if (length(tname) == 0)
    return(tags_colname(object))

  tname
}

#' @rdname tags
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   tags(s)
#'
forward_tags.robisample <- function(object) {
  to_keep = c(ids_colname(object),
              sort(forward_tags_colname(object)))
  select(object,to_keep)  %>% as_robidata()
}


#' @rdname tags
#' @export
reverse_tags <- function(object) {
  UseMethod("reverse_tags", object)
}


#' @details Extracts the name of the \code{\link[ROBITools2]{robitag_reverse}}
#' column from a \code{\link[ROBITools2]{robisample}} instance.
#'
#' @return a \code{character} string.
#'
#' @rdname tags
#' @export
reverse_tags_colname <- function(object) {
  UseMethod("reverse_tags_colname", object)
}


#' @rdname tags
#' @export
reverse_tags.default <- function(object) {
  stop(sprintf("No tags method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname tags
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   reverse_tags_colname(s)
#'
reverse_tags_colname.robisample <- function(object) {
  tname <- colnames(object)[which(vapply(object, is,
                                FUN.VALUE = TRUE,
                                class2 = "robitag_reverse"))]

  if (length(tname) == 0)
    return(tags_colname(object))

  tname
}

#' @rdname tags
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   tags(s)
#'
reverse_tags.robisample <- function(object) {
  to_keep = c(ids_colname(object),
              sort(reverse_tags_colname(object)))
  select(object,to_keep)  %>% as_robidata()
}


#' categories of a \code{robisample} instance
#'
#' @details Extracts the \code{\link[ROBITools2]{robiuniqueid}} column
#' from a \code{\link[ROBITools2]{robidata}} instance.
#'
#' @param object the object to analyse
#'
#' @return a \code{\link[tidyr]{tibble}} instance. Moreover the
#'        column containing te sample ids, it contains the
#'         \code{\link[ROBITools2]{robicategory}} column.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
categories <- function(object) {
  UseMethod("categories", object)
}


#' @details Extracts the name of the \code{\link[ROBITools2]{robicategory}}
#' column from a \code{\link[ROBITools2]{robisample}} instance.
#'
#' @return a \code{character} string.
#'
#' @rdname categories
#' @export
categories_colname <- function(object) {
  UseMethod("categories_colname", object)
}


#' @rdname categories
#' @export
categories.default <- function(object) {
  stop(sprintf("No categories method implemented for object object of class %s.",
               class(object)[1]))
}


#' @rdname categories
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   categories_colname(s)
#'
categories_colname.robisample <- function(object) {
  colnames(object)[which(vapply(object, is_robicategory, FUN.VALUE = TRUE))]
}

#' @rdname categories
#' @export
#'
#' @examples
#'
#'   s <- robisample(n=4)
#'   categories(s)
#'
categories.robisample <- function(object) {
  to_keep = c(ids_colname(object),
              sort(categories_colname(object)))
  select(object,to_keep) %>% as_robisample()
}
