#' @import dplyr
#' @import tibble
NULL


#' @author Eric Coissac
#' @export
read.sequence = function(con,keys,
                         n=-1,
                         skip = 0,
                         reader="fasta",
                         chunk.size=100000,
                         verbose=TRUE) {
  if (is.character(reader)){
    if (reader=="fasta")
      reader=Fasta.reader
    else
      stop('Unknown reader')
  }

  if (missing(keys))
    keys=NULL

  if (n > 0 && chunk.size > n)
    chunk.size = n

  iterator = reader(con=con,
                    keys=keys,
                    chunk.size=chunk.size)

  if (skip>0)
    for (i in seq_len(skip))
      seq=iterator()

  if (n==-1) {
    rep = vector(mode="list",length = 0)
    seq = iterator()
    i = 0
    while (! is.null(seq)) {
      i = i + 1
      rep[[i]]=seq
      seq = iterator()

      if (verbose && (i %% 10000)==0)
        message(paste(i,"sequences read"))
    }
  } else {
    rep = vector(mode="list",length = n)
    i = 0
    seq = iterator()
    while (i < n && ! is.null(seq)) {
      i = i + 1
      rep[[i]]=seq
      seq = iterator()

      if (verbose && (i %% 10000)==0)
        message(paste(i,"sequences read"))
    }
  }

  if (verbose )
    message(paste(i,"sequences read"))

  columns = Reduce(union,lapply(rep,names))
  ncol = length(columns)
  nrow = length(rep)
  data = vector(ncol*nrow,mode="list")
  dim(data)=c(nrow,ncol)
  colnames(data)=columns
  wide = integer(ncol)
  names(wide)=columns

  for (i in seq_along(rep)) {
    seq = rep[[i]]
    for (j in seq_along(columns)) {
      v = seq[[columns[j]]]
      data[i,j] = list(v)
      l = length(v)
      if (l > wide[j]) wide[j]=l
    }
  }

  if (verbose )
    message(paste(ncol,"variables extracted"))

  rep = vector(ncol,mode="list")
  names(rep)=columns
  for (j in seq_along(columns)) {
    if (wide[j]==1)
      rep[[j]]=simplify2array(data[,j])
    else {
        c = data[,j]
        subcols = sort(Reduce(union,lapply(c,names)))
        nsubcols= length(subcols)
        types = ordered(unique(vapply(c,class,"")),
                        levels=c("character",
                                 "numeric",
                                 "integer",
                                 "logical"))
        type = as.character(min(types))
        subtab = vector(nsubcols * nrow, mode = type)
        dim(subtab) = c(nrow,nsubcols)
        colnames(subtab)=subcols
        for (i in seq_along(c)) {
          d = c[[i]][subcols]
          subtab[i,] = d
        }
        rep[[j]]=subtab
    }
  }

  if (verbose )
    message("End of the data reformating")


  return(make.robiobject(rep,"robiseq_db"))
}

#' @author Eric Coissac
#' @export
`[.robiseq_db` = function(x, ..., drop = FALSE) {
  subset = function(d,i) {
    ds = length(dim(d))
    if (ds < 2) {
      return(d[i])
    }
    else
      return(d[i,,drop=FALSE])
  }

  size = function(d) {
    ds = length(dim(d))
    if (ds < 2)
      return(length(d))
    else
      return(nrow(d))
  }
  args = list(...)

  stopifnot(length(args)==1)

  i = args[[1]]

  if (is.character(i))
    i = match(i,x$id)

  rep = lapply(x, subset, i=i)

  if (drop && size(rep[[1]])==1)
    rep = do.call(build.robiseq,rep)
  else
    rep = make.robiobject(rep,"robiseq_db")

  return(rep)
}

#' @author Eric Coissac
#' @export
as.robimetabar.robiseq_db = function(data,motu.ids="id",read.count="merged_sample") {

  stopifnot(length(read.count)==1)

  rci = match(read.count,names(data))
  stopifnot(! is.na(rci))

  ii = match(motu.ids,names(data))
  stopifnot(! is.na(rci))

  rc = t(data[[rci]])
  rc[is.na(rc)]=0

  colnames(rc)=data[[ii]]

  motus = lapply(unmake.robiobject(data)[-rci],as.tibble)
  names.col = names(data)[-rci]

  for (i in seq_along(motus)){
    c=motus[[i]]
    if (ncol(c)==1) {
      colnames(c)=names.col[i]
      motus[[i]] = c
    }
  }

  motus = do.call(cbind,motus)
  samples = tibble(id = rownames(rc))

  rep = suppressWarnings(build.robimetabar(reads    = rc,
                                           motus    = motus,
                                           samples  = samples,
                                           motu.ids = motu.ids))

  return (rep)
}
