#' Defines all the methods available for the taxonomy object whatever their
#' implementation.
#'
#' @include utils.R
#'
NULL


#' @author Eric Coissac
#' @export
alternative_names = function(taxonomy) {
  UseMethod("alternative_names", taxonomy)
}

#' @author Eric Coissac
#' @export
alternative_taxids = function(taxonomy) {
  UseMethod("alternative_taxids", taxonomy)
}

#' @author Eric Coissac
#' @export
nodes = function(taxonomy) {
  UseMethod("nodes", taxonomy)
}

#' Converts a partial toxonomy to a full taxonomy.
#'
#' By looking into its reference taxonomy the partial
#' taxonomy is completed with the minimum set of taxa
#' allowing to respect the completude constraint asking
#' that parent taxa of every taxa are described in the
#' taxonomy.
#'
#' @param taxonomy
#'
#' @return a full taxonomy instance.
#'
#' @author Eric Coissac
#' @export
full_taxonomy = function(taxonomy) {
  UseMethod("full_taxonomy", taxonomy)
}


#' Predicat indicating is a taxonomy is full.
#'
#' A taxonomy is considered as full if all the taxid described
#' in its \code{parent} or \code{path} variables are described in
#' the taxonomy.
#'
#' Two verifications modes exist. A rapid one, that actually just check a
#' flag on the taxonomy instance, a \code{full_check} mode which actually
#' verify the constraint
#'
#' @param taxonomy the taxonomy to test.
#' @param full_check a logical value indicating the verification mode.
#'
#' @return a logical value.
#' @export
#'
#' @examples
#'
is_full_taxonomy <- function(taxonomy, full_check = FALSE) {
  UseMethod("is_full_taxonomy", taxonomy)
}


#' Returns the scientific name corresponding to a \emph{NCBI taxid}
#'
#' \code{scientificname} function in package \pkg{\link{ROBITaxonomy}} returns the
#' scientific name corresponding to a \emph{NCBI taxid}.
#'
#' @param taxonomy an object able to provide a taxonomy.
#' @param   taxid an integer value or a vector of integer representing NCBI
#'          taxonomic identifiers.
#' @return  The scientific name of the corresponding taxon as a string or a
#'          vector of string if the \code{taxid} argument is itself a vector
#'
#'
#' @author Eric Coissac
#' @export
scientific_name = function(data,taxonomy,results_in,...) {
  UseMethod("scientific_name", data)
}

#' Returns the taxonomic rank associated to a taxid
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   data a vector of taxid to analyse
#'
#' @return a vector of type \code{character} containing the taxonomic ranks
#'
#' @author Eric Coissac
#' @export
taxonomic_rank = function(data,taxonomy,results_in,...) {
  UseMethod("taxonomic_rank", data)
}


#' Returns the parent taxon corresponding to a \emph{NCBI taxid}
#'
#' \code{parent} function in package \pkg{\link{ROBITaxonomy}} returns the
#' parent taxon corresponding to a \emph{NCBI taxid}.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid an integer value or a vector of integer representing NCBI
#'          taxonomic identifiers.
#' @param   name A logical value \code{TRUE} or \code{FALSE} indicating if the
#'          method returns a taxid or a scientific name.
#'
#' @return  \describe{ \item{If \code{name==FALSE}}{the taxid of the
#'                            parent taxon as an integer or a vector of
#'                            integers if the \code{taxid} argument is itself
#'                            a vector}
#'                      \item{If \code{name==TRUE}}{the scientific name of the
#'                            parent taxon as a string or a vector of
#'                            string if  the \code{taxid} argument is itself a
#'                            vector} }
#'
#'
#' @author Eric Coissac
#' @export
taxonomic_parent = function(data,taxonomy, with_name = TRUE, results_in,...)  {
  UseMethod("taxonomic_parent", data)
}


#' Extracts the taxid at a specified taxonomic rank.
#'
#' The \code{taxonatrank} method of \code{\linkS4class{obitools.taxonomy}} class
#' returns the \emph{taxid} or the scientific name corresponding
#' to a \emph{taxid}.at a specified taxonomic rank
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid a vector of taxid to analyse
#' @param   rank a \code{character} indicating the desired rank
#' @param   name  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the method return a taxid or a scientific name.
#'
#' @return  \describe{
#'   \item{If \code{name==FALSE}}{the taxid of the corresponding
#'                                taxon as an integer or a vector of integers
#'                                if the \code{taxid} argument is itself
#'                                a vector}
#'   \item{If \code{name==TRUE}}{the scientific name of the corresponding
#'                               taxon as a string or a vector of string
#'                               if the \code{taxid} argument is itself
#'                               a vector}
#'  }
#'
#' @author Eric Coissac
#' @export
taxon_at_rank = function(data,taxonomy,rank,with_name=FALSE, results_in,...) {
  UseMethod("taxon_at_rank", data)
}



#' @author Eric Coissac
#' @export
expand_names <- function(taxonomy) {
  UseMethod("expand_names", taxonomy)
}




#' Checks that a \emph{taxid} is currently defined in a taxonomy.
#'
#' \code{taxid.validate} function  checks
#' that a \emph{taxid} is defined in the considered taxonomy.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid an integer value or a vector of integer representing NCBI
#'          taxonomic identifiers.
#'
#' @return  The taxid if it exists, NA otherwise. If the input taxid is a
#'          vector of integer returns an integer vector composed of validated
#'          taxids and NA values.
#'
#' @author Eric Coissac
#' @export
taxid.validate = function(taxonomy,taxid,na.rm=FALSE) {
  UseMethod("taxid.validate", taxonomy)
}

#'
#' @author Eric Coissac
#' @export
taxon.path = function(taxonomy,taxid,
                      name=FALSE,all.steps=TRUE,
                      na.rm=TRUE,
                      drop=TRUE) {
  UseMethod("taxon.path", taxonomy)
}


#' Computes the lowest common ancestor in the taxonomy tree between a set of taxa
#'
#' The \code{lowest.common.ancestor} function in package \pkg{ROBITaxonomy} computes
#' the lowest common ancestor of a set of taxids. The lowest common ancestor (LCA)
#' is the most precise taxonomic group shared by all the considered taxa. Tha
#' \code{lowest.common.ancestor} function implemented in the \pkg{ROBITaxonomy}
#' package, considers a fuzzy definition of the LCA as the most precise
#' taxonomic group shared by a quorum of the considered taxa.
#'
#' @param taxonomy an instance of \code{\linkS4class{obitools.taxonomy}}
#' @param taxid an integer value or a vector of integer representing NCBI
#'   taxonomic identifiers.
#' @param threshold a numeric value between 0.0 and 1.0 indicating the minimum
#' quorum of taxid that must belong the LCA.
#' @param error an integer value indicating the maximum count of taxids that
#' have not to belong the returned taxid. A \code{threshold} below 1.0 have
#'  priority on the \code{error} parameter.
#' @param name A logical value \code{TRUE} or \code{FALSE} indicating if the
#' method return a \emph{taxid} or a scientific name.
#'
#' @return Depending on the value of the \code{name} argument, set by default
#'         to \code{FALSE} the method returns :
#'    \describe{
#'           \item{If \code{name==FALSE}}{ the taxid of the taxon corresponding
#'             to the LCA as an integer value}
#'           \item{If \code{name==TRUE}}{ the scientific name of the taxon
#'                 corresponding to the LCA as a string}
#'          }
#'
#' @author Eric Coissac
#' @export
lowest_common_ancestor <- function(taxonomy,
                                   threshold = 1.0,
                                   error = 0,
                                   na.rm = TRUE,
                                   only_lca = FALSE) {
  UseMethod("lowest_common_ancestor", taxonomy)
}

#' Returns the list of taxonomic ranks
#'
#' The \code{rank.list} function returns the list of all taxonomic
#' ranks described in the taxonomy
#'
#' @param   taxonomy an object able to provide a taxonomy.
#'
#' @return a vector of type \code{character} containing the taxonomic rank names
#'
#' @author Eric Coissac
#' @export
rank.list = function(taxonomy) {
  UseMethod("rank.list", taxonomy)
}




#' Extracts the species corresponding to a taxid
#'
#' The \code{species} method of \code{\linkS4class{obitools.taxonomy}} class
#' returns the \emph{taxid} or the scientific name of the species corresponding
#' to a \emph{taxid}.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid a vector of taxid to analyse
#' @param   name  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the method return a taxid or a scientific name.
#'
#' @return  \describe{
#'   \item{If \code{name==FALSE}}{the taxid of the corresponding
#'                                taxon as an integer or a vector of integers
#'                                if the \code{taxid} argument is itself
#'                                a vector}
#'   \item{If \code{name==TRUE}}{the scientific name of the corresponding
#'                               taxon as a string or a vector of string
#'                               if the \code{taxid} argument is itself
#'                               a vector}
#'  }
#'
#' @author Eric Coissac
#' @export
species = function(taxonomy,taxid,name=FALSE,na.rm=TRUE) {
  UseMethod("species", taxonomy)
}

#' Extracts the genus corresponding to a taxid
#'
#' The \code{genus} method of \code{\linkS4class{obitools.taxonomy}} class
#' returns the \emph{taxid} or the scientific name of the genus corresponding
#' to a \emph{taxid}.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid a vector of taxid to analyse
#' @param   name  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the method return a taxid or a scientific name.
#'
#' @return  \describe{
#'   \item{If \code{name==FALSE}}{the taxid of the corresponding
#'                                taxon as an integer or a vector of integers
#'                                if the \code{taxid} argument is itself
#'                                a vector}
#'   \item{If \code{name==TRUE}}{the scientific name of the corresponding
#'                               taxon as a string or a vector of string
#'                               if the \code{taxid} argument is itself
#'                               a vector}
#'  }
#'
#' @author Eric Coissac
#' @export
genus = function(taxonomy,taxid,name=FALSE,na.rm=TRUE) {
  UseMethod("genus", taxonomy)
}


#' Extracts the family corresponding to a taxid
#'
#' The \code{family} method of \code{\linkS4class{obitools.taxonomy}} class
#' returns the \emph{taxid} or the scientific name of the family corresponding
#' to a \emph{taxid}.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid a vector of taxid to analyse
#' @param   name  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the method return a taxid or a scientific name.
#'
#' @return  \describe{
#'   \item{If \code{name==FALSE}}{the taxid of the corresponding
#'                                taxon as an integer or a vector of integers
#'                                if the \code{taxid} argument is itself
#'                                a vector}
#'   \item{If \code{name==TRUE}}{the scientific name of the corresponding
#'                               taxon as a string or a vector of string
#'                               if the \code{taxid} argument is itself
#'                               a vector}
#'  }
#'
#' @author Eric Coissac
#' @export
family = function(taxonomy,taxid,name=FALSE,na.rm=TRUE) {
  UseMethod("family", taxonomy)
}

#' Extracts the kingdom corresponding to a taxid
#'
#' The \code{kingdom} method of \code{\linkS4class{obitools.taxonomy}} class
#' returns the \emph{taxid} or the scientific name of the kingdom corresponding
#' to a \emph{taxid}.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid a vector of taxid to analyse
#' @param   name  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the method return a taxid or a scientific name.
#'
#' @return  \describe{
#'   \item{If \code{name==FALSE}}{the taxid of the corresponding
#'                                taxon as an integer or a vector of integers
#'                                if the \code{taxid} argument is itself
#'                                a vector}
#'   \item{If \code{name==TRUE}}{the scientific name of the corresponding
#'                               taxon as a string or a vector of string
#'                               if the \code{taxid} argument is itself
#'                               a vector}
#'  }
#'
#' @author Eric Coissac
#' @export
kingdom = function(taxonomy,taxid,name=FALSE,na.rm=TRUE) {
  UseMethod("kingdom", taxonomy)
}

#' Extracts the superkingdom corresponding to a taxid
#'
#' The \code{superkingdom} method of \code{\linkS4class{obitools.taxonomy}} class
#' returns the \emph{taxid} or the scientific name of the superkingdom corresponding
#' to a \emph{taxid}.
#'
#' @param   taxonomy an object able to provide a taxonomy.
#' @param   taxid a vector of taxid to analyse
#' @param   name  A logical value \code{TRUE} or \code{FALSE} indicating
#'          if the method return a taxid or a scientific name.
#'
#' @return  \describe{
#'   \item{If \code{name==FALSE}}{the taxid of the corresponding
#'                                taxon as an integer or a vector of integers
#'                                if the \code{taxid} argument is itself
#'                                a vector}
#'   \item{If \code{name==TRUE}}{the scientific name of the corresponding
#'                               taxon as a string or a vector of string
#'                               if the \code{taxid} argument is itself
#'                               a vector}
#'  }
#'
#' @author Eric Coissac
#' @export
superkingdom = function(taxonomy,taxid,name=FALSE,na.rm=TRUE) {
  UseMethod("superkingdom", taxonomy)
}
