#' @import utils
#' @import readr
#' @include taxonomy_robi.R
#' @useDynLib ROBITools2
#'
NULL

#' Read the rank list of a taxonomy
#'
#' @param taxoname a string indicating the path of the taxonomy
#'
#' @return a character vector
#'
#' @examples
#'
#' @rdname internal.read_ecopcr_rank
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
.read_ecopcr_rank <- function(taxoname) {
  vapply(.Call("R_raw_to_ecopcr_record",
               read_file_raw(paste0(taxoname,".rdx"))),
         rawToChar,
         FUN.VALUE = "")
}

#' Read the name list of a taxonomy
#'
#' @param taxoname a string indicating the path of the taxonomy
#'
#' @return a tibble with three column \code{taxid_idx},\code{name_class},\code{name}.
#'
#' @examples
#'
#' @rdname internal.read_ecopcr_name
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
.read_ecopcr_name <- function(taxoname) {
  data <- .Call("R_raw_to_ecopcr_tnames",
                read_file_raw(paste0(taxoname,".ndx")))

  names(data) <- c("taxid_idx","name_class","name")
  data$name_class <- factor(data$name_class)
  data$taxid_idx <- data$taxid_idx + 1
  as_tibble(data)
}

#' Reads an OBITools formated taxonomy.
#'
#'
#' @param taxoname a string indicating the path of the taxonomy
#' @param keep_alternative_names
#' @param keep_alternative_taxids
#' @param verbose
#'
#' @export
read_obitools_taxonomy <- function(taxoname,
                                   keep_alternative_names = FALSE,
                                   keep_alternative_taxids = FALSE,
                                   verbose = is_robi_verbose()) {
  if (verbose)
    message("Reading the taxonomy tree... ",appendLF = FALSE)

  data <- .Call("R_raw_to_ecopcr_tnodes",
                read_file_raw(paste0(taxoname,".tdx")))

  names(data) <- c("taxid","parent_idx","rank")
  data$taxid <- new_robitaxid_master(data$taxid)

  if (verbose)
    message(sprintf("%d taxa read.",length(data[[1]])))

  data$parent_idx <- as.integer(data$parent_idx + 1)

  if (verbose)
    message("Reading the taxonomy ranks... ",appendLF = FALSE)

  ranks <- .read_ecopcr_rank(taxoname)
  data$rank <- structure(as.integer(data$rank + 1),
                             levels = ranks,
                             class = "factor")

  if (verbose)
    message(sprintf("%d ranks read.",length(ranks)))

  if (verbose)
    message("Reading taxa names... ",appendLF = FALSE)

  taxnames <- .read_ecopcr_name(taxoname)

  if (verbose)
    message(sprintf("%d names read.",nrow(taxnames)))

  if (verbose)
    message("Indexing scientific names... ",appendLF = FALSE)

  taxnames %>%
    dplyr::filter(name_class == "scientific name") %>%
    dplyr::arrange(taxid_idx) %>%
    dplyr::pull(name) -> data$scientific_name

  if (verbose)
    message("Done.")

  if (verbose)
    message("Indexing alternative names... ",appendLF = FALSE)

  taxnames %>%
    dplyr::mutate(taxid = new_robitaxid(data$taxid[taxnames$taxid_idx])) %>%
    dplyr::select(-taxid_idx) -> taxnames

  if (verbose)
    message("Done.")

  if (.has_doParallel && getDoParRegistered()) {
      `%dp%` <- `%dopar%`
    }
  else{
      `%dp%` <- `%do%`
    }

  if (verbose)
    message("Computing path... ",appendLF = FALSE)


    paths <- .Call("R_compute_taxonomic_path",
                   data$taxid,
                   data$rank,
                   data$parent_idx)

    paths <- list(path =  lapply(paths[[1]],new_robiobject,
                                 class = "robipath"),
                  rpath = paths[[2]])


  if (verbose)
    message("Done.")

  data <- as_tibble(data)

  if (verbose)
    message("Indexing parent taxids... ",appendLF = FALSE)

  data %>%
    dplyr::mutate(parent = new_robitaxid(data$taxid[parent_idx])) %>%
    dplyr::select(-parent_idx) -> data

  if (verbose)
    message("Done.")

  data$path <- paths[[1]]
  data$rpath <- paths[[2]]

  if (!keep_alternative_names)
    taxnames = NULL

  new_robitaxonomy(data,
                   alternative_names = taxnames,
                   reference_taxonomy = NULL)
}
