#' @include robidata.R
#' @include type_robitaxid.R
#' @include taxonomy_methods.R
#' @import R6
#' @import tibble
#' @import dplyr
#' @import readr
NULL

library(R6)

#' Validates that a taxonomy instance respects some constraints.
#'
#'
#' @param object
#'
#' @return
#'
#' @examples
#' s <- robisample(n = 20)
#' check_robisample(s)
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
validate_robitaxonomy <- function(object, verbose = TRUE) {
  UseMethod("validate_robitaxonomy", object)
}

#' @details The \code{robitaxonomy} version of this methode checks that
#' taxonomy declared as full verified that constaint.
#'
#' @rdname validate_robitaxonomy
#' @export
validate_robitaxonomy.robitaxonomy <- function(object, verbose = TRUE) {
  if (is_full_taxonomy(object)) {
    if (!is_full_taxonomy(object, full_check = TRUE)) {
      stop("The taxonomy is declared as full but doesn't respect that constraint",
        call. = FALSE
      )
    }
  }
  object
}

#' Test if an object belongs \code{robitaxonomy} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robitaxonomy <- function(object) {
  inherits(object, "robitaxonomy")
}



.robitaxonomy <- R6Class("robitaxonomy",
  public = list(
    nodes = NULL,
    alternative_names = NULL,
    alternative_taxids = NULL,
    reference_taxonomy = NULL,

    initialize = function(nodes,
                          alternative_names,
                          alternative_taxids,
                          reference_taxonomy) {
      robiassert_arg(
        !missing(nodes), "nodes",
        "the node parametter is mandatory"
      )

      robiassert_arg(is_tibble(nodes) || is.null(nodes),
        "nodes",
        "Must be a tibble not a {wrongclass}.",
        wrongclass = class(node)[1]
      )

      self$nodes <- nodes

      if (!missing(alternative_names)) {
        robiassert_arg(is_tibble(alternative_names) ||
          is.null(alternative_names),
        "alternative_names",
        "Must be a tibble not a {wrongclass}.",
        wrongclass = class(alternative_names)[1]
        )

        self$alternative_names <- alternative_names
      }

      if (!missing(alternative_taxids)) {
        robiassert_arg(is_tibble(alternative_taxids) ||
          is.null(alternative_taxids),
        "alternative_taxids",
        "Must be a tibble not a {wrongclass}.",
        wrongclass = class(alternative_taxids)[1]
        )

        self$alternative_taxids <- alternative_taxids
      }

      if (!missing(reference_taxonomy)) {
        robiassert_arg(is_robitaxonomy(reference_taxonomy) ||
          is.null(reference_taxonomy),
        "reference_taxonomy",
        "Must be a robitaxonomy not a {wrongclass}.",
        wrongclass = class(alternative_names)[1]
        )

        self$reference_taxonomy <- reference_taxonomy
      }

      validate_robitaxonomy(self)
    },

    within = function(expression) {
      eval(
        expr = substitute(expression, environment()),
        envir = self,
        enclos = environment()
      )
    }
  )
)

#' Title
#'
#' @param object
#' @param alternative_names
#' @param reference_taxonomy
#' @param ...
#' @param class
#'
#' @return
#' @export
#'
#' @examples
new_robitaxonomy <- function(object,
                             alternative_names = NULL,
                             alternative_taxids = NULL,
                             reference_taxonomy = NULL,
                             ...) {
  .robitaxonomy$new(
    nodes = object,
    alternative_names = alternative_names,
    alternative_taxids = alternative_taxids,
    reference_taxonomy = reference_taxonomy
  )
}



#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
as_robitaxonomy <- function(object, ...) {
  UseMethod("as_robitaxonomy", object)
}

#' @export
as_robitaxonomy.robitaxid <- function(object, taxonomy) {
  new_robitaxonomy(taxonomy$nodes %>% filter(taxid %in% object),
    alternative_names = if (!is.null(taxonomy$alternative_names)) {
      taxonomy$alternative_names %>% filter(taxid %in% object)
    } else {
      NULL
    },
    reference_taxonomy = reference_taxonomy(taxonomy)
  )
}



#' @rdname scientific_name
#' @export
scientific_name.robitaxid <- function(data, taxonomy, results_in) {
  if (missing(results_in)) {
    results_in <- "scientific_name"
  }

  tibble(taxid = as_robitaxid_master(data)) %>%
    left_join(nodes(taxonomy) %>%
      select(taxid, !!results_in := scientific_name),
    by = "taxid"
    )
}

#' @rdname scientific_name
#' @export
scientific_name.numeric <- function(data, taxonomy, results_in) {
  fcall <- match.call()
  fcall[[1]] <- quote(scientific_name)
  fcall[[2]] <- quote(as_robitaxid_master(data))
  eval(fcall)
}

#' @rdname scientific_name
#' @export
scientific_name.tbl <- function(data, taxonomy, results_in, slot) {
  if (missing(results_in)) {
    results_in <- "scientific_name"
  }

  if (missing(slot)) {
    slot <- master_taxids_colname(data)
  }

  bys <- "taxid"
  names(bys) <- slot

  suppressWarnings(
    data %>% left_join(data %>%
      pull(slot) %>%
      unique() %>% na.omit() %>%
      as_robitaxid_master() %>%
      scientific_name(taxonomy, results_in), by = bys)
  )
}

#' @rdname scientific_name
#' @export
scientific_name.robimotu <- function(data, taxonomy, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(scientific_name)
  fcall[[2]] <- quote(as_tibble(data))
  eval(fcall) %>%
    as_robimotu()
}

#' @rdname scientific_name
#' @export
scientific_name.robimetabar <- function(data, taxonomy, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(scientific_name)
  fcall[[2]] <- quote(motus(data))
  eval(fcall)
}

#' @rdname taxonomic_rank
#' @export
taxonomic_rank.robitaxid <- function(data, taxonomy, results_in) {
  if (missing(results_in)) {
    results_in <- "rank"
  }

  tibble(taxid = as_robitaxid_master(data)) %>%
    left_join(nodes(taxonomy) %>%
      select(taxid, !!results_in := rank),
    by = "taxid"
    )
}

#' @rdname taxonomic_rank
#' @export
taxonomic_rank.numeric <- function(data, taxonomy, results_in) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxonomic_rank)
  fcall[[2]] <- quote(as_robitaxid_master(data))
  eval(fcall)
}

#' @rdname taxonomic_rank
#' @export
taxonomic_rank.tbl <- function(data, taxonomy, results_in, slot) {
  if (missing(results_in)) {
    results_in <- "rank"
  }

  if (missing(slot)) {
    slot <- master_taxids_colname(data)
  }

  bys <- "taxid"
  names(bys) <- slot

  suppressWarnings(
    data %>% left_join(data %>%
      pull(slot) %>%
      unique() %>% na.omit() %>%
      as_robitaxid_master() %>%
      taxonomic_rank(taxonomy, results_in), by = bys)
  )
}

#' @rdname taxonomic_rank
#' @export
taxonomic_rank.robimotu <- function(data, taxonomy, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxonomic_rank)
  fcall[[2]] <- quote(as_tibble(data))
  eval(fcall) %>%
    as_robimotu()
}


#' @rdname taxonomic_rank
#' @export
taxonomic_rank.robimetabar <- function(data, taxonomy, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxonomic_rank)
  fcall[[2]] <- quote(motus(data))
  eval(fcall)
}

#' @rdname taxonomic_parent
#' @export
taxonomic_parent.robitaxid <- function(data, taxonomy, with_name = TRUE, results_in) {
  reference <- reference_taxonomy(taxonomy)
  if (missing(results_in)) {
    results_in <- "parent"
  }

  bys <- "taxid"
  names(bys) <- results_in

  tibble(taxid = as_robitaxid_master(data)) %>%
    left_join(nodes(taxonomy) %>%
      select(taxid, !!results_in := parent),
    by = "taxid"
    ) -> results

  if (with_name) {
    results %>% left_join(results %>%
      pull(results_in) %>%
      unique() %>% na.omit() %>%
      as_robitaxid_master() %>%
      scientific_name(reference,
        results_in = paste0(results_in, "_name")
      ),
    by = bys
    ) -> results
  }

  results
}

#' @rdname taxonomic_parent
#' @export
taxonomic_parent.numeric <- function(data, taxonomy, with_name = TRUE, results_in) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxonomic_parent)
  fcall[[2]] <- quote(as_robitaxid_master(data))
  eval(fcall)
}

#' @rdname taxonomic_parent
#' @export
taxonomic_parent.tbl <- function(data, taxonomy, with_name = TRUE, results_in, slot) {
  if (missing(results_in)) {
    results_in <- "parent"
  }

  if (missing(slot)) {
    slot <- master_taxids_colname(data)
  }

  bys <- "taxid"
  names(bys) <- slot

  suppressWarnings(
    data %>% left_join(data %>%
      pull(slot) %>%
      unique() %>% na.omit() %>%
      as_robitaxid_master() %>%
      taxonomic_parent(taxonomy, with_name, results_in), by = bys)
  )
}

#' @rdname taxonomic_parent
#' @export
taxonomic_parent.robimotu <- function(data, taxonomy, with_name = TRUE, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxonomic_parent)
  fcall[[2]] <- quote(as_tibble(data))
  eval(fcall) %>%
    as_robimotu()
}

#' @rdname taxonomic_parent
#' @export
taxonomic_parent.robimetabar <- function(data, taxonomy, with_name = TRUE, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxonomic_parent)
  fcall[[2]] <- quote(motus(data))
  eval(fcall)
}



#' @rdname taxon_at_rank
#' @export
taxon_at_rank.robitaxid <- function(data, taxonomy, rank, with_name = FALSE, results_in) {
  reference <- reference_taxonomy(taxonomy)

  ranks <- levels(nodes(reference)$rank)
  rankid <- match(rank, ranks, nomatch = 0)

  robiassert_arg(rankid > 0,
    "rank",
    "The specified rank {rank} doesn't correspond to any known rank",
    rank = rank
  )

  if (missing(results_in)) {
    results_in <- rank
  }

  bys <- "taxid"
  names(bys) <- results_in



  tibble(taxid = as_robitaxid_master(data)) %>%
    left_join(nodes(taxonomy) %>%
      select(taxid, trank = rank, path, rpath),
    by = "taxid"
    ) -> results

  results[[results_in]] <- vapply(
    seq_along(results$taxid),
    function(i) {
      p <- results$path[[i]]
      if (results$trank[[i]] == rank) {
        return(results$taxid[[i]])
      }
      if (length(p > 0)) {
        p[match(
          rankid,
          results$rpath[[i]],
          NA
        )][1]
      } else {
        NA
      }
    }, 0
  ) %>% as_robitaxid()

  if (with_name) {
    suppressWarnings(
      results %>% left_join(results %>%
        pull(results_in) %>%
        unique() %>% na.omit() %>%
        as_robitaxid_master() %>%
        scientific_name(reference,
          results_in = paste0(results_in, "_name")
        ),
      by = bys
      )
    ) -> results
  }

  results %>% select(-path, -rpath, -trank)
}

#' @rdname taxon_at_rank
#' @export
taxon_at_rank.numeric <- function(data, taxonomy, rank, with_name = FALSE, results_in) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxon_at_rank)
  fcall[[2]] <- quote(as_robitaxid_master(data))
  eval(fcall)
}


#' @rdname taxonomic_parent
#' @export
taxon_at_rank.tbl <- function(data, taxonomy, rank, with_name = FALSE, results_in, slot) {
  reference <- reference_taxonomy(taxonomy)
  ranks <- levels(nodes(reference)$rank)
  rankid <- match(rank, ranks, nomatch = 0)

  robiassert_arg(rankid > 0,
    "rank",
    "The specified rank {rank} doesn't correspond to any known rank",
    rank = rank
  )

  if (missing(results_in)) {
    results_in <- rank
  }

  if (missing(slot)) {
    slot <- master_taxids_colname(data)
  }

  bys <- "taxid"
  names(bys) <- slot

  suppressWarnings(
    data %>% left_join(data %>%
      pull(slot) %>%
      unique() %>% na.omit() %>%
      as_robitaxid_master() %>%
      taxon_at_rank(taxonomy, rank, with_name, results_in), by = bys)
  )
}

#' @rdname taxonomic_parent
#' @export
taxon_at_rank.robimotu <- function(data, taxonomy, rank, with_name = FALSE, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxon_at_rank)
  fcall[[2]] <- quote(as_tibble(data))
  eval(fcall) %>%
    as_robimotu()
}

#' @rdname taxonomic_parent
#' @export
taxon_at_rank.robimetabar <- function(data, taxonomy, rank, with_name = FALSE, results_in, slot) {
  fcall <- match.call()
  fcall[[1]] <- quote(taxon_at_rank)
  fcall[[2]] <- quote(motus(data))
  eval(fcall)
}



#' @rdname as_robitaxid
#' @export
as_robitaxid.robitaxonomy <- function(x) {
  taxids <- nodes(x)$taxid
  names(taxids) <- nodes(x)$scientific_name
  as_robitaxid(taxids)
}

#' @rdname as_robitaxid
#' @export
as_robitaxid_master.robitaxonomy <- function(x) {
  taxids <- nodes(x)$taxid
  names(taxids) <- nodes(x)$scientific_name
  taxids
}

#' @export
parent.robitaxonomy <- function(taxonomy) {
  reference <- reference_taxonomy(taxonomy)
  suppressWarnings(nodes(taxonomy) %>%
    dplyr::left_join(reference,
      by = c(parent = "taxid")
    )) %>%
    dplyr::mutate(
      son = new_robitaxid(taxid),
      taxid = new_robitaxid_master(parent),
      parent = parent.y,
      scientific_name = scientific_name.y,
      rank = rank.y,
      path = path.y
    ) %>%
    select(taxid, parent, rank, path, son) -> candidates

  alternative <- alternative_names(taxonomy)
  if (!is.null(alternative)) {
    alternative %>%
      dplyr::filter(taxid %in% candidates$taxid) -> alternative
  }


  new_robitaxonomy(candidates,
    alternative_names = alternative,
    reference_taxonomy = reference
  )
}

#' @export
max.robitaxonomy <- function(x, ..., na.rm = FALSE) {
  taxos <- lapply(c(list(x), list(...)), nodes)
  foreach(
    i = seq_along(taxos),
    .combine = rbind
  ) %do%
    {
      taxos[[i]] %>%
        dplyr::filter(taxid == max(taxid, na.rm = na.rm))
    } %>%
    dplyr::filter(taxid == max(taxid, na.rm = na.rm)) -> candidates

  suppressWarnings(candidates %>%
    dplyr::distinct_(.dots = names(.))) -> candidates

  do.call(rbind, lapply(taxos, alternative_names)) -> alternative


  if (!is.null(alternative)) {
    suppressWarnings(
      dplyr::filter(taxid %in% candidates$taxid) %>%
        dplyr::distinct_(.dots = names(.))
    ) %>%
      arrange(name) -> alternative
  }

  new_robitaxonomy(candidates,
    alternative_names = alternative,
    reference_taxonomy = reference_taxonomy(x)
  )
}

#' @export
min.robitaxonomy <- function(x, ..., na.rm = FALSE) {
  taxos <- lapply(c(list(x), list(...)), nodes)
  foreach(
    i = seq_along(taxos),
    .combine = rbind
  ) %do%
    {
      taxos[[i]] %>%
        dplyr::filter(taxid == min(taxid, na.rm = na.rm))
    } %>%
    dplyr::filter(taxid == min(taxid, na.rm = na.rm)) -> candidates

  suppressWarnings(candidates %>%
    dplyr::distinct_(.dots = names(.))) -> candidates

  do.call(rbind, lapply(taxos, alternative_names)) -> alternative


  if (!is.null(alternative)) {
    suppressWarnings(
      dplyr::filter(taxid %in% candidates$taxid) %>%
        dplyr::distinct_(.dots = names(.))
    ) %>%
      arrange(name) -> alternative
  }

  new_robitaxonomy(candidates,
    alternative_names = alternative,
    reference_taxonomy = reference_taxonomy(x)
  )
}
#' @export
spread_names.robitaxonomy <- function(taxonomy) {
  suppressWarnings(nodes(taxonomy) %>%
    dplyr::left_join(alternative_names(taxonomy),
      by = "taxid"
    )) %>%
    dplyr::spread(key = name_class, value = name) -> candidates

  new_robitaxonomy(candidates,
    alternative_names = alternative_names(taxonomy),
    reference_taxonomy = reference_taxonomy(taxonomy)
  )
}

#' @rdname lowest_common_ancestor
#' @export
lowest_common_ancestor.robitaxonomy <- function(taxonomy,
                                                threshold = 1.0,
                                                error = 0,
                                                na.rm = TRUE,
                                                only_lca = FALSE) {
  if (threshold < 1) {
    error <- ceiling(nrow(taxonomy) * (1. - threshold))
  }

  paths <- nodes(taxonomy)$path
  lens <- vapply(paths, FUN = length, FUN.VALUE = 0)
  lmax <- max(lens)
  paths <- mapply(function(p, l, m) p[(l - m + 1):l], paths, lens, lmin)

  errors <- apply(paths,
    MARGIN = 1,
    function(i) length(i) - max(table(i[i > 0]))
  )

  most <- as.integer(apply(paths,
    MARGIN = 1,
    function(i) names(table(i[i > 0]))[which.max(table(i[i > 0]))]
  ))


  rep <- most[errors <= error]

  if (only_lca) rep <- rep[1]

  new_robilca(rep,
    reference_taxonomy = reference_taxonomy(taxonomy),
    errors = errors[errors <= error],
    ntaxa = nrow(taxonomy)
  )
}


# lowest_common_ancestor.robitaxonomy <- function(taxonomy,
#                                 threshold = 1.0,
#                                 error = 0,
#                                 na.rm = TRUE,
#                                 only_lca = FALSE) {
#
#   if (threshold < 1)
#     error <- ceiling(nrow(taxonomy) * (1. - threshold))
#
#   paths <- nodes(taxonomy)$path
#   lmax  <- max(vapply(paths,FUN = length,FUN.VALUE = 0))
#   paths <- sapply(paths, function(x) c(rep(0, lmax - length(x)), x))
#
#   errors <- apply(paths,
#                   MARGIN = 1,
#                   function(i) length(i) - max(table(i[i > 0])))
#
#   most <- as.integer(apply(paths,
#                            MARGIN = 1,
#                            function(i) names(table(i[i > 0]))[which.max(table(i[i > 0]))]))
#
#
#   rep <- most[errors <= error]
#
#   if (only_lca) rep <- rep[1]
#
#   new_robilca(rep,
#                reference_taxonomy = reference_taxonomy(taxonomy),
#                errors = errors[errors <= error],
#                ntaxa = nrow(taxonomy))
# }



#' @rdname alternative_names
#' @export
alternative_names.robitaxonomy <- function(taxonomy) {
  taxonomy$alternative_names
}

#' @rdname alternative_names
#' @export
alternative_taxids.robitaxonomy <- function(taxonomy) {
  taxonomy$alternative_taxids
}

#' @rdname nodes
#' @export
nodes.robitaxonomy <- function(taxonomy) {
  taxonomy$nodes
}


#' @export
dim.robitaxonomy <- function(x) dim(nodes(x))

#' @rdname is_full_taxonomy
#' @export
is_full_taxonomy.robitaxonomy <- function(taxonomy, full_check = FALSE) {
  if (full_check) {
    all(unique(unlist(nodes(taxonomy)$path,
      recursive = TRUE
    )) %in% nodes(taxonomy)$taxid)
  } else {
    is.null(taxonomy$reference_taxonomy)
  }
}

#' @rdname full_taxonomy
#' @export
full_taxonomy.robitaxonomy <- function(taxonomy) {
  new_robitaxid_master(unique(unlist(list(
    nodes(taxonomy)$taxid,
    nodes(taxonomy)$path
  ),
  recursive = TRUE
  ))) %>%
    tibble(taxid = .) %>%
    left_join(reference_taxonomy(taxonomy) %>% nodes(),
      by = "taxid"
    ) -> candidates

  alternative <- alternative_names(taxonomy)
  if (!is.null(alternative)) {
    alternative %>%
      dplyr::filter(taxid %in% candidates$taxid) -> alternative
  }


  new_robitaxonomy(candidates,
    alternative_names = alternative,
    reference_taxonomy = NULL
  )
}


#' Saves a robitaxonomy instance.
#'
#' Save a robitaxonomy instance into a file on disque. Only full taxonomy
#' can be saved. If a partial taxonomy is passed as argument, it will be
#' enventally completed before saving
#'
#' @param taxonomy the `robitaxonomy` object to save.
#' @param file      a (writable binary-mode) connection or the name of the file
#'   where the data will be saved (when tilde expansion is done).
#' @param only_full a logical value indicating if only full taxonomy can be
#'   saved. If set to `FALSE` and a partial taxonomy is saved, it is before
#'   completed using the `full_taxonomy()` function.
#' @param verbose   if `TRUE` message can be printed out during the function
#'   execution in interactive mode.
#'
#' @return the provided taxonomy in invisible mode.
#'
#' @md
#' @seealso `full_taxonomy()`
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
write_robitaxonomy <- function(taxonomy, file,
                               only_full = FALSE,
                               verbose = is_robi_verbose()) {
  if (! is_full_taxonomy(taxonomy) && verbose) {
      message("Partial taxonomy was completed before writing...")
  }

  save(full_taxonomy(taxonomy), file = file)

  invisible(taxonomy)
}

#' @export
read_robitaxonomy <- function(file) {
  get(load(file))
}
