#' @importFrom  vctrs new_vctr vec_c
#' @importFrom  rlang is_atomic
#' @include type_robiobject.R
NULL


#' Robiatomic vectors.
#'
#' `robiatomic` is an abstract class in the ROBITools2 package. It is the ancestor
#' of every vector values. It is relying on the [`vctrs`] library. and most
#' of the `robiatomic` subclasses are also subclasses of
#' [`vctrs::vctrs_vctr`]. It is notably the super class of :
#'
#' * [`robicategory`]
#' * [`robitag`]
#' * [`robitaxid`]
#' * [`robiuniqueid`]
#'
#' @details the `robiatomic()` constructor just raises an error.
#'
#' @param length A non-negative integer specifying the desired length. Double
#'   values will be coerced to integer: supplying an argument of length other
#'   than one is an error.
#'
#' @return the `robiatomic()` constructor returns nothing but raises an error.
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
#' @md
#'
robiatomic <- function(length=0) {
  abort("robiatomic is a super class, no instance of that class can be created.")
}



#' @param object the object used as base to build the `robitatomic` instance
#'
#' @param ... Passed on to [`base::structure()`]. It allows for
#'   defining attributes, specified in tag = value form, which will be attached
#'   to data.
#' @param class Subclasses to assign to the new object. For the `robiatomic`
#'   constructor, `class`must be specied as `robiatomic` is an abstract class.
#'
#' @return `new_robiatomic` returns a new instance of `robiatomic`.
#'
#' @rdname robiatomic
#' @export
#'
#' @examples
#'
#' xd <- new_robiatomic(1:10,class="robifoo")
#'
#' @md
new_robiatomic <- function(object,...,class) {

  robiassert_arg(!missing(class),
                 "class",
                 paste("robiatomic is a super class,",
                        "no instance of that class can be created.",
                        "The class attribute cannot be empty"
                      ))

  vc <- new_vctr(object,
                 class = c(class, "robiatomic","robiobject"))

  validate_robiobject(vc)

}

#' @details The [`robiatomic`] version of that method
#'     ckecks that the instance is an [`atomic`][rlang::is_atomic] object.
#'
#' @export
#' @rdname validate_robiobject
#'
#' @examples
#' xd <- new_robiatomic(1:10,class="robifoo")
#' validate_robiobject(xd)
#'
#' @md
validate_robiobject.robiatomic <- function(object, verbose=is_robi_verbose()) {
  robiassert_arg(is_atomic(object),
                 "object",
                 "To be upgraded to the class robiatomic objects have to be atomic")
  object
}

#' @details `is_robiatomic` tests if an object belongs the `robiatomic` class
#'
#' @param `object` the object to test
#' @return a [`logical`] value
#'
#' @examples
#' is_robiatomic(xd)
#' is_robiatomic(1:10)
#'
#' @rdname robiatomic
#' @export
#' @md
is_robiatomic <- function(object) {
  inherits(object, "robiatomic")
}

#' @details `as_robiatomic` throw an error to prevent creation of an abstract
#'   class instance.
#'
#' @rdname robiatomic
#' @export
#' @md
as_robiatomic <- function(object) {
  abort("robiatomic is an abstract class, no pure instance can be created")
}



