#' @include utils.R
NULL


#' Validates a new object
#'
#' Methods \code{validate_robiobject} chek that all the
#' invariants on an object are respected. If not
#' an error is raised, otherwise the object is returned
#'
#' @param object the object to validate
#' @param verbose a logical indicating if some messages
#'        have to be printed during the checks.
#'
#' @return the input object
#' @export
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#'
validate_robiobject <- function(object, verbose=is_robi_verbose()) {
  if (length(object) == 0)
  return(object)

  UseMethod("validate_robiobject",object)
}

#' @details The default version only returns the
#'     checked object without any test
#' @export
#'
#' @rdname validate_robiobject
#' @examples
#'    x <- new_robiuniqid(c("A","B","C"))
#'    validate_robiobject(x)
validate_robiobject.default <- function(object, verbose=is_robi_verbose()) {
  object
}


#' Tests if a robiobject is valid.
#'
#' The function returns `TRUE` if the object is valid `FALSE` otherwise.
#'
#' @param object the instance to test
#'
#' @return a logical value
#' @export
#'
#' @md
#' @rdname validate_robiobject
#' @examples
#'    x <- new_robiuniqid(c("A","B","C"))
#'    is_a_valid_robiobject(x)
#'
is_a_valid_robiobject <- function(object) {
  suppressWarnings(
    tryCatch({validate_robiobject(object); TRUE},
             error = function(condition) FALSE
             )
  )
}

#' Tests if an object belongs \code{robiobject} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#' is_robiobject(3)
#' o <- 1:10
#' a <- new_robiobject(a)
#' is_robiobject(a)
#' @seealso \code{\link[ROBITools2]{new_robiobject}},
#'          \code{\link[ROBITools2]{unclass_robiobject}}
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robiobject <- function(object) {
  inherits(object, "robiobject")
}

#' Makes an object a member of the \code{robiobject} class.
#'
#' @param object the object to modify
#' @param ... Passed on to \code{\link[base]{structure}}()
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#' o <- 1:10
#' a <- new_robiobject(a)
#' is_robiobject(a)
#' @seealso \code{\link[ROBITools2]{unclass_robiobject}},
#'          \code{\link[ROBITools2]{is_robiobject}}
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
new_robiobject <- function(object, ..., class = character()) {
  if (is_robiobject(object)) {
    object <- unclass_robiobject(object)
  }

  structure(
    object,
    ...,
    class = c(class, "robiobject", class(object))
  )
}

#' Removes to an object its association to the \code{robiobject} class hierarchy.
#'
#' @param object the object to modify
#'
#' @examples
#' o <- 1:10
#' a <- new_robiobject(a)
#' is_robiobject(a)
#' a <- unclass_robiobject(a)
#' is_robiobject(a)
#' @seealso \code{\link[ROBITools2]{new_robiobject}},
#'          \code{\link[ROBITools2]{is_robiobject}}
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
unclass_robiobject <- function(object) {
  if (is_robiobject(object)) {
    classes <- class(object)
    nclasses <- length(classes)
    rclass <- which("robiobject" == classes) + 1
    class(object) <- classes[rclass:nclasses]
  }

  object
}

#' Converts an object to a \code{as_robiobject} instance
#'
#' @param object the object to convert
#' @return a \code{robiuniqueid} instance
#'
#' @examples
#' x <- as_robiobject(c("A", "B", "C"))
#' class(x)
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
as_robiobject <- function(object) {
  UseMethod("as_robiobject", object)
}

#' @rdname as_robiobject
#' @export
as_robiobject.robiobject <- function(object) {
  classes <- class(object)
  nclasses <- length(classes)
  rclass <- which("robiobject" == classes) + 1
  class(object) <- classes[rclass:nclasses]
  object
}

#' @rdname as_robiobject
#' @export
as_robiobject.default <- function(object) {
  new_robiobject(object)
}


