#' @include type_robiatomic.R
#' @importFrom  vctrs vec_data vec_cast vec_default_ptype2 vec_default_cast new_factor
#' @importFrom glue glue
#' @import stringr
NULL

#' Vector of tags for multiplexing PCR.
#'
#' A `robitag` is character vector that is for storing short oligo nucleotides
#' used to tag sample. Tags are always represented as uppercase letters and
#' contain only A, C, G, and T and every tag
#'
#' @details The `robitag()` constructor builds a default `robitag` instance of
#'   length defined by the `length` argument and composed of tags of length
#'   zero.
#'
#' @param length A non-negative integer specifying the desired length. Double
#'   values will be coerced to integer: supplying an argument of length other
#'   than one is an error.
#'
#' @return the `robitag()`and `new_robitag()` functions return a `robitag`
#'   instance.
#'
#' @examples
#'
#' rt <- robitag(10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
#' @md
robitag <- function(length = 0) {
  new_robitag(character(length))
}


#' @details The contructor `new_robitag` is the main function to create a
#'   `robitag` instance. For simpler conversion to that class, it can be easier
#'   to use the `as_robitag()` utility function.
#'
#' @param object a character vector or something coercible.
#' @param adapte_length if `TRUE` longuest tags are shortened by trimming them
#'   on the left side.
#' @param ... Passed on to [`base::structure()`]. It allows for defining
#'   attributes, specified in tag = value form, which will be attached to data
#'   (see [`robiatomic`]).
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   constructor takes some decision because of missing values and estimates
#'   default values these missing data.
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#'
#' rt <- new_robitag(c("ATC", "ATG", "CCC", "CTG"))
#'
#' @rdname robitag
#' @export
#' @md
new_robitag <- function(object, adapte_length=TRUE,
                        ...,
                        verbose = is_robi_verbose(),
                        class = character()) {

  object <- toupper(vec_cast(object,character()))

  lx <- nchar(object)
  lx_min <- if (length(object) == 0) 0 else min(lx, na.rm = TRUE)

  if (lx_min > 0 && any(lx != lx_min  & !is.na(object))) {
    if (adapte_length) {
      if (verbose)
        warning(sprintf("Every tags have not the same length. They will be all shortened to %dbp",
                        lx_min))

      object <- str_sub(object,-lx_min,-1)
    }
  }

  new_robiatomic(as.character(object),
      ...,
      class = c(class, "robitag","character")
    )

}

#' @details The [`robitag`] version of that method ckecks that :
#'
#' * All the tags have the same length
#' * Every tags are composed only of upper case A, C, G, or T
#'
#' @examples
#'
#' rt <- new_robitag(c("AGT","GGT","CGT","GTA","CCC"))
#'
#' @rdname validate_robiobject
#' @export
#' @md
validate_robiobject.robitag <- function(object) {

  object <- validate_robiobject.robiatomic(object)

  lx <- nchar(object)
  same_length <- all(duplicated(lx)[-1])

  if (!same_length) {
    strange = object[which(duplicated(lx)[-1]) + 1]
    if (length(strange) > 20)
      strange = c(strange[1:19], "and others...")
  }

  robiassert_arg(same_length,
                 "object",
                 "Every tags must have the same length ({nchar(object[1])}). ({paste(strange,collapse=', ')} seem different)")


  if (any(is.na(str_match(object,"^[ACGT]*$")) &&
          !is.na(object) &&
          object != "__NA__")) {
    stop("Only uppercase A, C, G, and T are allowed in tags.",
    call. = TRUE
    )
  }

  object
}


#' @@details `is_robitag` tests if an object belongs the `robitag` class
#'
#' @param `object` the object to test.
#'
#' @return the `is_robitag()`, `is_robitag_forward()`, and `is_robitag_reverse()`
#'    functions return a [`logical`] value.
#'
#' @examples
#'
#' is_robitag(rt)
#' is_robitag("toto")
#'
#' @rdname robitag
#' @export
#' @md
is_robitag = function(object) {
  inherits(object,'robitag')
}

#' @details `as_robitag` converts an object to a `robitag` instance.
#'
#' @param object the object to convert.
#'
#' @return the `as_robitag()` funtion returns a `robitag` value.
#'
#' @examples
#'
#' x <- c("AGT","GGT","CGT","GTA","CCC")
#' rt <- as_robitag(x)
#'
#' @rdname robitag
#' @export
#' @md
as_robitag <- function(object) {
  vec_cast(object,robitag())
}



#########################
#
# robitag_forward
#
#########################

#' @details `robitag_forward` is a specialisation of the class `robitag`. When
#'   PCR products are tagged on both extremities, that class allows for
#'   distinguising tags on the 5' end of the PCR product (`robitag_forward`)
#'   from those at the 3' end extremity (`reverse_tag`). Except that semantic
#'   distinction no difference exists with `robitag`.
#'
#' @return the `robitag_forward()`and `new_robitag_forward()` functions return a
#'   `robitag_forward` instance.
#'
#' @rdname robitag
#' @export
#' @md
robitag_forward <- function(length = 0) {
  new_robitag_forward(character(length))
}

#' @examples
#'
#' rf <- new_robitag_forward(c("AGT","GGT","CGT","GTA","CCC"))
#'
#' @rdname robitag
#' @export
#' @md
new_robitag_forward <- function(object, adapte_length=TRUE,
                        ..., verbose = is_robi_verbose(),
                        class = character())  {
  new_robitag(object = object,
              adapte_length = adapte_length,
      ...,
      verbose = verbose,
      class = c(class, "robitag_forward")
    )
}

#' @@details `is_robitag_formard` tests if an object belongs the `robitag_formard` class
#'
#' @examples
#'
#' is_robitag_forward(rf)
#' is_robitag_forward("toto")
#'
#' @rdname robitag
#' @export
#' @md
is_robitag_forward = function(object) {
  inherits(object,'robitag_forward')
}


#' @details `as_robitag` converts an object to a `robitag` instance.
#'
#' @param object the object to convert
#' @return the `as_robitag_forward()` funtion returns a `robitag_forward` value.
#'
#' @examples
#' x <- c("AA","CC","CG")
#' as_robitag_forward(x)
#'
#' @rdname robitag
#' @export
#' @md
as_robitag_forward = function(object) {
  vec_cast(object,robitag_forward())
}

#########################
#
# robitag_reverse
#
#########################

#' @details `robitag_reverse` is a specialisation of the class `robitag`.
#'    It plays the same role than `robitag_forward` but for the reverse
#'    tag located at the three prime end of the PCR product.
#'
#' @return the `robitag_reverse()`and `new_robitag_reverse()` functions
#'   return a `robitag_reverse` instance.
#'
#' @rdname robitag
#' @export
robitag_reverse <- function(length = 0) {
  new_robitag_reverse(character(length))
}

#' @examples
#'
#' rr <- new_robitag_reverse(c("AGT","GGT","CGT","GTA","CCC"))
#'
#' @rdname robitag
#' @export
new_robitag_reverse <- function(object, adapte_length=TRUE,
                        ..., verbose = is_robi_verbose(),
                        class = character())  {
  new_robitag(object=object,
              adapte_length = adapte_length,
      ...,
      verbose = verbose,
      class = c(class, "robitag_reverse")
    )
}

#' @details `is_robitag_reverse` tests if an object belongs the `is_robitag_reverse` class
#'
#' @examples
#'
#' is_robitag_reverse(rr)
#' is_robitag_reverse("toto")
#'
#' @rdname robitag
#' @export
#' @md
is_robitag_reverse = function(object) {
  inherits(object,'robitag_reverse')
}


#' @details `as_robitag` converts an object to a `robitag` instance.
#'
#' @param object the object to convert
#' @return the `as_robitag_reverse()` funtion returns a `robitag_reverse` value.
#'
#' @examples
#' x <- c("AA","CC","CG")
#' as_robitag_reverse(x)
#'
#' @rdname robitag
#' @export
#' @md
as_robitag_reverse = function(object) {
  vec_cast(object,robitag_reverse())
}


##
#
# Declaration of coersion rules when combining vectors
#

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag.robitag <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag.character <- function(x, y, ...) character()

#' @rdname robitag_casts
#' @export
vec_ptype2.character.robitag <- function(x, y, ...) character()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_forward.robitag_forward <- function(x, y, ...) robitag_forward()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_forward.character <- function(x, y, ...) character()

#' @rdname robitag_casts
#' @export
vec_ptype2.character.robitag_forward <- function(x, y, ...) character()


#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_forward.robitag <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag.robitag_forward <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_reverse.robitag_reverse <- function(x, y, ...) robitag_reverse()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_reverse.character <- function(x, y, ...) character()

#' @rdname robitag_casts
#' @export
vec_ptype2.character.robitag_reverse <- function(x, y, ...) character()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_reverse.robitag <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag.robitag_reverse <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_reverse.robitag <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag.robitag_reverse <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_reverse.robitag_forward <- function(x, y, ...) robitag()

#' @rdname robitag_casts
#' @export
vec_ptype2.robitag_forward.robitag_reverse <- function(x, y, ...) robitag()


#
#
# Casts for the robitag class
#
#

#' @rdname robitag_casts
#' @export
vec_cast.robitag.robitag <- function(x, to, ...) x

#' @rdname robitag_casts
#' @export
vec_cast.robitag.character <- function(x, to, ...)
  new_robitag(x)

#' @rdname robitag_casts
#' @export
vec_cast.character.robitag <- function(x, to, ...)
  vec_data(x)


#
#
# Casts for the robitag_forward class
#
#

#' @rdname robitag_casts
#' @export
vec_cast.robitag_forward.robitag_forward <- function(x, to, ...) x

#' @rdname robitag_casts
#' @export
vec_cast.robitag_forward.character <- function(x, to, ...)
  new_robitag_forward(x)

#' @rdname robitag_casts
#' @export
vec_cast.character.robitag_forward <- function(x, to, ...)
  vec_data(x)

#' @rdname robitag_casts
#' @export
vec_cast.robitag_forward.robitag <- function(x, to, ...)
  new_robitag_forward(vec_data(x))

#' @rdname robitag_casts
#' @export
vec_cast.robitag.robitag_forward <- function(x, to, ...)
  new_robitag(vec_data(x))





#
#
# Casts for the robitag_reverse class
#
#

#' @rdname robitag_casts
#' @export
vec_cast.robitag_reverse.robitag_reverse <- function(x, to, ...) x

#' @rdname robitag_casts
#' @export
vec_cast.robitag_reverse.character <- function(x, to, ...)
  new_robitag_reverse(x)

#' @rdname robitag_casts
#' @export
vec_cast.character.robitag_reverse <- function(x, to, ...)
  vec_data(x)

#' @rdname robitag_casts
#' @export
vec_cast.robitag_reverse.robitag <- function(x, to, ...)
  new_robitag_reverse(vec_data(x))

#' @rdname robitag_casts
#' @export
vec_cast.robitag.robitag_reverse <- function(x, to, ...)
  new_robitag(vec_data(x))

##
#
# Functions related to the printing of a robitag
#

#' @rdname robitag_casts
#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.robitag <- function(x, ...) {
  out <- format(x)
  out[is.na(x)] <- NA
  pillar::new_pillar_shaft_simple(out, align = "left")
}

#' @rdname robitag_casts
#' @export
vec_ptype_abbr.robitag <- function(x, ...) {
  "robitag"
}

#' @rdname robitag_casts
#' @export
vec_ptype_abbr.robitag_forward <- function(x, ...) {
  "robitag_forward"
}

#' @rdname robitag_casts
#' @export
vec_ptype_abbr.robitag_reverse <- function(x, ...) {
  "robitag_reverse"
}

#' @rdname robitag_casts
#' @importFrom pillar type_sum
#' @export
type_sum.robitag <- function(x) {
  glue("tag{<nchar(x[1])>}",.open = "<",.close = ">")
}

#' @rdname robitag_casts
#' @importFrom pillar type_sum
#' @export
type_sum.robitag_forward <- function(x) {
  glue("Ftag{<nchar(x[1])>}",.open = "<",.close = ">")
}

#' @rdname robitag_casts
#' @importFrom pillar type_sum
#' @export
type_sum.robitag_reverse <- function(x) {
  glue("Rtag{<nchar(x[1])>}",.open = "<",.close = ">")
}

