#' @import vctrs
#' @import rlang
#' @include type_robiatomic.R
#'
NULL

#' Build a \code{robitaxid} instance
#'
#' a \code{robitaxid} is an integer vector that is for storing numerical taxonomic
#' identifiers. They are usually used to store NCBI taxids. They have to be strictly
#' positive.
#'
#' @param length A non-negative integer specifying the desired length. Double
#'   values will be coerced to integer: supplying an argument of length other
#'   than one is an error.
#'
#' @return a \code{robitaxid} instance
#'
#' @examples
#'
#'   robitaxid(1:10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
robitaxid <- function(length = 0) {
  robiassert_arg(length(length)==1,arg = "length",
                 message = "The length of the length argument must be equal to 1.")

  new_robitaxid(rep(1,length))
}

#'
#' @param object an integer vector
#' @param reference
#' @param ... Passed on to [`base::structure()`]. It allows for defining
#'   attributes, specified in tag = value form, which will be attached to data
#'   (see [`robiatomic`]).
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   constructor takes some decision because of missing values and estimates
#'   default values these missing data.
#' @param class Subclasses to assign to the new object, default: none
#'
#' @examples
#' new_robitaxid(1:10)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @rdname robitaxid
#' @export
new_robitaxid <- function(object, reference = NULL,
                          ...,
                          verbose = is_robi_verbose(),
                          class = character()) {

    object <- vec_cast(object,integer())

    if (length(object) > 0 && !is.null(reference)) {

      robimessage("Cheking taxids against a reference taxonomy..")

      ok <- match(object,nodes(reference)$taxid)

      robiassert_arg(!any(are_na(ok)),
                     "reference",
                     ngettext(sum(are_na(ok)) > 1,
                              "The taxid ({object[are_na(ok)]}) is a wrong taobjectid and substituted by NA",
                              "The taxid ({paste(object[are_na(ok)],collapse=", ")}) are a wrong taobjectid and substituted by NA"
                             ),
                     .abort = warn)

      names(object) <- reference$scientific_name[ok]

      robimessage("done")
  }


  new_robiatomic(object,
      ...,
      class = c(class, "robitaxid","integer"))
}



#' Test if an object belongs \code{robitaxid} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#'   x <- robitaxid(1:10)
#'   is_robitaxid(x)
#'   is_robitaxid("toto")
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robitaxid = function(object) {
  inherits(object,'robitaxid')
}

#' @rdname validate_robiobject
#' @export
validate_robiobject.robitaxid <- function(object) {

  object <- validate_robiobject.robiatomic(object, verbose=is_robi_verbose())

  if (any(!are_na(object) & vec_cast(object,integer()) < 1)) {
    bad = unique(object[!is.na(object) && vec_cast(object,integer()) < 1])
    stop(ngettext(length(bad),
                  sprintf("The taxid %d is negative or null",
                          bad[1]),
                  sprintf("The taxids (%s) are negative or null",
                          paste(bad,collapse = ","))))
  }

  object
}


#' @details `as_robitaxid` converts an object to a `robitaxid` instance.
#'
#' @param object the object to convert.
#'
#' @return the `as_robitaxid()` funtion returns a `robitaxid` value.
#'
#' @examples
#'
#' x <- c(1,10,33)
#' rt <- as_robitaxid(x)
#'
#' @rdname robitaxid
#' @export
#' @md
as_robitaxid = function(object) {
  UseMethod("as_robitaxid", object)
}

#' @rdname as_robitaxid
#' @export
as_robitaxid.default = function(object) {
  vec_cast(object,robitaxid())
}

#########################
#
# robitaxid_master
#
#########################

#' @rdname robitaxid
#' @export
robitaxid_master <- function(length = 0) {
  robiassert_arg(length(length)==1,arg = "length",
                 message = "The length of the length argument must be equal to 1.")

  new_robitaxid_master(rep(1,length))
}

#' @rdname robitaxid
#' @export
new_robitaxid_master <- function(object, reference = NULL,
                          ...,
                          verbose = is_robi_verbose(),
                          class = character()) {
  new_robitaxid(object,
                reference = reference,
                ...,
                class = c(class, "robitaxid_master")
               )
}



#' @return the `as_robitaxid_master()` funtion returns a `robitaxid_master` value.
#'
#' @examples
#'
#' x <- c(1,10,33)
#' rt <- as_robitaxid_master(x)
#'
#' @rdname robitaxid
#' @export
#' @md
as_robitaxid_master = function(object) {
  UseMethod("as_robitaxid_master", object)
}

#' @rdname robitaxid
#' @export
as_robitaxid_master.default = function(object) {
  vec_cast(object,robitaxid_master())
}

#########################
#
# robipath
#
#########################

#' @rdname robitaxid
#' @export
robipath <- function(length = 0) {
  robiassert_arg(length(length)==1,arg = "length",
                 message = "The length of the length argument must be equal to 1.")

  new_robipath(rep(1,length))
}

#' @rdname robitaxid
#' @export
new_robipath <- function(object, reference = NULL,
                          ...,
                          verbose = is_robi_verbose(),
                          class = character()) {
  new_robitaxid(object,
                reference = reference,
                ...,
                class = c(class, "robipath")
               )
}

#' @rdname as_robitaxid
#' @export
as_robipath = function(object) {
  UseMethod("as_robipath", object)
}

#' @rdname as_robitaxid
#' @export
as_robipath.default = function(object) {
  vec_cast(object,robipath())
}

#########################
#
# robilca
#
#########################

#' @rdname robitaxid
#' @export
robilca <- function(length = 0) {
  robiassert_arg(length(length)==1,arg = "length",
                 message = "The length of the length argument must be equal to 1.")

  new_robilca(rep(1,length))
}

#' @rdname robitaxid
#' @export
new_robilca <- function(object, reference = NULL,
                          ...,
                          verbose = is_robi_verbose(),
                          class = character()) {
  new_robitaxid(object,
                reference = reference,
                ...,
                class = c(class, "robilca")
               )
}

#' @export
as_robilca = function(object) {
  UseMethod("as_robilca", object)
}

#' @rdname as_robitaxid
#' @export
as_robilca.default = function(object) {
  vec_cast(object,robilca())
}



##
#
# Declaration of coersion rules when combining vectors
#

#' @rdname robitaxid_casts
#' @export
vec_ptype2.robitaxid.robitaxid <- function(x, y, ...) robitaxid()

#' @rdname robitaxid_casts
#' @export
vec_ptype2.robitaxid.integer <- function(x, y, ...) integer()

#' @rdname robitaxid_casts
#' @export
vec_ptype2.integer.robitaxid <- function(x, y, ...) integer()

#' @rdname robitaxid_casts
#' @export
vec_ptype2.robitaxid.double <- function(x, y, ...) double()

#' @rdname robitaxid_casts
#' @export
vec_ptype2.double.robitaxid <- function(x, y, ...) double()


#' @export
vec_ptype2.robitaxid_master.robitaxid_master <- function(x, y, ...)  robitaxid_master()

#' @export
vec_ptype2.robitaxid_master.robitaxid <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robitaxid.robitaxid_master <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robitaxid_master.integer <- function(x, y, ...) integer()

#' @export
vec_ptype2.integer.robitaxid_master <- function(x, y, ...) integer()

#' @export
vec_ptype2.robitaxid_master.double <- function(x, y, ...) double()

#' @export
vec_ptype2.double.robitaxid_master <- function(x, y, ...) double()



#' @export
vec_ptype2.robipath.robipath <- function(x, y, ...)  robipath()

#' @export
vec_ptype2.robipath.robitaxid <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robitaxid.robipath <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robipath.robitaxid_master <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robitaxid_master.robipath <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robipath.integer <- function(x, y, ...) integer()

#' @export
vec_ptype2.integer.robipath <- function(x, y, ...) integer()

#' @export
vec_ptype2.robipath.double <- function(x, y, ...) double()

#' @export
vec_ptype2.double.robipath <- function(x, y, ...) double()




#' @export
vec_ptype2.robilca.robilca <- function(x, y, ...)  robilca()

#' @export
vec_ptype2.robilca.robitaxid <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robitaxid.robilca <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robilca.robitaxid_master <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robitaxid_master.robilca <- function(x, y, ...) robitaxid()

#' @export
vec_ptype2.robilca.integer <- function(x, y, ...) integer()

#' @export
vec_ptype2.integer.robilca <- function(x, y, ...) integer()

#' @export
vec_ptype2.robilca.double <- function(x, y, ...) double()

#' @export
vec_ptype2.double.robilca <- function(x, y, ...) double()


#
#
# Casts for the robitaxid class
#
#

#' @export
vec_cast.robitaxid.robitaxid <- function(x, to, ...) x

#' @export
vec_cast.robitaxid.integer <- function(x, to, ...)
  new_robitaxid(x)

#' @export
vec_cast.integer.robitaxid <- function(x, to, ...)
  vec_data(x)

#' @export
vec_cast.robitaxid.double <- function(x, to, ...)
  new_robitaxid(x)

#' @export
vec_cast.double.robitaxid <- function(x, to, ...)
  vec_cast(vec_data(x),double())

#' @export
vec_cast.robitaxid.character <- function(x, to, ...)
  new_robitaxid(vec_cast(x,integer()))

#' @export
vec_cast.character.robitaxid <- function(x, to, ...)
  as.character(vec_data(x))




#' @export
vec_cast.robitaxid_master.robitaxid_master <- function(x, to, ...) x

#' @export
vec_cast.robitaxid_master.robitaxid <- function(x, to, ...)
  new_robitaxid_master(x)

#' @export
vec_cast.robitaxid.robitaxid_master <- function(x, to, ...)
  new_robitaxid(vec_data(x))

#' @export
vec_cast.robitaxid_master.integer <- function(x, to, ...)
  new_robitaxid_master(x)

#' @export
vec_cast.integer.robitaxid_master <- function(x, to, ...)
  vec_data(x)

#' @export
vec_cast.robitaxid_master.double <- function(x, to, ...)
  new_robitaxid_master(x)

#' @export
vec_cast.double.robitaxid_master <- function(x, to, ...)
  vec_cast(vec_data(x),double())

#' @export
vec_cast.robitaxid_master.character <- function(x, to, ...)
  new_robitaxid_master(vec_cast(x,integer()))

#' @export
vec_cast.character.robitaxid_master <- function(x, to, ...)
  as.character(vec_data(x))




#' @export
vec_cast.robipath.robipath <- function(x, to, ...) x

#' @export
vec_cast.robipath.robitaxid <- function(x, to, ...)
  new_robipath(x)

#' @export
vec_cast.robitaxid.robipath <- function(x, to, ...)
  new_robitaxid(vec_data(x))

#' @export
vec_cast.robipath.integer <- function(x, to, ...)
  new_robipath(x)

#' @export
vec_cast.integer.robipath <- function(x, to, ...)
  vec_data(x)

#' @export
vec_cast.robipath.double <- function(x, to, ...)
  new_robipath(x)

#' @export
vec_cast.double.robipath <- function(x, to, ...)
  vec_cast(vec_data(x),double())

#' @export
vec_cast.robipath.character <- function(x, to, ...)
  new_robipath(vec_cast(x,integer()))

#' @export
vec_cast.character.robipath <- function(x, to, ...)
  as.character(vec_data(x))




#' @export
vec_cast.robilca.robilca <- function(x, to, ...) x

#' @export
vec_cast.robilca.robitaxid <- function(x, to, ...)
  new_robilca(x)

#' @export
vec_cast.robitaxid.robilca <- function(x, to, ...)
  new_robitaxid(vec_data(x))

#' @export
vec_cast.robilca.robitaxid_master <- function(x, to, ...)
  new_robilca(x)

#' @export
vec_cast.robitaxid_master.robilca <- function(x, to, ...)
  new_robitaxid_master(vec_data(x))

#' @export
vec_cast.robilca.integer <- function(x, to, ...)
  new_robilca(x)

#' @export
vec_cast.integer.robilca <- function(x, to, ...)
  vec_data(x)

#' @export
vec_cast.robilca.double <- function(x, to, ...)
  new_robilca(x)

#' @export
vec_cast.double.robilca <- function(x, to, ...)
  vec_cast(vec_data(x),double())

#' @export
vec_cast.robilca.character <- function(x, to, ...)
  new_robilca(vec_cast(x,integer()))

#' @export
vec_cast.character.robilca <- function(x, to, ...)
  as.character(vec_data(x))

# ------------------------------------------------------------------------------
# Printing


#' @export
format.robitaxid <- function(x, ...) {
  out <- formatC(vec_data(x),format = "d")
  out[is.na(x)] <- NA
  out[!is.na(x)] <- paste0("taxon:",out[!is.na(x)])
  out
}



#' @export
vec_ptype_abbr.robitaxid <- function(x, ...) {
  "taxid"
}

#' @export
vec_ptype_full.robitaxid <- function(x, ...) {
  "robitaxid"
}

#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.robitaxid <- function(x, ...) {
  out <- formatC(x)
  out[is.na(x)] <- NA
  pillar::new_pillar_shaft_simple(out, align = "right")
}

#' @export
vec_ptype_full.robitaxid_master <- function(x, ...) {
  "robitaxid_master"
}

#' @importFrom pillar type_sum
#' @export
vec_ptype_abbr.robitaxid_master <- function(x, ...) {
  "TAXID"
}


#' @importFrom pillar type_sum
#' @export
vec_ptype_abbr.robipath <- function(x, ...) {
  "path"
}



#' @importFrom pillar type_sum
#' @export
type_sum.robipath <- function(x) {
  sprintf("tpath[%d]",length(x))
}

#' @export
sort.robipath <- function(x, decreasing = FALSE, ...) {
  stop("robipath are not sortable")
}


#' @importFrom pillar type_sum
#' @export
vec_ptype_abbr.robilca <- function(x, ...) {
  "lca"
}



#' @importFrom pillar type_sum
#' @export
type_sum.robilca <- function(x) {
  sprintf("lca[%d]",x[1])
}


#' @rdname as_robitaxonomy
#' @export
as_robitaxonomy.robilca <- function(object) {
  ref <- reference_taxonomy(object)
  suppressWarnings(tibble(taxid=object) %>%
    left_join(nodes(ref),
              by = "taxid")) %>%
    mutate(taxid = new_robitaxid_master(taxid),
           error = attr(object,"errors"),
           ntaxa = attr(object,"ntaxa")) %>%
    select(taxid, rank,
           scientific_name,
           parent, path, error, ntaxa) -> candidates

  new_robitaxonomy(candidates,
                   reference_taxonomy = ref)
}




