#' @include type_robiatomic.R
#' @import vctrs
#' @import tibble
#' @import dplyr
NULL

#' Build a \code{robiuniqueid} instance
#'
#' a \code{robiuniqueid} is character vector that ensure that each value is
#' unique. It is used for storing sample or MOTU ids in
#' \code{\link[ROBITools2]{robisample}} and \code{\link[ROBITools2]{robimotus}}
#' instance
#'
#'
#' @param length A non-negative integer specifying the desired length. Double
#'   values will be coerced to integer: supplying an argument of length other
#'   than one is an error.
#'
#' @return a \code{robiuniqueid} instance
#'
#' @examples
#'
#' robiuniqueid(c("A", "A", "B", "C"))
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
robiuniqueid <- function(length=0) {
  new_robiuniqueid(character(length))
}

#' Creates a new instance of `robiuniqueid`.
#'
#' @param object the object to modify
#' @param make_unique If some values are not unique the unicity is forced by
#'   adding a suffix to the duplicated item.
#' @param sep a `character` separator to be added before the unicity suffix.
#' @param ... Passed on to \code{\link[base]{structure}}()
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   constructor takes some decision because of missing values and estimates
#'   default values these missing data.
#' @param class Subclasses to assign to the new object, default: none
#'
#'
#' @return a `robiuniqueid` instance.
#'
#' @examples
#' new_robiuniqueid(c("A", "A", "B", "C"))
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @md
#' @export
new_robiuniqueid <- function(object = character(0),
                             make_unique = TRUE,
                             sep = "_",
                             ...,
                             verbose = is_robi_verbose(),
                             class = character()) {

  object <- vec_cast(object,character())

  if (any(duplicated(object))) {
    if (make_unique) {
      robimessage("Several ids are not unique : {paste(object[duplicated(object)], collapse = ', ')}"
        )
      object <- make.unique(object, sep = sep)
    }
  }

  new_robiatomic(object,
      ...,
      class = c(class, "robiuniqueid","character")
    )
}



#' @details The \code{\link[ROBITools2]{robiuniqueid}} version
#'     ckecks that the instance contains no duplicated identifers.
#'
#' @export
#'
#' @rdname validate_robiobject
validate_robiobject.robiuniqueid <- function(object, verbose=is_robi_verbose()) {

  object <- validate_robiobject.robiatomic(object)

  robiassert_arg(!any(is.na(object)),
                 "object",
                 "robiuniqueid cannot contain NA value" )

  no_dup <- anyDuplicated(object) == 0
  errmsg <- ""
  if (!no_dup) {
    multi <- unique(object[duplicated(object)])
    if (length(multi) > 20)
      multi <- c(multi[1:19],"and more...")
    multi_txt <- paste(multi, collapse = ", ")

    errmsg <- paste0("All values in a robiuniqueid must be unique. ",
                        ngettext(
                          length(multi),
                          sprintf("(%s) is used several times", multi_txt),
                          sprintf("(%s) are used several times", multi_txt)
                        ))
               }

  robiassert_arg(anyDuplicated(object) == 0, "object", errmsg )

  object
}

#' Test if an object belongs \code{robiuniqueid} class
#'
#' @param object the object to test
#' @return a \code{logical} value
#'
#' @examples
#' x <- new_robiuniqueid(c("A", "B", "C"))
#' is_robiuniqueid(x)
#' is_robiuniqueid("toto")
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
is_robiuniqueid <- function(object) {
  inherits(object, "robiuniqueid")
}


#' Converts an object to a \code{robiuniqueid} instance
#'
#' @param object the object to convert
#' @return a \code{robiuniqueid} instance
#'
#' @examples
#'   x <- robiuniqueid(c("A","B","C"))
#'   as_robiuniqueid(x)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
as_robiuniqueid = function(object) {
  UseMethod("as_robiuniqueid", object)
}

#' @rdname as_robiuniqueid
#' @export
as_robiuniqueid.default = function(object) {
  vec_cast(object,robiuniqueid())
}

##
#
# Declaration of coersion rules when combining vectors
#


#' @export
vec_ptype2.robiuniqueid.robiuniqueid <- function(x, y, ...) robiuniqueid()

#' @export
vec_ptype2.robiuniqueid.character <- function(x, y, ...) character()

#' @export
vec_ptype2.character.robiuniqueid <- function(x, y, ...) character()


#
#
# Casts for the robiuniquei class
#
#

#' @export
vec_cast.robiuniqueid.robiuniqueid <- function(x, to, ...) x

#' @export
vec_cast.robiuniqueid.character <- function(x, to, ...)
  new_robiuniqueid(x)

#' @export
vec_cast.character.robiuniqueid <- function(x, to, ...)
  vec_data(x)

#' @export
vec_cast.robiuniqueid.numeric <- function(x, to, ...)
  new_robiuniqueid(as.character(x))

# ------------------------------------------------------------------------------
# Printing

#' @export
vec_ptype_abbr.robitaxid <- function(x, ...) {
  "robiuid"
}

#' @export
vec_ptype_full.robiuniqueid <- function(x, ...) {
  "robiuniqueid"
}

#' @importFrom pillar pillar_shaft
#' @export
pillar_shaft.robiuniqueid <- function(x, ...) {
  out <- format(x)
  out[is.na(x)] <- NA
  pillar::new_pillar_shaft_simple(out, align = "left")
}


#' @export
vec_restore.robiuniqueid <- function(x, to, ...) {
  if (anyDuplicated(na.omit(x)) > 0)
    stop("Duplicated id")
  NextMethod()
}
