#' @include ROBITools2.R
#'
#' @importFrom glue glue
#' @importFrom rlang abort warn inform
#'
#'
NULL

#
# https://stackoverflow.com/questions/11885207/get-all-parameters-as-list
#
.allargs <- function(orig_values = FALSE) {
  # get formals for parent function
  parent_formals <- formals(sys.function(sys.parent(n = 1)))

  # Get names of implied arguments
  fnames <- names(parent_formals)

  # Remove '...' from list of parameter names if it exists
  fnames <- fnames[-which(fnames == '...')]

  # Get currently set values for named variables in the parent frame
  args <- evalq(as.list(environment()), envir = parent.frame())

  # Get the list of variables defined in '...'
  args <- c(args[fnames], evalq(list(...), envir = parent.frame()))

  if(orig_values) {
    # get default values
    defargs <- as.list(parent_formals)
    defargs <- defargs[unlist(lapply(defargs, FUN = function(x) class(x) != "name"))]
    args[names(defargs)] <- defargs
    setargs <- evalq(as.list(match.call())[-1], envir = parent.frame())
    args[names(setargs)] <- setargs
  }
  return(args)
}



#' Builds a serie of unique words
#'
#' \code{n} words of size \code{nletters} are built as a combination of the
#' uppercase letters.
#'
#' @param n an integer value indicating the number of words to generate
#'
#' @param nletters an integer value indicating the number of letters for each
#'   words
#' @return a \code{character} vector of length \code{n}
#'
#' @examples
#'    combine_LETTERS(10)
#'    combine_LETTERS(30)
#'    combine_LETTERS(10,3)
#'    combine_LETTERS(30,3)
#'
#' @author Eric Coissac <eric.coissac@metabarcoding.org>
#' @export
combine_LETTERS <- function(n,
                           nletters = floor(log(n) / log(26)) + 1) {
  f <- function(x) {
    if (x == 1)
      return(LETTERS)

    l <- expand.grid(LETTERS,f(x - 1))
    l <- paste0(l[,2],l[,1])

    if (length(l) > n)
      l[1:n]
    else
      l
  }

  f(nletters)[1:n]
}

#' Sets the verbose status
#'
#' @param state a logical value indicating the new verbose mode
#'
#' @export
#'
#' @examples
#'
#' state <- getOption("ROBITools2.verbose")
#' is_robi_verbose()
#' set_robi_verbose(TRUE)
#' is_robi_verbose()
#' set_robi_verbose(FALSE)
#' is_robi_verbose()
#' set_robi_verbose(state)
#'
set_robi_verbose <- function(state) {
  robiassert_arg(is.logical(state),
                 arg = "state",
                 message = "value must be a logical")
  robiassert_arg(length(state)==1,
                 arg = "state",
                 message = "value must have a length of one")

   options("ROBITools2.verbose"=value)
}

#' Indicates if ROBITools send messages to the terminal
#'
#' Mainly indicates if messages are sent to the user terminal
#' to indicate computation progression.
#'
#' @return a logical value
#' @export
#'
#' @rdname set_robi_verbose
#' @examples
#'   is_robi_verbose()
#'
is_robi_verbose <- function()
{
    isTRUE(getOption("ROBITools2.verbose")) &&
    interactive() &&
    !isTRUE(getOption("rstudio.notebook.executing")) &&
    !isTRUE(getOption("knitr.in.progress"))
}


#' Generate an error on false assertion.
#'
#' @param predicat a logical value
#' @param message a character used as error message using the
#'   \code{\link[glue]{glue}} formater
#' @param ... supplementary variables used by the error message
#' @param call	logical, indicating if the call should become part of the error
#' @param .envir the environment where glue evaluate expressions
#' @param .abort the methode to call to raise the error. can be
#'               \code{\link[rlang]{abort}}, the default,
#'               \code{\link[rlang]{warm}}, or \code{\link[rlang]{inform}}
#' @return the predicat value
#' @export
#'
#' @examples
#'    robiassert(3==3,"The value {a} is different to value {b}",
#'               a=3,b=4)
#'
robiassert <- function(predicat,message,..., .envir = parent.frame(), .abort = abort) {
  if (!predicat)
    .abort(glue(message,...,.envir = .envir))

  invisible(predicat)
}

#' @param arg a character string indicating the name of the argument generating
#'   the error
#' @rdname robiassert
#' @export
#'
#' @examples
#'    robiassert_arg(3==3,"titi","The value {a} is different to value {b}",
#'                   a=3,b=4)
#'
robiassert_arg <- function(predicat,arg,message,..., .envir = parent.frame(), .abort = abort) {
  message = glue("[{arg}] : {message}",.envir = environment())
  robiassert(predicat,message,...,.envir = .envir, .abort = .abort)
}

#' Writes a message on the console in verbose mode
#'
#' The message is written to the console if the code
#' is executed in the verbose mode. The verbose mode
#' can be specified by declaring a `verbose` logical
#' value to `TRUE`. If no `verbose` variable is defined,
#' the result of the `ROBITools2::is_robi_verbose`
#' function is used to decide of the verbose status.
#'
#'
#' @param message a string template sent to the `glue::glue` preprocessor.
#' @param ... other values to be concatenated to the end of the message
#' @param .envir the environment passed to the  `glue::glue` preprocessor
#'
#' @md
#' @export
#'
#' @examples
#'    verbose = TRUE
#'    robimessage("Hello world !")
#'    verbose = FALSE
#'    robimessage("Hello world !")
#'    rm(verbose)
#'    v_mode = is_robi_verbose()
#'    v_mode
#'    robimessage("Hello world !")
#'    options(ROBITools2.verbose = FALSE)
#'    robimessage("Hello world !")
#'    options(ROBITools2.verbose = TRUE)
#'    robimessage("Hello world !")
#'    options(ROBITools2.verbose = v_mode)
robimessage <- function(message,..., .envir = parent.frame()) {
  verbose = tryCatch(get("verbose",envir = .envir),
                     error = function(e) {
                       print("coucou");is_robi_verbose()})

  if (verbose) {
    message(glue(message,..., .envir = .envir))
  }
}

#' @export
clone <- function(object) {
  UseMethod("clone",object)
}
