#' @importFrom stringr str_sub
NULL


#' Writes a tibble containing sequences in a fasta file.
#'
#' To be writable as a fasta file a `tibble` requires to have
#' a sequence column containing the `character` strings. The
#' other columns that can be used during the formating are:
#'
#' - `id` : used as sequence id located just after the `>`
#'          sign,
#' - `definition` : which is added at the end of the title line,
#' - `features` : the features must be a `tibble` or a `data.frame`,
#'                each column of the table will be translated to
#'                a *key*, *value* feature following the extended
#'                obifasta format used by *OBITools*
#'
#' @param data the `tibble` to format
#' @param file a writable `connection` or a character string naming the
#'        file to write to, or NULL to print on stdout.
#' @param make_unique insure that the ids of the sequences are unique
#' @param verbose    if `TRUE` warnings and messages are emitted when the
#'   function takes some decision.
#'
#' @export
#'
#' @md
#' @examples
#'    filename <- system.file("extdata", "B5.fasta.gz",
#'                            package = "ROBITools2")
#'
#'    seqs <- read_obifasta(filename)
#'    write_obifasta(seqs)
#'
write_obifasta <- function(data,file = NULL,
                           make_unique = TRUE,
                           verbose = is_robi_verbose()) {

 robiassert("sequence" %in% colnames(data),
            "to be written in fasta format data must contain a `sequence` column",
            .abort = stop)

 fmt_data <- function(v) {
    if (is.list(v)) {
       v <- v[[1]]
       if (is.null(names(v)))
          paste0("[",paste(v,collapse = ", "),"]")
       else
          paste0("{",paste(paste0("'",names(v),"'"),
                           v,
                           sep = ":",collapse = ", "),"}")
    } else {
       as.character(v)
    }
 }

 if (! "id" %in% colnames(data))  {
   ids <- new_robiuniqueid(rep("seq",nrow(data)),
                              make_unique = TRUE)
 } else {
   ids <- if (make_unique)
     new_robiuniqueid(data$id,
                      make_unique = TRUE)
   else data$id
 }

 feature_names <- colnames(data$features)
 is_a_list <- sapply(data$features,is.list)

 suppressWarnings(sink(file))
 for (i in seq_len(nrow(data))) {
    b <- seq(from = 1,
             to = nchar(data$sequence[i]),
             by = 60)

    f <- data$features[i,]
    valid <- ! sapply(f,is.na)

    writeLines(paste0(">",ids[i]," ",
                     paste(feature_names[valid],
                           sapply(f[,valid],fmt_data),
                           sep="=", collapse = "; "),
                     "; ",
                     data$definition[i]))

    writeLines(paste0(str_sub(data$sequence[i],
                              start = b,
                              end = b + 59),
                              collapse = "\n"),
                     )
 }
 suppressWarnings(sink())
}
