#include <R.h>
#include <Rinternals.h>
#include <Rdefines.h>
#include <stdint.h>
#include <string.h>
#include <stdio.h>

typedef struct {
 	int32_t  is_scientificname;
	int32_t  namelength;
	int32_t  classlength;
	int32_t  taxid;
	char     names[1];
} econameformat_t;

typedef const econameformat_t* econameformat_p;

typedef struct {
	int32_t  taxid;
	int32_t  rank;
	int32_t	 parent;
	int32_t  namelength;
	char     name[1];
} ecotxformat_t;

typedef const ecotxformat_t* ecotxformat_p;


#define SWAPINT32(x)     ((((x) << 24) & 0xFF000000) | (((x) <<  8) & 0xFF0000) | \
                         (((x) >>  8) & 0xFF00)     | (((x) >> 24) & 0xFF))


static int32_t is_big_endian()
{
	int32_t i=1;
	return (int32_t)((char*)&i)[0];
}

/*
static int32_t swap_int32_t(int32_t i)
{
	return SWAPINT32(i);
}
*/

SEXP R_raw_to_ecopcr_record(SEXP raw)
{
  int32_t   big;
	int32_t   size;
  int32_t   i;
  int32_t   rsize;
	const void* rawdata;
	const void* pointer;
	void* precord;

	SEXP  record_list;
	SEXP  record;

	big = is_big_endian();

  if (! IS_RAW(raw))
      error("the raw parameter must be of raw type");

	rawdata = RAW(raw);

	size = *((const int32_t*) rawdata);

	if (big) size = SWAPINT32(size);


	record_list = PROTECT(NEW_LIST(size));
	pointer = rawdata + sizeof(int32_t);

  for(i=0; i < size; i++) {
	  rsize = *((const int32_t*) pointer);
	  if (big) rsize = SWAPINT32(rsize);

	  pointer += sizeof(int32_t);

	  record = PROTECT(NEW_RAW(rsize));
	  precord = RAW(record);
	  memcpy(precord,pointer,rsize);
	  SET_VECTOR_ELT(record_list, i, record);
	  UNPROTECT(1);

	  pointer += rsize;
	}


  UNPROTECT(1);

	return record_list;
}

SEXP R_raw_to_ecopcr_tnames(SEXP raw)
{
  int32_t   big;
	int32_t   size;
	int32_t   rsize;
  int32_t   i;
  int32_t   name_length;
  int32_t   class_length;
  int32_t   taxid;

  char buffer[1000];

	const void* rawdata;
	const void* pointer;
	econameformat_p rpointer;

	SEXP  name_table;
	SEXP  names;
	SEXP  classes;
	SEXP  taxids;

	big = is_big_endian();

  if (! IS_RAW(raw))
      error("the raw parameter must be of raw type");

  name_table = PROTECT(NEW_LIST(3));

	rawdata = RAW(raw);

	size = *((const int32_t*) rawdata);

	if (big) size = SWAPINT32(size);


	names   = PROTECT(NEW_STRING(size));
	classes = PROTECT(NEW_STRING(size));
	taxids  = PROTECT(NEW_INTEGER(size));

	pointer = rawdata + sizeof(int32_t);


  for (i=0; i < size; i++) {
	  rsize    = *((const int32_t*) pointer);
    pointer += sizeof(int32_t);
    rpointer = (econameformat_p) pointer;

  	  if (big) {
  	    rsize        = SWAPINT32(rsize);

  	    name_length  = SWAPINT32(rpointer->namelength);
  	    class_length = SWAPINT32(rpointer->classlength);
  	    taxid        = SWAPINT32(rpointer->taxid);
  	  }
  	  else {
  	    name_length  = rpointer->namelength;
  	    class_length = rpointer->classlength;
  	    taxid        = rpointer->taxid;
  	  }

	  if (name_length >= 1000)
	    error("Taxon name must be less than 1000 characters long");

	  memcpy(buffer,rpointer->names,name_length);
    buffer[name_length]=0;

	  SET_STRING_ELT(names, i,
                   mkChar(buffer));


	  if (class_length >= 1000)
	    error("Taxon class name must be less than 1000 characters long");

	  memcpy(buffer,
           (rpointer->names + name_length),
           class_length);
    buffer[class_length]=0;

	  SET_STRING_ELT(classes, i,
                   mkChar(buffer));


	  SET_INTEGER_ELT(taxids,i,taxid);

	  pointer += rsize;
	}

  SET_VECTOR_ELT(name_table,0,taxids);
  SET_VECTOR_ELT(name_table,1,classes);
  SET_VECTOR_ELT(name_table,2,names);

  UNPROTECT(4);

	return name_table;
}

SEXP R_raw_to_ecopcr_tnodes(SEXP raw)
{
  int32_t   big;
	int32_t   size;
	int32_t   rsize;
  int32_t   i;
  int32_t   taxid;
  int32_t   rank_idx;
  int32_t   parent_idx;

	const void* rawdata;
	const void* pointer;
	ecotxformat_p rpointer;

	SEXP  taxon_table;
	SEXP  taxids;
	SEXP  parents;
	SEXP  ranks;

	big = is_big_endian();

  if (! IS_RAW(raw))
      error("the raw parameter must be of raw type");

  taxon_table = PROTECT(NEW_LIST(3));

	rawdata = RAW(raw);

	size = *((const int32_t*) rawdata);

	if (big) size = SWAPINT32(size);


	taxids   = PROTECT(NEW_INTEGER(size));
	parents  = PROTECT(NEW_INTEGER(size));
	ranks    = PROTECT(NEW_INTEGER(size));

	pointer = (char*)rawdata + sizeof(int32_t);

  for (i=0; i < size; i++) {
	  rsize    = *((const int32_t*) pointer);
    pointer += sizeof(int32_t);
    rpointer = (ecotxformat_p) pointer;

  	  if (big) {
  	    rsize        = SWAPINT32(rsize);

  	    taxid      = SWAPINT32(rpointer->taxid);
  	    parent_idx = SWAPINT32(rpointer->parent);
  	    rank_idx   = SWAPINT32(rpointer->rank);
  	  }
  	  else {
  	    taxid      = rpointer->taxid;
  	    parent_idx = rpointer->parent;
  	    rank_idx   = rpointer->rank;
  	  }

	  SET_INTEGER_ELT(taxids,i,taxid);
	  SET_INTEGER_ELT(parents,i,parent_idx);
	  SET_INTEGER_ELT(ranks,i,rank_idx);

	  pointer += rsize;
	}

  SET_VECTOR_ELT(taxon_table,0,taxids);
  SET_VECTOR_ELT(taxon_table,1,parents);
  SET_VECTOR_ELT(taxon_table,2,ranks);

  UNPROTECT(4);

	return taxon_table;
}

SEXP R_compute_taxonomic_path(SEXP taxids, SEXP ranks, SEXP parents) {

  SEXP paths;
  SEXP rpaths;
  SEXP path;
  SEXP rpath;
  SEXP spath;
  SEXP srpath;

  SEXP results;

  int32_t n;
  int32_t p;
  int32_t i,j;

  if (LENGTH(taxids) != LENGTH(parents) || LENGTH(taxids) != LENGTH(ranks))
    error("taxids, tanks, and parents parameters must have the same lenght");

  n = LENGTH(taxids);
  paths = PROTECT(NEW_LIST(n));
  rpaths = PROTECT(NEW_LIST(n));

  path = PROTECT(NEW_INTEGER(1000));
  rpath = PROTECT(NEW_INTEGER(1000));

  for (i=0; i < n; i++) {
    p = INTEGER(parents)[i] - 1;
    j = 0;
    INTEGER(path)[j] = INTEGER(taxids)[p];
    INTEGER(rpath)[j] = INTEGER(ranks)[p];

    while (p != 0) {
      j = j + 1;
      p = INTEGER(parents)[p] - 1;
      INTEGER(path)[j] = INTEGER(taxids)[p];
      INTEGER(rpath)[j] = INTEGER(ranks)[p];
    }

    j++;

    spath = PROTECT(NEW_INTEGER(j));
    memcpy(INTEGER(spath),INTEGER(path),j*sizeof(int));
    SET_VECTOR_ELT(paths,i,spath);

    srpath = PROTECT(NEW_INTEGER(j));
    memcpy(INTEGER(srpath),INTEGER(rpath),j*sizeof(int));
    SET_VECTOR_ELT(rpaths,i,srpath);

    UNPROTECT(2);
  }

  results = PROTECT(NEW_LIST(2));
  SET_VECTOR_ELT(results,0,paths);
  SET_VECTOR_ELT(results,1,rpaths);

  UNPROTECT(5);

  return results;
}
