#' @import utils
#' @import readr
#' @include taxonomy_robi.R
#' @useDynLib ROBITools2
#'
NULL

.NCBIURL <- "https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz"


#' Untar a NCBI taxdump archive in a temporary directory.
#'
#' @param archive
#' @param verbose a logical indicating if progession messages have to be printed
#'   on the console.
#'
#' @return the name of the temporary directory containing the extracted archive.
#'
#' @rdname internal.untar_ncbi_taxdump
#' @examples
.untar_ncbi_taxdump <- function(archive,
                               verbose = is_robi_verbose()) {

  destdir <- tempfile(pattern = "taxdump_", fileext = ".dir")
  success <- untar(tarfile = archive, exdir = destdir)

  if (success > 0) {
    if (file.exists(destdir)) {
      unlink(destdir,
             recursive = TRUE,
             force = TRUE)
    }
    stop("Failing to untar the taxdump file downladed from NCBI")
  }

  destdir
}

#' Download the NCBI taxdump archive.
#'
#' @param url a string indicating the URL to download
#' @param verbose a logical indicating if progession messages have to be printed
#'   on the console.
#'
#' @return the name of the temporary directory containing the extracted archive.
#' @export
#'
#' @examples
#'     taxdump <- download_ncbi_taxdump()
#'     file.remove(taxdump)
#'
download_ncbi_taxdump <- function(url = .NCBIURL,
                                  verbose = is_robi_verbose()) {
  destfile <- tempfile(pattern = "taxdump_", fileext = ".tar.gz")

  success <- if (verbose)
                download.file(url = url, destfile = destfile)
             else
               suppressMessages(download.file(url = url, destfile = destfile))

  if (success > 0) {
    if (file.exists(destfile))
      file.remove(destfile)

    stop("Failing to download taxdump file from NCBI")
  }

  destdir <- .untar_ncbi_taxdump(destfile,
                                 verbose = verbose)

  file.remove(destfile)

  destdir
}

#' Reads the nodes file.
#'
#' The nodes file describes the topology of the taxonomy tree.
#' For every taxa, its parent taxon an toxonomic rank is stored.
#'
#' @param directory a string pointing to the directory containing the file
#' @param file a string corresponding to the file name in the directory
#'
#' @return a tibble instance.
#'
#' @rdname internal.read_ncbi_nodes_dump
#' @examples
.read_ncbi_nodes_dump <- function(directory = ".", file = "nodes.dmp") {
  read_delim(file.path(directory,file),
            delim = "|",
            trim_ws = TRUE,
            col_types = cols( taxid = col_integer(),
                              parent = col_integer(),
                              rank = col_factor(),
                              embl_code = col_skip(),
                              division = col_skip(),
                              inherited_div = col_skip(),
                              genetic_code = col_skip(),
                              inherited_GC = col_skip(),
                              mito_genetic_code = col_skip(),
                              inherited_mito_GC = col_skip(),
                              hidden = col_skip(),
                              hiden_subtree = col_skip(),
                              comment = col_skip(),
                              null = col_skip()
                            ),
            col_names = c("taxid","parent","rank",
                          "embl_code","division","inherited_div",
                          "genetic_code","inherited_GC",
                          "mito_genetic_code","inherited_mito_GC",
                          "hidden","hiden_subtree",
                          "comment","null")
            )
}

#' Reads the taxa names file.
#'
#' The taxa names file contains the names for all the taxa.
#' For each taxon several names are recorded. Among them the
#' scientific name as a special status and is always kept in
#' a \code{\link[ROBITools2]{robitaxonomy}} instance.
#'
#' @param directory a string pointing to the directory containing the file
#' @param file a string corresponding to the file name in the directory
#'
#' @return a tibble instance.
#'
#' @rdname internal.read_ncbi_names_dump
#' @examples
.read_ncbi_names_dump <- function(directory = ".", file = "names.dmp") {
  readr::read_delim(file.path(directory,file),
                  delim = "|",
                  trim_ws = TRUE,
                  col_types = cols(taxid = col_integer(),
                                   name  = col_character(),
                                   uniq_name = col_skip(),
                                   name_class = col_factor(),
                                   null = col_skip()
                                   ),
                  col_names = c("taxid","name",
                                "uniq_name","name_class",
                                "null")
                  ) %>%
  dplyr::mutate(taxid = new_robitaxid_master(taxid)) %>%
  dplyr::select(name_class,name,taxid)
}

#' Reads the deleted nodes file.
#'
#' The deleted nodes file contains information aboiut taxids that
#' doesn't exist anymore and that have not be transfered to a new taxid.
#'
#' @param directory a string pointing to the directory containing the file
#' @param file a string corresponding to the file name in the directory
#'
#' @return a two columns tibble instance.
#'
#' @rdname internal.read_ncbi_delnodes_dump
#' @examples
.read_ncbi_delnodes_dump <- function(directory = ".", file = "delnodes.dmp") {
  readr::read_delim(file.path(directory,file),
                  delim = "|",
                  trim_ws = TRUE,
                  col_types = cols(taxid = col_integer(),
                                   null = col_skip()
                                  ),
                  col_names = c("old_taxid","null")
                  ) %>%
    dplyr::mutate(old_taxid = new_robitaxid(old_taxid),
                  new_taxid = robitaxid_master(NA))
}

#' Reads the merged nodes file.
#'
#' The merged nodes file contains information aboiut taxids that
#' doesn't exist anymore but have to transfered to a new taxid.
#'
#' @param directory a string pointing to the directory containing the file
#' @param file a string corresponding to the file name in the directory
#'
#' @return a two columns tibble instance.
#'
#' @rdname internal.read_ncbi_merged_dump
#' @examples
.read_ncbi_merged_dump <- function(directory = ".", file = "merged.dmp") {
  dplyr::read_delim(file.path(directory,file),
                  delim = "|",
                  trim_ws = TRUE,
                  col_types = cols(old_taxid = col_integer(),
                                   new_taxid = col_integer(),
                                   null = col_skip()
                                  ),
                  col_names = c("old_taxid","null")
                  ) %>%
    dplyr::mutate(old_taxid = new_robitaxid(old_taxid),
                  new_taxid = new_robitaxid_master(new_taxid))
}

#' Reads a NCBI taxonomy databse dump.
#'
#' NCBI provides ASCII dump of their database. They can be downloaded from their
#' FTP site as a compressed (gzip) tar archive at the following URL:
#'
#' > https://ftp.ncbi.nlm.nih.gov/pub/taxonomy/taxdump.tar.gz
#'
#' The function can download the archive, untar it or only reads the files from
#' a directory according to the string passed through the parametter
#' \code{file}.
#'
#' @param file a string pointing to a directory, a tar file, or an URL
#' @param nodes_dump the name of the dump of the nodes constituting the tree.
#' @param name_dump the name of the dump containing the names of the taxa.
#' @param keep_alternative_names a logical indicating if at the end of the
#'   process names, other than the scientific names have to be kept in memory.
#' @param keep_alternative_taxids a logical indicating if deleted taxids and
#'   merged taxids have to be stored to allow for looking in them and converting
#'   them to current taxids.
#' @param verbose a logical indicating if progession messages have to be printed
#'   on the console.
#'
#' @return a \code{\link[ROBITools2]{robitaxonomy}} instance.
#'
#' @export
#'
#' @examples
#'
#' taxo <- read_ncbi_taxdump()
#'
read_ncbi_taxdump <- function(file = .NCBIURL,
                              keep_alternative_names = FALSE,
                              keep_alternative_taxids = FALSE,
                              nodes_dump = "nodes.dmp",
                              name_dump = "names.dmp",
                              verbose = is_robi_verbose()) {

  to_be_cleaned = FALSE

  if (dir.exists(file))
    directory <- file
  else {
    to_be_cleaned = TRUE

    if (file.exists(file) &&
        str_ends(file, pattern = "\\.(tar|tar\\.gz|tgz)")) {
      robimessage("Untaring the taxonomy taxdump... ")

      directory <- .untar_ncbi_taxdump(file,
                                       verbose = verbose)
    }
    else if (str_starts(file, "(https?|ftp)://")) {
      robimessage("Downloading the taxonomy from NCBI web site... ")

      directory <- download_ncbi_taxdump(file,
                                         verbose = verbose)
    }
    else
      stop("[file] : doesn't seem to correspond neither to a directory, a tar archive neither to an URL")
  }


  robimessage("Reading the taxonomy tree... ")

  nodes <- .read_ncbi_nodes_dump(directory,
                                 file = nodes_dump)

  robimessage("Done.")


  robimessage("Reading the taxon names... ")

  names <- .read_ncbi_names_dump(directory,
                                 file = name_dump)

  robimessage("Done.")

  if (to_be_cleaned)
    unlink(directory,recursive = TRUE)

  robimessage("Indexing scientific names... ")

  suppressWarnings(names %>%
      dplyr::filter(name_class == "scientific name") %>%
      dplyr::right_join(nodes, by = "taxid")) %>%
  dplyr::mutate(taxid = new_robitaxid_master(taxid),
                parent = new_robitaxid(parent)) %>%
  dplyr::select(taxid,
                rank,
                scientific_name=name,
                parent) -> taxonomy

  robimessage("Done.")

  robimessage("Indexing parent taxids... ")

  o <- order(nodes$taxid)
  names(o) <- nodes$taxid
  relparent <- o[as.character(nodes$parent)]

  robimessage("Done.")

  robimessage("Computing path... ")

  paths <- .Call("R_compute_taxonomic_path",
                 nodes$taxid,
                 nodes$rank,
                 relparent)

  paths <- list(path =  lapply(paths[[1]],new_robiobject,
                               class = "robipath"),
                rpath = paths[[2]])

  robimessage("Done.")

  taxonomy$path <- paths[[1]]
  taxonomy$rpath <- paths[[2]]

  if (!keep_alternative_names)
    names = NULL

  new_robitaxonomy(taxonomy,
                   alternative_names = names,
                   reference_taxonomy = NULL)
}

